AC_DEFUN([NORDUGRID_INIT],
[

AC_MSG_CHECKING([for SYSV init scripts])
AC_ARG_WITH(sysv-scripts-location,
  [  --with-sysv-scripts-location=<PATH>       Location of the SYSV init scripts (e.g. /etc/init.d)],
  [ initddir="$withval" ],
  [
    initddir=
    for i in /etc/init.d /etc/rc.d/init.d; do
      if test -d "$i" ; then
	initddir="$i"
	break
      fi
    done
  ]
)
if test ! -d "$initddir"; then
  AC_MSG_ERROR(could not find a suitable location for the SYSV init scripts)
fi
AC_MSG_RESULT($initddir)
AC_SUBST(initddir)

AC_ARG_WITH(monitor-prefix,
  [  --with-monitor-prefix=<PATH>         Specify the location of the Grid Monitor WWW scripts],
  [ monitor_prefix="$withval" ],
  [ monitor_prefix="/var/www/html/gridmonitor" ]
)
AC_SUBST(monitor_prefix)

AC_ARG_WITH(monitor-local-prefix,
  [  --with-monitor-local-prefix=<PATH>   Specify the relative location of the Grid Monitor WWW scripts],
  [ monitor_local_prefix="$withval" ],
  [ monitor_local_prefix=`basename $monitor_prefix` ]
)
AC_SUBST(monitor_local_prefix)

dnl
dnl gSOAP
dnl
https_prefix="# "
https_init_script=""
AC_ARG_WITH(gsoap-location,
  [  --with-gsoap-location=<PATH>         Specify the gSOAP installation path],
  [ AC_PATH_PROG(soapcpp2, soapcpp2, not found, $with_gsoap_location/bin) ],
  [ AC_PATH_PROG(soapcpp2, soapcpp2, not found) ]
)
if test "x$soapcpp2" = "xnot found"; then
  gsoap_location=""
else
  gsoap_location=`echo "$soapcpp2" | sed 's%/bin/soapcpp2$%%;t;s%.*%%'`
  SOAPCPP2=$gsoap_location/bin/soapcpp2
  WSDL2H=$gsoap_location/bin/wsdl2h
  if test "x$gsoap_location" = "x/usr"; then
    :
  else
    GSOAP_INCLUDES=-I$gsoap_location/include
    AC_SUBST(GSOAP_INCLUDES)
  fi
dnl Is the gSoap source available?
  if test -r "$gsoap_location/include/stdsoap2.cpp"; then
    GSOAP_LIBS='$(top_builddir)/grid-manager/external/libgsoap++.la'
  else
    GSOAP_LIBS=$gsoap_location/lib/libgsoap++.a
  fi
  AM_CONDITIONAL(HAVE_GSOAP_SOURCE, test -r "$gsoap_location/include/stdsoap2.cpp")
  https_prefix=
  https_init_script=httpsd
fi
AC_SUBST(https_init_script)
GSOAP_LOCATION="$gsoap_location"
if test "x$soapcpp2" = "x"; then
  AC_MSG_ERROR([soapcpp2 is required. Please install gSOAP.])
fi

dnl
dnl gSOAP version
dnl
AC_MSG_CHECKING(for gSOAP version)
gsoap_version=`$gsoap_location/bin/soapcpp2 -v 2>&1 | sed 's/.*version \([[^ ]]*\).*/\1/;t new;s/^.*Compiler for C and C++ *\([[^ ]]*\).*/\1/;t matched;s/.*//;:matched;:new' | grep '.'`
if test "x$gsoap_version" = "x"; then
  AC_MSG_ERROR([Could not determine version.])
fi
gsoap_version_major=`echo "$gsoap_version" | sed 's/^\([[^\.]]\).*/\1/;t match;s/.*/0/;:match'`
gsoap_version_minor=`echo "$gsoap_version" | sed 's/^[[^\.]]*\.\([[^\.]]*\).*/\1/;t match;s/.*/0/;:match'`
gsoap_version_subminor=`echo "$gsoap_version" | sed 's/^[[^\.]]*\.[[^\.]]*\.\([[^\.]]*\).*/\1/;t match;s/.*/0/;:match'`
if test "$gsoap_version_major" -lt "2"; then
  AC_MSG_RESULT($gsoap_version_major.$gsoap_version_minor)
  AC_MSG_ERROR([gSOAP is too old])
fi
if test "$gsoap_version_major" -eq "2"; then
  if test "$gsoap_version_minor" -lt "5" ; then
    AC_MSG_ERROR([gSOAP with version below 2.5 is not supported anymore])
  else
    if test "$gsoap_version_minor" -eq "5" ; then
      if test "$gsoap_version_minor" -lt "2" ; then
        AC_MSG_ERROR([gSOAP with version below 2.5.2 has a bug critical for this application])
      fi
    fi
  fi
fi
AC_MSG_RESULT($gsoap_version_major.$gsoap_version_minor.$gsoap_version_subminor)
AC_SUBST(https_prefix)
AC_SUBST(SOAPCPP2)
AC_SUBST(WSDL2H)
AC_SUBST(GSOAP_LIBS)
AC_SUBST(GSOAP_LOCATION)

AC_ARG_WITH(mysql-location,
  [  --with-mysql-location=<PATH>         Specify the MySQL installation path],
  [ AC_PATH_PROG(mysql_config, mysql_config, not found, $with_mysql_location/bin) ],
  [ AC_PATH_PROG(mysql_config, mysql_config, not found) ]
)
MYSQL_INCLUDES=
MYSQL_LIBS=
if test "x$mysql_config" != "xnot found"; then
  AC_MSG_CHECKING(for mysql includes)
  mysql_include=`$mysql_config --cflags | sed "s/^-I\([[^ ]]*\).*/\1\/mysql.h/;s/'//g"`
  if ! test -f "$mysql_include"; then
    AC_MSG_RESULT(not found)
    mysql_config=""
  else
    AC_MSG_RESULT(found)
    MYSQL_INCLUDES=`$mysql_config --cflags | sed "s/'//g"`
    MYSQL_LIBS=`$mysql_config --libs | sed "s/'//g"`
  fi
else
  mysql_config=""
fi
AC_SUBST(MYSQL_INCLUDES)
AC_SUBST(MYSQL_LIBS)

XML_LIBS=
AC_ARG_WITH(xml-location,
  [  --with-libxml-location=<PATH>        Specify the libxml2 installation path],
  [ AC_PATH_PROG(xml_config, xml2-config, not found, $with_libxml_location/bin) ],
  [ AC_PATH_PROG(xml_config, xml2-config, not found) ]
)
if test "x$xml_config" != "xnot found"; then
  XML_INCLUDES=`$xml_config --cflags`
  # Remove /usr/include
  XML_INCLUDES=`echo $XML_INCLUDES | sed -e 's, *-I/usr/include/*$, ,g' -e 's, *-I/usr/include/* , ,g'`
  XML_LIBS=`$xml_config --libs`
fi
AC_SUBST(XML_INCLUDES)
AC_SUBST(XML_LIBS)
AM_CONDITIONAL(HAVE_LIBXML, test "x$xml_config" != "xnot found")

dnl
dnl GACL_LOCATION
dnl
gridsite=yes
DEFAULT_GRIDSITE_LOCATION=/opt/gridsite
AC_SUBST(DEFAULT_GRIDSITE_LOCATION)
AC_MSG_CHECKING(for GACL)
AC_ARG_WITH(gacl-location,
  [  --with-gacl-location=<PATH>          Specify the GACL installation path],
  [ GACL_LOCATION=$with_gacl_location ],
  [
    if test "x$GACL_LOCATION" = "x"; then
      GACL_LOCATION=$DEFAULT_GRIDSITE_LOCATION
    fi
  ]
)
if test ! -d $GACL_LOCATION; then
  GACL_LOCATION="$PWD/grid-manager/external/gacl-0.1.5"
  gridsite=
else
  if test ! -f $GACL_LOCATION/lib/libgridsite.a; then
    GACL_LOCATION="$PWD/grid-manager/external/gacl-0.1.5"
    gridsite=
  fi
fi
AC_MSG_RESULT($GACL_LOCATION)  
AC_SUBST(GACL_LOCATION)
AM_CONDITIONAL(NG_GACL, test "x$gridsite" = "x")
AM_CONDITIONAL(GRIDSITE_GACL, test "x$gridsite" != "x")
if test "x$gridsite" = "x"; then
  AC_DEFINE(NG_GACL,,[GACL patched by NG])
else
  AC_DEFINE(GRIDSITE_GACL,,[Gridsite GACL (also patched by NG)])
fi

dnl
dnl NORDUGRID_LOCATION
dnl
AC_ARG_WITH(nordugrid-location,
  [  --with-nordugrid-location=<PATH>     Specify the nordugrid installation path],
  [ prefix=$with_nordugrid_location ],
  [ prefix=/usr ]
)
AC_DEFINE_UNQUOTED(DEFAULT_NORDUGRID_LOCATION,"$prefix",[Default NorduGrid installation path])

dnl
dnl GLOBUS_LOCATION
dnl
DEFAULT_GLOBUS_LOCATION=/opt/globus
AC_DEFINE_UNQUOTED(DEFAULT_GLOBUS_LOCATION,"$DEFAULT_GLOBUS_LOCATION",[Default Globus installation path])
AC_SUBST(DEFAULT_GLOBUS_LOCATION)
AC_MSG_CHECKING(for GLOBUS_LOCATION)
AC_ARG_WITH(globus-location,
  [  --with-globus-location=<PATH>        Specify the globus installation path],
  [
    GLOBUS_LOCATION="$withval"
    export GLOBUS_LOCATION
  ],
  [
    if test "x$GLOBUS_LOCATION" = "x"; then
      GLOBUS_LOCATION=$DEFAULT_GLOBUS_LOCATION
      export GLOBUS_LOCATION
    fi
  ]
)
if test ! -d $GLOBUS_LOCATION; then
  AC_MSG_ERROR(could not find GLOBUS_LOCATION)
fi
AC_MSG_RESULT($GLOBUS_LOCATION)  
AC_SUBST(GLOBUS_LOCATION)

dnl
dnl GPT_LOCATION
dnl
DEFAULT_GPT_LOCATION=/opt/gpt
AC_SUBST(DEFAULT_GPT_LOCATION)
AC_MSG_CHECKING(for GPT_LOCATION)
AC_ARG_WITH(gpt-location,
  [  --with-gpt-location=<PATH>           Specify the gpt installation path],
  [
    if test "x$withval" = "x"; then
      if test "x$GPT_LOCATION" = "x"; then
	GPT_LOCATION=$DEFAULT_GPT_LOCATION
	if test ! -d $GPT_LOCATION; then
	  GPT_LOCATION=$GLOBUS_LOCATION
	fi
	export GPT_LOCATION
      fi
    else
      GPT_LOCATION="$withval"
      export GPT_LOCATION
    fi
  ],
  [
    if test "x$GPT_LOCATION" = "x"; then
      GPT_LOCATION=$DEFAULT_GPT_LOCATION
      if test ! -d $GPT_LOCATION; then
	GPT_LOCATION=$GLOBUS_LOCATION
      fi
      export GPT_LOCATION
    fi
  ]
)
AC_MSG_RESULT($GPT_LOCATION)
AC_SUBST(GPT_LOCATION)

dnl
dnl globus-makefile-header
dnl
AC_ARG_WITH(globus-makefile-header,
  [  --with-globus-makefile-header        Use globus-makefile-header to generate makefile stubs],
  [
    if test "x$withval" = "xno" ; then
      USE_GLOBUS_GPT=.pre
      GLOBUS_MAKEFILE_HEADER=:
    else
      AC_CHECK_PROG(GLOBUS_MAKEFILE_HEADER, globus-makefile-header, $GLOBUS_LOCATION/bin/globus-makefile-header, not found, $GLOBUS_LOCATION/bin)
      if test "x$GLOBUS_MAKEFILE_HEADER" = "xnot found"; then
	AC_CHECK_PROG(GLOBUS_MAKEFILE_HEADER, globus-makefile-header, $GLOBUS_LOCATION/sbin/globus-makefile-header, not found, $GLOBUS_LOCATION/sbin)
	if test "x$GLOBUS_MAKEFILE_HEADER" = "xnot found"; then
	  USE_GLOBUS_GPT=.pre
	  GLOBUS_MAKEFILE_HEADER=:
	else
	  USE_GLOBUS_GPT=
	fi
      else
	USE_GLOBUS_GPT=
      fi
    fi
  ],
  [
    AC_CHECK_PROG(GLOBUS_MAKEFILE_HEADER, globus-makefile-header, $GLOBUS_LOCATION/bin/globus-makefile-header, not found, $GLOBUS_LOCATION/bin)
    if test "x$GLOBUS_MAKEFILE_HEADER" = "xnot found"; then
      AC_CHECK_PROG(GLOBUS_MAKEFILE_HEADER, globus-makefile-header, $GLOBUS_LOCATION/sbin/globus-makefile-header, not found, $GLOBUS_LOCATION/sbin)
      if test "x$GLOBUS_MAKEFILE_HEADER" = "xnot found"; then
	USE_GLOBUS_GPT=.pre
	GLOBUS_MAKEFILE_HEADER=:
      else
	USE_GLOBUS_GPT=
      fi
    else
      USE_GLOBUS_GPT=
    fi
  ]
)
AC_SUBST(GLOBUS_MAKEFILE_HEADER)

AC_ARG_WITH(flavor,
  [  --with-flavor=<flavor>               Specify the globus build flavor],
  [
    GLOBUS_FLAVOR_NAME=$withval
    if test ! -f "$GLOBUS_LOCATION/etc/globus_core/flavor_$GLOBUS_FLAVOR_NAME.gpt"; then
      echo "ERROR: Flavor $GLOBUS_FLAVOR_NAME has not been installed" >&2
      exit 1
    fi
  ],
  [
    if test "x$GLOBUS_FLAVOR_NAME" = "x"; then
      AC_CHECK_PROG(GLOBUS_FLAVOR_CONF, globus-flavor-configuration, $GLOBUS_LOCATION/sbin/globus-flavor-configuration, not found, $GLOBUS_LOCATION/sbin)
      if test "x$GLOBUS_FLAVOR_CONF" = "xnot found"; then
	AC_CHECK_PROG(GPT_FLAVOR_CONF, gpt-flavor-configuration, $GPT_LOCATION/sbin/gpt-flavor-configuration, not found, $GPT_LOCATION/sbin)
	if test "x$GPT_FLAVOR_CONF" = "xnot found"; then
	  echo "Please specify a globus build flavor" >&2
  	  exit 1
	fi
	FLAVOR_CONF=$GPT_FLAVOR_CONF
      else
	FLAVOR_CONF=$GLOBUS_FLAVOR_CONF
      fi
      GLOBUS_FLAVOR_NAME=`$FLAVOR_CONF | grep '^[[a-zA-Z]].*:$' | cut -f1 -d: | grep thr | tail -1`
    fi
  ]
)
AC_MSG_CHECKING(for Globus flavor)
AC_MSG_RESULT($GLOBUS_FLAVOR_NAME)
AC_SUBST(GLOBUS_FLAVOR_NAME)

GLOBUS_INCLUDES="-I$GLOBUS_LOCATION/include/$GLOBUS_FLAVOR_NAME -I$GLOBUS_LOCATION/include/$GLOBUS_FLAVOR_NAME/openssl"
AC_SUBST(GLOBUS_INCLUDES)

CPPFLAGS=$GLOBUS_INCLUDES
AC_CHECK_HEADERS([globus_gsi_proxy.h])
AC_CHECK_HEADERS([globus_replica_catalog.h])
AC_CHECK_HEADERS([globus_rls_client.h])

AC_MSG_CHECKING(for globus_io version)
IO_VERSION=`$GPT_LOCATION/sbin/gpt-query -name globus_io 2>/dev/null | \
  grep 'ver\(sion\)\?:' 2>/dev/null | head -n 1 2>/dev/null | \
  sed 's/^.*ver\(sion\)\?: \([[^ ]]*\).*/\2/' 2>/dev/null`
IO_VERSION_MAJOR=`echo "$IO_VERSION" | sed 's/^\([[^.]]\).*/\1/'`;
if test "x$IO_VERSION_MAJOR" = "x"; then
  IO_VERSION_MAJOR=0
fi
AC_DEFINE_UNQUOTED(GLOBUS_IO_VERSION,$IO_VERSION_MAJOR,[Globus IO version])
#if test "$IO_VERSION_MAJOR" -lt "4"; then
#  AC_DEFINE(OLD_GLOBUS_IO,,[Globus IO version])
#fi
AC_MSG_RESULT($IO_VERSION)

dnl
dnl Enable VOMS
dnl
voms=true
ng_voms=
edg_voms=
AC_ARG_ENABLE(voms,
  [  --disable-voms                     Disable VOMS support],  
  [
    case "${enableval}" in
      yes) voms=true ;;
      no)  voms=false; AC_MSG_WARN(VOMS will be disabled) ;;
      *)   AC_MSG_ERROR(bad value ${enableval} for --disable-voms) ;;
    esac
  ],
  [ voms=true ]
)

if test "x$voms" = xtrue ; then

dnl
dnl VOMS_LOCATION
dnl
DEFAULT_VOMS_LOCATION=/opt/voms
AC_SUBST(DEFAULT_VOMS_LOCATION)
AC_ARG_WITH(voms-location,
  [  --with-voms-location=<PATH>          Specify the VOMS installation path],
  [ 
    VOMS_LOCATION=$with_voms_location
    if test ! -d $VOMS_LOCATION; then
      AC_MSG_WARN([VOMS_LOCATION ($VOMS_LOCATION) does not exist])
      VOMS_LOCATION=
    fi
  ],[
    if test "x$VOMS_LOCATION" = "x"; then
      VOMS_LOCATION=$DEFAULT_VOMS_LOCATION
    fi
    if test ! -d $VOMS_LOCATION; then
      VOMS_LOCATION=
    fi
  ]
)

dnl
dnl Check for VOMS api availability and type
dnl
VOMS_INCLUDES=""
if test "x$VOMS_LOCATION" != "x"; then
  VOMS_INCLUDES=-I$VOMS_LOCATION/include
fi
AC_SUBST(VOMS_INCLUDES)

VOMS_LIBS=
VOMS_FLAVOR=
CPPFLAGS=$VOMS_INCLUDES
AC_LANG(C++)
AC_CHECK_HEADERS([voms_api.h],
  CPPFLAGS="$VOMS_INCLUDES -I$GLOBUS_LOCATION/include/$GLOBUS_FLAVOR_NAME"
  [AC_CHECK_HEADERS([api_util.h],
    ng_voms=true,
    edg_voms=true
    voms_flavor="_$GLOBUS_FLAVOR_NAME"
  )]
  LDFLAGS=
  if test "x$VOMS_LOCATION" != "x"; then
    LDFLAGS=-L$VOMS_LOCATION/lib
  fi
  AC_CHECK_LIB(vomsc$voms_flavor,VOMS_Init, VOMS_LIBS="$LDFLAGS -lvomsapi$voms_flavor",voms=false,)
  ,
  voms=false
)
AC_SUBST(VOMS_LIBS)
AC_LANG(C)

fi

if test "x$voms" != xtrue ; then
  ng_voms=
  edg_voms=
  AC_MSG_WARN(VOMS will be disabled)
fi

if test "x$ng_voms" != "x" ; then
  AC_DEFINE(NG_VOMS,,[voms-1.1.x]),
fi
if test "x$edg_voms" != "x" ; then
  AC_DEFINE(EDG_VOMS,,[voms-1.2.x])
fi
AM_CONDITIONAL(NG_VOMS, test "x$ng_voms" != "x")
AM_CONDITIONAL(EDG_VOMS, test "x$edg_voms" != "x")

#
# Qt checks
#
have_qt=
QT_INCLUDES=
QT_LIBS=
# If QTDIR is not specified environment is used
# If not available in this shell then try a login shell
if test "x$QTDIR" = "x"; then
  eval `sh -l -c printenv|grep ^QTDIR=`
fi
AC_ARG_WITH(qtdir,
  [  --with-qtdir=<PATH>         Specify the QT installation path],
  [QTDIR=$with_qtdir]
)

# Fallback for Debian
if test "x$QTDIR" = "x"; then
  QTDIR=/usr/share/qt
fi

if ! test -d "$QTDIR"; then
  AC_MSG_WARN([Qt directory not found - no gui])
else
  QT_INCLUDES=-I$QTDIR/include
  CPPFLAGS=$QT_INCLUDES
  AC_CHECK_HEADER([qglobal.h],
    [
      AC_PATH_PROG([MOC], [moc], [not found], $QTDIR/bin:$PATH)
      if test "x$MOC" = "xnot found"; then
        AC_MSG_WARN([No moc - no gui])
      else
        LDFLAGS="-L$QTDIR/lib -L/usr/X11R6/lib -lm"
#        AC_CHECK_LIB(qt-mt,theMainWindow,
          QT_LIBS="$LDFLAGS -lqt-mt"
          have_qt=true
#        )
      fi       
    ],
    AC_MSG_WARN([Qt common include file not found - no gui])
  )
fi
AC_SUBST(QT_INCLUDES)
AC_SUBST(QT_LIBS)
AM_CONDITIONAL(HAVE_QT, test "x$have_qt" != "x")

dnl
dnl Check how signal SIG_CHLD is treated in threads
dnl
nonposix_signals_in_threads=
LIBS=-lpthread
AC_MSG_CHECKING(for signals in threads)
AC_TRY_RUN(
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

pid_t pid = -1;
int res = -2;

void* parent_thread(void *arg) {
  pid=fork();
  if(pid == -1) {
    _exit(-1);
  };
  if(pid == 0) {
    sleep(5);
    exit(0);
  };
  sleep(1000);
}

void* waiter_thread(void *arg) {
  int status;
  pid_t p;
  sleep(2);
  p=waitpid(pid,&status,0);
  if(p == -1) {
    if(errno != ECHILD) { res=-1; } else { res=1; };
    _exit(res);
  };
  if(p == 0) {
    res=-1;
    _exit(-1);
  };
  res=0;
}

int main(void) {
  pthread_t parent_thr;
  pthread_t waiter_thr;
  int n;
  if(pthread_create(&parent_thr,NULL,&parent_thread,NULL) != 0) {
    _exit(-1);
  };
  if(pthread_create(&waiter_thr,NULL,&waiter_thread,NULL) != 0) {
    _exit(-1);
  };
  for(n=0;n<30;n++) {
    sleep(1);
    if(res!=-2) break;
  };
  _exit(res);
}
  ,
  AC_MSG_RESULT([quite POSIX]),
  [AC_DEFINE(NONPOSIX_SIGNALS_IN_THREADS,,[Old Linux pthreads])
  AC_MSG_RESULT([not a POSIX])],
  AC_MSG_ERROR([cross-compilation is not supported])
)
AM_CONDITIONAL(NONPOSIX_SIGNALS_IN_THREADS, test "x$nonposix_signals_in_threads" != "x")


dnl
dnl Experimental code
dnl
AC_ARG_ENABLE(experimental,
  [  --enable-experimental               Enable experimental code],  
  [
    case "${enableval}" in
      yes) experimental=true ;;
      no)  experimental=false ;;
      *)   AC_MSG_ERROR(bad value ${enableval} for --enable-experimental) ;;
    esac
  ],
  [ experimental=false ]
)
experimental_prefix="# "
if test "x$experimental" = "xtrue"; then
  if test "x$gsoap_location" = "x"; then
    AC_MSG_ERROR(Currently there is no experimental code to be built without gSOAP)
  fi
dnl  AC_MSG_NOTICE([Turning on experimental code])
  experimental_prefix=""
fi
AC_SUBST(experimental_prefix)
AM_CONDITIONAL(EXPERIMENTAL, test x$experimental = xtrue)
AM_CONDITIONAL(BUILD_LOGGER, test "x$mysql_config" != "x" && test "x$gsoap_location" != "x")
AM_CONDITIONAL(BUILD_HTTPSD, test "x$gsoap_location" != "x")
AM_CONDITIONAL(BUILD_SE, test "x$xml_config" != "x" && test "x$gsoap_location" != "x")
AM_CONDITIONAL(BUILD_GUI, test "x$have_qt" != "x" && test "x$experimental" = "xtrue")

gridssh_init_script=""
gridssh=true
AC_ARG_ENABLE(gridssh,
  [  --disable-gridssh	Disable gridssh client and server],  
  [
    case "${enableval}" in
      yes) gridssh=true ;;
      no)  gridssh=false; AC_MSG_WARN(gridssh will be disabled) ;;
      *)   AC_MSG_ERROR(bad value ${enableval} for --disable-gridssh) ;;
    esac
  ],
  [ gridssh=true ]
)
gridssh_prefix="# "
if test "x$gridssh" = "xtrue"; then
dnl  AC_MSG_NOTICE([enabling compilation of gridssh client and server])
gridssh_init_script="gridsshd"
gridssh_prefix=""
fi
AC_SUBST(gridssh_prefix)
AC_SUBST(gridssh_init_script)
AM_CONDITIONAL(BUILD_GRIDSSH, test x$gridssh = xtrue)

dnl Add $GLOBUS_LOCATION/lib to search for system dynamic libraries.
dnl Is used to avoid rpath of the specified paths
  
AC_ARG_ENABLE(rpath-hack,
  [  --enable-rpath-hack                 Do not rpath Globus libraries in binaries],
  [
    case "${enableval}" in
      yes) rpath_hack=true ;;
      no)  rpath_hack=false ;;
      *)   AC_MSG_ERROR(bad value ${enableval} for --enable-rpath-hack) ;;
    esac
  ],
  [ rpath_hack=true ]
)
if test "x$rpath_hack" = "xtrue"; then
dnl  AC_MSG_NOTICE([Adding GLOBUS_LOCATION/lib to system dynamic library search path])
  cp -p libtool libtool.orig
  cat libtool.orig | \
    sed "s,\(^sys_lib_dlsearch_path_spec=.*\).$,\1 $GLOBUS_LOCATION/lib\"," > \
    libtool
fi

]
)
