#ifndef __GM_AUTH_H__
#define __GM_AUTH_H__

#include <string>
#include <list>

#include <gssapi.h>

#ifdef EDG_VOMS
#endif

#ifdef NG_VOMS
  #define HAVE_VOMS
  #include <voms_api.h>
#else
 #ifdef EDG_VOMS
  #define HAVE_VOMS
  #include <voms_api.h>
 #else
  #undef HAVE_VOMS
 #endif
#endif


#define AAA_POSITIVE_MATCH 1
#define AAA_NEGATIVE_MATCH -1
#define AAA_NO_MATCH 0
#define AAA_FAILURE 2

class AuthUser {
 private:
  typedef int (AuthUser:: * match_func_t)(const char* line);
  typedef struct {
    const char* cmd;
    match_func_t func;
  } source_t;
  class group_t {
   public:
    const char* voms;             //
    const char* vo;               //
    const char* role;             //
    const char* capability;       //
    const char* vgroup;           //
    string name;                  //
    group_t(const char* name_,const char* vo_,const char* role_,const char* cap_,const char* vgrp_):name(name_?name_:""),vo(vo_?vo_:""),role(role_?role_:""),capability(cap_?cap_:""),vgroup(vgrp_?vgrp_:"") { };
  };
  string subject;
  string filename;
  bool proxy_file_was_created;
  static source_t sources[];
  int match_all(const char* line);
  int match_group(const char* line);
  int match_subject(const char* line);
  int match_file(const char* line);
  int match_ldap(const char* line);
  int match_voms(const char* line);
  int match_plugin(const char* line);
#ifdef HAVE_VOMS
  vector<struct voms> voms_data;
  bool voms_extracted;
#endif
  list<group_t> groups;
  const char* default_voms_;
  const char* default_vo_;
  const char* default_role_;
  const char* default_capability_;
  const char* default_vgroup_;
  const char* default_group_;
 public:
  AuthUser(const AuthUser&);
  AuthUser(const char* subject = NULL,const char* filename = NULL);
  ~AuthUser(void);
  void operator=(const AuthUser&);
  void operator=(gss_cred_id_t cred);
  void set(const char* subject,gss_cred_id_t cred);
  int evaluate(const char* line);
  const char* DN(void) const { return subject.c_str(); };
  const char* proxy(void) const { return filename.c_str(); };
  void add_group(const char* grp) {
    groups.push_back(group_t(grp,default_vo_,default_role_,default_capability_,default_vgroup_));
  };
  void add_group(const string& grp) { add_group(grp.c_str()); };
  void clear_groups(void) { groups.clear(); default_group_=NULL; };
  bool check_group(const char* grp) const {
    for(list<group_t>::const_iterator i=groups.begin();i!=groups.end();++i) {
      if(strcmp(i->name.c_str(),grp) == 0) return true;
    };
    return false;
  };
  bool check_group(const string& grp) const { return check_group(grp.c_str());};
  const char* default_voms(void) const { return default_voms_; };
  const char* default_vo(void) const { return default_vo_; };
  const char* default_role(void) const { return default_role_; };
  const char* default_capability(void) const { return default_capability_; };
  const char* default_vgroup(void) const { return default_vgroup_; };
  const char* default_group(void) const { return default_group_; };
  const char* default_subject(void) const { return subject.c_str(); };
#ifdef HAVE_VOMS
  const vector<struct voms>& voms(void);
#endif
};

class AuthEvaluator {
 private:
  list<string> l;
  string name;
 public:
  AuthEvaluator(void);
  AuthEvaluator(const char* name);
  ~AuthEvaluator(void);
  void add(const char*);
  int evaluate(AuthUser &) const;
  bool operator==(const char* s) { return (strcmp(name.c_str(),s)==0); };
  bool operator==(const string& s) { return (name == s); };
  const char* get_name() { return name.c_str(); };
};

#endif 
