#include "../std.h"

#include "../misc/url_options.h"
#include "../misc/stringtoint.h"
#include "../misc/globus_error_utils.h"
#include "../misc/globus_modules.h"
#include "../misc/log_time.h"
#include "datapoint.h"
#include "datahandle.h"

bool DataHandle::analyze(analyze_t &arg) {
  if(!url) return false;
  const char* cur_url = url->current_location();
  const string cur_url_s(cur_url);
  string value;
#ifdef SKIP_HTTPG
  if(!strncasecmp("gsiftp://",cur_url,9)) {
#else
  if((strncasecmp("gsiftp://",cur_url,9)==0) ||
     (strncasecmp("http://",cur_url,7)==0) ||
     (strncasecmp("https://",cur_url,8)==0) ||
     (strncasecmp("httpg://",cur_url,8)==0) ||
     (strncasecmp("se://",cur_url,5)==0) ) {
#endif
    unsigned int threads;
    if(get_url_option(cur_url_s,"threads",0,value) == 0) {
      if(stringtoint(value,threads)) {
        if(threads < 1) threads = 1;
        if(threads > MAX_PARALLEL_STREAMS) threads = MAX_PARALLEL_STREAMS;
      }
      else { threads = 1; };
      arg.bufnum=threads;
    };
  };
  arg.bufsize=-1;
  if((strncasecmp("http://",cur_url,7)==0) ||
     (strncasecmp("https://",cur_url,8)==0) ||
     (strncasecmp("httpg://",cur_url,8)==0) ||
     (strncasecmp("se://",cur_url,5)==0)) {
     // for HTTP(S/G) bufsize directly affects transaction's length
     // Hence it is better to have it bigger
     arg.bufsize=MAX_BLOCK_SIZE;
  };
  if(get_url_option(cur_url_s,"blocksize",0,value) == 0) {
    int n;
    if(stringtoint(value,n)) {
      if(n<0) n=0;
      if(n > MAX_BLOCK_SIZE) n = MAX_BLOCK_SIZE;
      arg.bufsize=n;
    };
  };
  arg.cache=true;
  if(get_url_option(cur_url_s,"cache",0,value) == 0) {
    if(strcasecmp(value.c_str(),"no") == 0) arg.cache=false;
  };
  arg.readonly=true;
  if(get_url_option(cur_url_s,"readonly",0,value) == 0) {
    if(strcasecmp(value.c_str(),"no") == 0) arg.readonly=false;
  };
  if(!strcmp("-",cur_url)) { arg.cache=false; arg.readonly=false; };
  arg.local=false;
  if(!strncasecmp("file:/",cur_url,6)) {
    arg.local=true; arg.cache=false;
  };
  return true;
}

bool DataHandle::deinit_handle(void) {
  if(ftp_active) {
    globus_ftp_client_handle_destroy(&ftp_handle);
    globus_ftp_client_operationattr_destroy(&ftp_opattr);
//    globus_ftp_client_plugin_destroy(&ftp_plugin);
  };
#ifdef SKIP_HTTPG
  if(http_active) {
    if(http_handle_was_initialized) {
      odlog(2)<<"DataHandle::deinit_handle: destroy http_handle"<<endl;
      globus_gass_transfer_request_destroy(http_handle);
    };
  };
#endif
  return true;
}

bool DataHandle::init_handle(void) {
  if(!url) return false;
  const char* cur_url = url->current_location();
  const string cur_url_s(cur_url);
  cacheable=true;
  linkable=true;
  string value;
  if(get_url_option(cur_url_s,"cache",0,value) == 0) {
    if(strcasecmp("no",value.c_str()) == 0) cacheable=false;
  };
  if(get_url_option(cur_url_s,"readonly",0,value) == 0) {
    if(strcasecmp("no",value.c_str()) == 0) linkable=false;
  };
  c_url=cur_url;
  out_of_order(out_of_order());
  if(strcmp("-",cur_url)) if(::canonic_url(c_url) != 0) return false;
  if((!strncasecmp("ftp://",cur_url,6)) || 
     (!strncasecmp("gsiftp://",cur_url,9))) {
    url_proto=url_is_ftp;
    is_secure=false;
    if(!strncasecmp("gsiftp://",cur_url,9)) is_secure=true;
    if(!ftp_active) {
      globus_result_t res;
      if((res=globus_ftp_client_handle_init(
                           &ftp_handle,GLOBUS_NULL)) != GLOBUS_SUCCESS) {
        odlog(0)<<"init_handle: globus_ftp_client_handle_init failed"<<endl;
        odlog(0)<<"Globus error: "<<GlobusResult(res)<<endl;
        ftp_active=false; return false;
      };
      if((res=globus_ftp_client_operationattr_init(
                           &ftp_opattr)) != GLOBUS_SUCCESS) {
        odlog(0)<<"init_handle: globus_ftp_client_operationattr_init failed"<<endl;
        odlog(0)<<"Globus error: "<<GlobusResult(res)<<endl;
        globus_ftp_client_handle_destroy(&ftp_handle);
        ftp_active=false; return false;
      };
    };
    ftp_active=true; ftp_threads=1;
    if(allow_out_of_order) {
      if(get_url_option(cur_url_s,"threads",0,value) == 0) {
        if(stringtoint(value,ftp_threads)) {
          if(ftp_threads < 1) ftp_threads = 1;
          if(ftp_threads > MAX_PARALLEL_STREAMS) ftp_threads = MAX_PARALLEL_STREAMS;
        }
        else { ftp_threads = 1; };
      };
    };
    globus_ftp_control_parallelism_t paral;
    if(ftp_threads > 1) {
      paral.fixed.mode=GLOBUS_FTP_CONTROL_PARALLELISM_FIXED;
      paral.fixed.size=ftp_threads;
    } 
    else {
      paral.fixed.mode=GLOBUS_FTP_CONTROL_PARALLELISM_NONE;
      paral.fixed.size=1;
    };
    globus_ftp_client_operationattr_set_parallelism(&ftp_opattr,&paral);
    globus_ftp_client_operationattr_set_striped(&ftp_opattr,GLOBUS_FALSE);
    /*   globus_ftp_client_operationattr_set_layout         */
    /*   globus_ftp_client_operationattr_set_tcp_buffer     */
    globus_ftp_client_operationattr_set_type(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_TYPE_IMAGE);
    if(!is_secure) { // plain ftp protocol
      globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_STREAM);
      globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_PROTECTION_CLEAR);
      globus_ftp_client_operationattr_set_control_protection(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_PROTECTION_CLEAR);
      // need to set dcau to none in order Globus libraries not to send 
      // it to pure ftp server
      globus_ftp_control_dcau_t dcau;
      dcau.mode=GLOBUS_FTP_CONTROL_DCAU_NONE;
      globus_ftp_client_operationattr_set_dcau(&ftp_opattr,&dcau);
    }
    else { // gridftp protocol

      char* subj = getenv("SUBJECT");
      if(subj) {
        globus_ftp_client_operationattr_set_authorization(&ftp_opattr,
          GSS_C_NO_CREDENTIAL,NULL,NULL,NULL,subj);
      };
      if(get_url_option(cur_url_s,"secure",0,value) == 0) {
        if(value == "yes") {
          globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
          globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                      GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
          odlog(2)<<"Using secure data transfer (requested in url)"<<endl;
        } else {
          if(force_passive) {
            globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_STREAM);
          } else {
            globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
          };
          globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                      GLOBUS_FTP_CONTROL_PROTECTION_CLEAR);
          odlog(2)<<"Using insecure data transfer (requested in url)"<<endl;
        };
      } else {
        if(force_secure) {
          globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
          globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                      GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
        } else {
          if(force_passive) {
            globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_STREAM);
          } else {
            globus_ftp_client_operationattr_set_mode(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_MODE_EXTENDED_BLOCK);
          };
          globus_ftp_client_operationattr_set_data_protection(&ftp_opattr,
                                      GLOBUS_FTP_CONTROL_PROTECTION_CLEAR);
          odlog(2)<<"Using insecure data transfer"<<endl;
        };
      };
      globus_ftp_client_operationattr_set_control_protection(&ftp_opattr,
                                    GLOBUS_FTP_CONTROL_PROTECTION_PRIVATE);
    };
    /*   globus_ftp_client_operationattr_set_dcau                         */
    /*   globus_ftp_client_operationattr_set_resume_third_party_transfer  */
    /*   globus_ftp_client_operationattr_set_authorization                */
    globus_ftp_client_operationattr_set_append(&ftp_opattr,GLOBUS_FALSE);
    return true;
  }
#ifdef SKIP_HTTPG
  else if((!strncasecmp("http://",cur_url,7)) || 
          (!strncasecmp("https://",cur_url,8))) {
    url_proto=url_is_http;
    if(!http_active) {
      http_handle_was_initialized=false;
      http_active=true;
    };
    return true;
  }
#else
  else if((!strncasecmp("http://",cur_url,7)) || 
          (!strncasecmp("https://",cur_url,8)) ||
          (!strncasecmp("httpg://",cur_url,8)) ||
          (!strncasecmp("se://",cur_url,5))) {
odlog(1)<<"init_handle: httpg, https, http or se"<<endl;
    url_proto=url_is_httpg;
    if(!strncasecmp("se://",cur_url,5)) {
      c_url.replace(0,2,"httpg");
      int n = c_url.find('?'); if(n != string::npos) c_url.replace(n,1,"/");
    };
    ftp_threads=1;
    if(allow_out_of_order) {
      if(get_url_option(cur_url_s,"threads",0,value) == 0) {
        if(stringtoint(value,ftp_threads)) {
          if(ftp_threads < 1) ftp_threads = 1;
          if(ftp_threads > MAX_PARALLEL_STREAMS) ftp_threads = MAX_PARALLEL_STREAMS;
        }
        else { ftp_threads = 1; };
      };
    };
odlog(1)<<"init_handle: http(s,g)/se: threads: "<<ftp_threads<<endl;
    return true;
  }
#endif
  else if(!strncasecmp("file:/",cur_url,6)) {
    url_proto=url_is_file;
    cacheable=false;
    return true;
  }
  else if(!strcmp("-",cur_url)) {
    url_proto=url_is_channel;
    cacheable=false;
    linkable=false;
    return true;
  } else {
    odlog(1)<<"Protocol in url "<<cur_url<<" is not supported"<<endl;
  };
  return false;
}

DataHandle::DataHandle(DataPoint* url_) {
  url=url_;
  /* connection is not initialised here - only if request is made */
  no_checks=false;
  reading=false;
  writing=false;
  ftp_active=false; http_active=false; 
  if(!ftp_mod.active()) {
    odlog(0)<<"GLOBUS_FTP_CLIENT_MODULE activation failed"<<endl;
    url=NULL;
  };
#ifdef SKIP_HTTPG
  if(!http_mod.active()) {
    odlog(0)<<"GLOBUS_GASS_TRANSFER_MODULE activation failed"<<endl;
    url=NULL;
  };
#endif
  is_secure=false; force_secure=true; force_passive=false;
  allow_out_of_order=false; // safe setting
#ifndef SKIP_HTTPG
  httpg_stat=NULL;
#endif
  failure_code=common_failure;
  range_start=0;
  range_end=0;
}

DataHandle::~DataHandle(void) {
  /*  Stop any transfer */
  stop_reading();
  stop_writing();
  /* !!!!!!!!!!! destroy activated handles, etc here !!!!!!!! */
  deinit_handle();
}

bool DataHandle::start_reading(DataBufferPar &buffer) {
  failure_code=common_failure;
  if(reading || writing || (url==NULL)) return false;
  if(!init_handle()) return false;
  reading=true;
  switch(url_proto) {
    case url_is_channel:
    case url_is_file: {
      if(!start_reading_file(buffer)) { reading=false; return false; };
    }; break;
    case url_is_ftp: {
      if(!start_reading_ftp(buffer)) { reading=false; return false; };
    }; break;
#ifdef SKIP_HTTPG
    case url_is_http: {
      if(!start_reading_http(buffer)) { reading=false; return false; };
    }; break;
#else
    case url_is_httpg: {
      if(!start_reading_httpg(buffer)) { reading=false; return false; };
    }; break;
#endif
    default: reading=false; return false;
  };  
  return true;
}

bool DataHandle::check(void) {
  failure_code=common_failure;
  if(reading || writing || (url==NULL)) return false;
  if(!init_handle()) return false;
  switch(url_proto) {
    case url_is_channel: {
      return false;
    }; break;
    case url_is_file: {
      if(!check_file()) { return false; };
    }; break;
    case url_is_ftp: {
      if(!check_ftp()) { return false; };
    }; break;
#ifdef SKIP_HTTPG
    case url_is_http: {
      if(!check_http()) { return false; };
    }; break;
#else
    case url_is_httpg: {
      if(!check_httpg()) { return false; };
    }; break;
#endif
    default: {
      odlog(2)<<"DataHandle::check: unknown protocol"<<endl;
      return false;
    };
  };
  return true;
}

bool DataHandle::remove(void) {
  failure_code=common_failure;
  if(reading || writing || (url==NULL)) return false;
  if(!init_handle()) return false;
  switch(url_proto) {
    case url_is_channel: {
      return false;
    }; break;
    case url_is_file: {
      if(!remove_file()) { return false; };
    }; break;
    case url_is_ftp: {
      if(!remove_ftp()) { return false; };
    }; break;
#ifdef SKIP_HTTPG
    case url_is_http: {
      return false;
    }; break;
#else
    case url_is_httpg: {
odlog(1)<<"remove: httpg"<<endl;
      if(!remove_httpg()) { return false; };
    }; break;
#endif
    default: return false;
  };
  return true;
}

bool DataHandle::start_writing(DataBufferPar &buffer,DataCallback *space_cb) {
  failure_code=common_failure;
  if(reading || writing || (url==NULL)) return false;
  if(!init_handle()) return false;
  writing=true;
  switch(url_proto) {
    case url_is_channel:
    case url_is_file: {
      if(!start_writing_file(buffer,space_cb)) { writing=false; return false; };
    }; break;
    case url_is_ftp: {
      if(!start_writing_ftp(buffer)) { writing=false; return false; };
    }; break;
#ifdef SKIP_HTTPG
    case url_is_http: {
      if(!start_writing_http(buffer)) { writing=false; return false; };
    }; break;
#else
    case url_is_httpg: {
odlog(1)<<"start_writing: httpg"<<endl;
      if(!start_writing_httpg(buffer)) { writing=false; return false; };
    }; break;
#endif
    default: writing=false; return false;
  };
  return true;
}

bool DataHandle::stop_reading(void) {
  if(!reading) return false;
  reading=false;   
  switch(url_proto) {
    case url_is_channel:
    case url_is_file: {
      return stop_reading_file();
    }; break;
    case url_is_ftp: {
      return stop_reading_ftp();
    }; break;
#ifdef SKIP_HTTPG
    case url_is_http: {
      return stop_reading_http();
    }; break;
#else
    case url_is_httpg: {
      return stop_reading_httpg();
    }; break;
#endif
  };
  return true;
}

bool DataHandle::stop_writing(void) {
  if(!writing) return false;
  writing=false;
  switch(url_proto) {
    case url_is_channel:
    case url_is_file: {
      return stop_writing_file();
    }; break;
    case url_is_ftp: {
      return stop_writing_ftp();
    }; break;
#ifdef SKIP_HTTPG
    case url_is_http: {
      return stop_writing_http();
    }; break;
#else
    case url_is_httpg: {
      return stop_writing_httpg();
    }; break;
#endif
  };
  return true;
}

bool DataHandle::out_of_order(void) {
  const char* cur_url = url->current_location();
#ifdef SKIP_HTTPG
  if((strncasecmp(cur_url,"gsiftp://",9)==0) || 
     (strncasecmp(cur_url,"file:/",6)==0)) return true;
#else
  if((strncasecmp(cur_url,"gsiftp://",9)==0) || 
     (strncasecmp(cur_url,"file:/",6)==0) || 
     (strncasecmp(cur_url,"http://",7)==0) || 
     (strncasecmp(cur_url,"https://",8)==0) || 
     (strncasecmp(cur_url,"httpg://",8)==0) ||
     (strncasecmp(cur_url,"se://",5)==0)) return true;
#endif
  return false;
}

void DataHandle::out_of_order(bool val) {
  allow_out_of_order=val;
}

bool DataHandle::secure(void) {
  return is_secure;
}

void DataHandle::secure(bool val) {
  force_secure=val;
}

void DataHandle::passive(bool val) {
  force_passive=val;
}

bool DataHandle::list_files(list<DataPoint::FileInfo> &files,bool resolve) {
  if(!init_handle()) return false;
  if(url_proto == url_is_channel) {
    return false;
  } else if(url_proto == url_is_file) {
    return list_files_file(files,resolve);
  } else if(url_proto == url_is_ftp) {
    return list_files_ftp(files,resolve);
#ifndef SKIP_HTTPG
  } else if(url_proto == url_is_httpg) { // not every "httpg" proto supports listing
    return list_files_httpg(files,resolve);
#endif
  } else {
    odlog(0)<<"File listing for this protocol is not supported"<<endl;
    return false;
  };
  return true;
}

