#include "../std.h"
#include <iostream>

#include "../misc/url_options.h"
#include "../misc/stringtoint.h"
#include "../misc/globus_error_utils.h"
#include "../misc/log_time.h"
#include "datapoint.h"
#include "databufferpar.h"
#include "datahandle.h"
#include "datacallback.h"


void DataHandle::read_http(void* arg,globus_gass_transfer_request_t handle,
    globus_byte_t* buffer, globus_size_t length, globus_bool_t last_data) {
#ifdef SKIP_HTTPG
/* maybe should put locks here in case callbacks overlap ???? */
  odlog(2)<<"DataHandle: read_http: buffer: "<<(unsigned long int)buffer<<endl;
  odlog(2)<<"DataHandle: read_http: length: "<<length<<endl;
  odlog(2)<<"DataHandle: read_http: last  : "<<last_data<<endl;
  DataHandle* it = (DataHandle*)arg;
  if(buffer != NULL) {
    odlog(2)<<"DataHandle: read_http: releasing buffer"<<endl;
    it->buffer->is_read((char*)buffer,length,it->http_offset);
    it->http_offset+=length;
  };
  if(last_data) {  /* success */
    odlog(2)<<"DataHandle: read_http: finish: set eof"<<endl;
    it->buffer->eof_read(true);
    odlog(2)<<"DataHandle: read_http: finish: leave"<<endl;
    http_callback(arg,handle);
    return;
  };
  int h;
  unsigned int l;
  odlog(2)<<"DataHandle: read_http: claiming buffer"<<endl;
  globus_thread_blocking_will_block();
  if(!it->buffer->for_read(h,l,true)) {
    odlog(2)<<"DataHandle: read_http: claiming buffer: failed"<<endl;
    /* failed to get buffer - must be error or request to exit */
    it->buffer->error_read(true);
    globus_gass_transfer_fail(it->http_handle,&http_callback,arg);
    return;
  };
  if(it->buffer->error()) {
    odlog(2)<<"DataHandle: read_http: buffer error"<<endl;
    it->buffer->is_read(h,0,0);
    globus_gass_transfer_fail(it->http_handle,&http_callback,arg);
    return;
  };
  odlog(2)<<"DataHandle: read_http: register buffer/callback"<<endl;
  int res = globus_gass_transfer_receive_bytes(it->http_handle,
             (globus_byte_t*)((*(it->buffer))[h]),l,l,&read_http,arg);
  if(res != 0) {
    odlog(2)<<"Failed to register buffer with globus_gass_transfer_receive_bytes"<<endl;
    odlog(1)<<"Globus error: "<<globus_gass_error_string(res)<<endl;
    it->buffer->is_read(h,0,0);
    it->buffer->error_read(true);
    globus_gass_transfer_fail(it->http_handle,&http_callback,arg);
    return;
  };
#endif
}

void DataHandle::write_http(void* arg,globus_gass_transfer_request_t handle,
    globus_byte_t* buffer, globus_size_t length, globus_bool_t last_data) {
#ifdef SKIP_HTTPG
/* maybe should put locks here in case callbacks overlap ???? */
  odlog(2)<<"DataHandle: write_http: buffer: "<<(unsigned long int)buffer<<endl;
  odlog(2)<<"DataHandle: write_http: length: "<<length<<endl;
  odlog(2)<<"DataHandle: write_http: last  : "<<last_data<<endl;
  DataHandle* it = (DataHandle*)arg;
  if(buffer != NULL) {
    odlog(2)<<"DataHandle: write_http: releasing buffer"<<endl;
    it->buffer->is_written((char*)buffer);
    it->http_offset+=length;
  };
  if(last_data) {  
    odlog(2)<<"DataHandle: write_http: finish: set eof"<<endl;
    it->buffer->eof_write(true);
    odlog(2)<<"DataHandle: write_http: finish: leave"<<endl;
    http_callback(arg,handle);
    return;
  };
  int h;
  unsigned int l;
  unsigned long long int o;
  odlog(2)<<"DataHandle: write_http: claiming buffer"<<endl;
  globus_bool_t have_eof=GLOBUS_FALSE;
  globus_thread_blocking_will_block();
  if(!it->buffer->for_write(h,l,o,true)) {
    odlog(2)<<"DataHandle: write_http: claiming buffer: failed"<<endl;
    /* failed to get buffer - must be error or eof */
    if(it->buffer->error()) {
      it->buffer->error_write(true);
      globus_gass_transfer_fail(it->http_handle,&http_callback,arg);
      return;
    };
    have_eof=GLOBUS_TRUE;
  };
  if(it->buffer->eof_read()) have_eof=GLOBUS_TRUE;
  odlog(2)<<"DataHandle: write_http: register buffer/callback"<<endl;
  int res = globus_gass_transfer_send_bytes(it->http_handle,
             (globus_byte_t*)((*(it->buffer))[h]),l,have_eof,&write_http,arg);
  if(res != 0) {
    odlog(2)<<"Failed to register buffer with globus_gass_transfer_send_bytes"<<endl;
    odlog(1)<<"Globus error: "<<globus_gass_error_string(res)<<endl;
    it->buffer->is_written(h);
    it->buffer->error_write(true);
    globus_gass_transfer_fail(it->http_handle,&http_callback,arg);
    return;
  };
#endif
}

bool DataHandle::check_http(void) {
  odlog(2)<<"check_http: start"<<endl;
#ifdef SKIP_HTTPG
  ftp_completed.reset();
/*

TODO - obtain creation/expiration information

*/
  odlog(2)<<"check_http: register get"<<endl;
  int res = globus_gass_transfer_register_get(
    &http_handle,GLOBUS_NULL,(char*)(c_url.c_str()),&http_check_callback,this);
  odlog(2)<<"check_http: register get finished: "<<res<<endl;
  if(res != 0) {
    odlog(1)<<"Globus error: "<<globus_gass_error_string(res)<<endl;
    return false;
  };
  int c_res;
  bool c_ex = false;

  if(!ftp_completed.wait(c_res,300000)) {
    odlog(0)<<"check_http: timeout waiting for connection"<<endl;
//    globus_gass_transfer_fail(http_handle,&http_check_callback,this);
//    ftp_completed.wait(c_res,-1);
  }
  else {
    c_ex=(c_res == 0);
    odlog(2)<<"check_http: finished with error code: "<<c_res<<endl;
  };
  globus_gass_transfer_fail(http_handle,&http_check_callback,this);
  for(;;) {
    globus_gass_transfer_request_status_t s =
               globus_gass_transfer_request_get_status(http_handle);
    odlog(2)<<"check_http: state: "<<s<<endl;
    if(s == GLOBUS_GASS_TRANSFER_REQUEST_INVALID) break;
    if(ftp_completed.wait(c_res,1000)) break;
  };
  if(globus_gass_transfer_request_destroy(http_handle) ==
                                  GLOBUS_SUCCESS) for(;;) {
    globus_gass_transfer_request_status_t s =
               globus_gass_transfer_request_get_status(http_handle);
    if(s == GLOBUS_GASS_TRANSFER_REQUEST_INVALID) break;
    globus_thread_yield();
  };
  return c_ex;
#else
  return false;
#endif
}

void DataHandle::http_check_callback(void* arg,globus_gass_transfer_request_t handle) {
#ifdef SKIP_HTTPG
  DataHandle* it = (DataHandle*)arg;
  switch(globus_gass_transfer_request_get_status(it->http_handle)) {
    case GLOBUS_GASS_TRANSFER_REQUEST_DENIED:
    case GLOBUS_GASS_TRANSFER_REQUEST_FAILED:
    case GLOBUS_GASS_TRANSFER_REQUEST_INVALID:
    case GLOBUS_GASS_TRANSFER_REQUEST_REFERRED: /* not supported (yet?) */
    default: {
      /* Failed to initiate transfer in some way */
      it->ftp_completed.signal(1);
    }; break;
    case GLOBUS_GASS_TRANSFER_REQUEST_PENDING: 
      /* Can it somehow jump over starting here ?! - just in case */
    case GLOBUS_GASS_TRANSFER_REQUEST_DONE:
      /* and how about done ?! */
    case GLOBUS_GASS_TRANSFER_REQUEST_STARTING: {
      it->ftp_completed.signal(0);
    }; break;
  };
#endif
}

void DataHandle::http_callback(void* arg,globus_gass_transfer_request_t handle) {
#ifdef SKIP_HTTPG
  DataHandle* it = (DataHandle*)arg;
  switch(globus_gass_transfer_request_get_status(it->http_handle)) {
    case GLOBUS_GASS_TRANSFER_REQUEST_DENIED:
    case GLOBUS_GASS_TRANSFER_REQUEST_FAILED:
    case GLOBUS_GASS_TRANSFER_REQUEST_INVALID:
    case GLOBUS_GASS_TRANSFER_REQUEST_REFERRED: /* not supported (yet?) */
    default: {
      odlog(1)<<"request: some error (DENIED,FAILED,INVALID,REFERRED): "<<it->c_url<<endl;
      switch(globus_gass_transfer_request_get_type(handle)) {
        case GLOBUS_GASS_TRANSFER_REQUEST_TYPE_GET: {
          odlog(2)<<"request: some error: get"<<endl;
          it->buffer->error_read(true);
          it->buffer->eof_read(true);
          it->file_thread_exited.signal();
        }; break;
        case GLOBUS_GASS_TRANSFER_REQUEST_TYPE_PUT: {
          odlog(2)<<"request: some error: put"<<endl;
          it->buffer->error_write(true);
          it->buffer->eof_write(true);
          it->file_thread_exited.signal();
        }; break;
        default: {
          odlog(2)<<"request: done: ????"<<endl;
        };
      };
    }; break;
    case GLOBUS_GASS_TRANSFER_REQUEST_STARTING: {
      odlog(2)<<"request: starting"<<endl;
    }; break;
    case GLOBUS_GASS_TRANSFER_REQUEST_PENDING: {
      odlog(2)<<"request: pending"<<endl;
      /* authenticated - now go for reading or writing */
      switch(globus_gass_transfer_request_get_type(handle)) {
        case GLOBUS_GASS_TRANSFER_REQUEST_TYPE_GET: {
          odlog(2)<<"request: pending: get"<<endl;
          read_http(arg,it->http_handle,NULL,0,GLOBUS_FALSE);
        }; break;
        case GLOBUS_GASS_TRANSFER_REQUEST_TYPE_PUT: {
          odlog(2)<<"request: pending: put"<<endl;
          write_http(arg,it->http_handle,NULL,0,GLOBUS_FALSE);
        }; break;
        default: {
          odlog(2)<<"request: pending: ????"<<endl;
        };
      };
    }; break;
    case GLOBUS_GASS_TRANSFER_REQUEST_DONE: {
      /*  should never get here - anyway, leave quietly */
      odlog(2)<<"request: done"<<endl;
      switch(globus_gass_transfer_request_get_type(handle)) {
        case GLOBUS_GASS_TRANSFER_REQUEST_TYPE_GET: { 
          it->buffer->eof_read(true);
          it->file_thread_exited.signal();
        }; break;
        case GLOBUS_GASS_TRANSFER_REQUEST_TYPE_PUT: {
          it->buffer->eof_write(true);
          it->file_thread_exited.signal();
        }; break;
        default: {
          odlog(2)<<"request: done: ????"<<endl;
        };
      };
    }; break;
  };
#endif
}

bool DataHandle::start_reading_http(DataBufferPar &buf) {
  odlog(2)<<"start_reading_http: start"<<endl;
#ifdef SKIP_HTTPG
  buffer=&buf;
  file_thread_exited.reset();
  http_offset=0;
  if(http_handle_was_initialized) {
    odlog(2)<<"DataHandle: start_reading_http: destroy handle"<<endl;
    globus_gass_transfer_request_destroy(http_handle);
    http_handle_was_initialized=false;
  };
  odlog(2)<<"start_reading_http: register get: "<<c_url<<endl;
  int res = globus_gass_transfer_register_get(
    &http_handle,GLOBUS_NULL,(char*)(c_url.c_str()),&http_callback,this);
  odlog(2)<<"start_reading_http: register get finished: "<<res<<endl;
  if(res != 0) {
    odlog(1)<<"Globus error: "<<globus_gass_error_string(res)<<endl;
    buffer->error_read(true);
    buffer->eof_read(true);
    return false;
  };
  http_handle_was_initialized=true;
  return true;
#else
  return false;
#endif
}

bool DataHandle::stop_reading_http(void) {
#ifdef SKIP_HTTPG
  if(!buffer->eof_read()) {
    buffer->error_read(true);
    if(globus_gass_transfer_fail(http_handle,&http_callback,this) != 0) {
      buffer->eof_read(true);
    };
  };
  file_thread_exited.wait(); 
  url->meta_validtill(time(NULL)+(60*10)); // give it 10 minutes
  /* bloody globus */
  if(globus_gass_transfer_request_destroy(http_handle) == 
                                  GLOBUS_SUCCESS) for(;;) {
    globus_gass_transfer_request_status_t s = 
	       globus_gass_transfer_request_get_status(http_handle);
    if(s == GLOBUS_GASS_TRANSFER_REQUEST_INVALID) break;
    globus_thread_yield();
  };
  http_handle_was_initialized=false;
#endif
  return true;
}

bool DataHandle::start_writing_http(DataBufferPar &buf) {
  odlog(2)<<"start_writing_http: start"<<endl;
#ifdef SKIP_HTTPG
  buffer=&buf;
  file_thread_exited.reset();
  http_offset=0;
  if(http_handle_was_initialized) {
    odlog(2)<<"DataHandle: start_writing_http: destroy handle"<<endl;
    globus_gass_transfer_request_destroy(http_handle);
    http_handle_was_initialized=false;
  };
  odlog(2)<<"start_writing_http: register get"<<endl;
  int res = globus_gass_transfer_register_put(
    &http_handle,GLOBUS_NULL,(char*)(c_url.c_str()),
    GLOBUS_GASS_TRANSFER_LENGTH_UNKNOWN,&http_callback,this);
  odlog(2)<<"start_writing_http: register put finished: "<<res<<endl;
  if(res != 0) {
    odlog(1)<<"Globus error: "<<globus_gass_error_string(res)<<endl;
    buffer->error_write(true);
    buffer->eof_write(true);
    return false;
  };
  http_handle_was_initialized=true;
  return true;
#else
  return false;
#endif
}

bool DataHandle::stop_writing_http(void) {
#ifdef SKIP_HTTPG
  if(!buffer->eof_write()) {
    buffer->error_write(true);
    if(globus_gass_transfer_fail(http_handle,&http_callback,this) != 0) {
      buffer->eof_write(true);
    };
  };
  file_thread_exited.wait();
  /* bloody globus */
  if(globus_gass_transfer_request_destroy(http_handle) ==
                                  GLOBUS_SUCCESS) for(;;) {
    globus_gass_transfer_request_status_t s =
               globus_gass_transfer_request_get_status(http_handle);
    if(s == GLOBUS_GASS_TRANSFER_REQUEST_INVALID) break;
    globus_thread_yield();
  };
  http_handle_was_initialized=false;
#endif
  return true;
}

