#include "../std.h"
#include <iostream>

#include "../misc/condition.h"
#include "../misc/log_time.h"
#include "../misc/checksum.h"
#include "../misc/inttostring.h"
#include "../https/se/file_soapH.h"
extern SOAP_NMAC struct Namespace file_soap_namespaces[];
#include "../https/client/client.h"
#include "../misc/time_utils.h"
#include "../../userinterface/CertInfo.h"
#include "databufferpar.h"
#include "datapoint.h"
#include "datahandle.h"


class httpg_info_t;
static int read_callback(unsigned long long offset,unsigned long long size,unsigned char** buf,unsigned long long* bufsize,void* arg);
static void* read_thread(void *arg);
static void* write_thread(void *arg);

typedef struct {
  httpg_info_t* stat;
  unsigned long long offset;
//  unsigned long long end;
  char* buffer;
  int buffer_h;
  unsigned int buffer_size;
  unsigned int buffer_used;
  pthread_t thr;
  HTTP_Client* s;
} httpg_state_t;

class httpg_info_t {
 friend int read_callback(unsigned long long offset,unsigned long long size,unsigned char** buf,unsigned long long* bufsize,void* arg);
 friend void* read_thread(void *arg);
 friend void* write_thread(void *arg);
 friend class DataHandle;
 private:
  int streams;
  unsigned int chunk_size;
  unsigned long long int requested_size;
  int threads;
  bool have_threads;
  CondSimple lock;
  DataBufferPar* buffer;
  const char* url;
  httpg_state_t* channels;
  bool cancel;
  unsigned long long int content_size;
  bool have_content_size;
  DataHandle* handle;
  DataPoint* point;
  DataHandle::failure_reason_t failure_code;
 public:
  httpg_info_t(void);
  ~httpg_info_t(void);
};

httpg_info_t::httpg_info_t(void) {
  url=NULL;
  buffer=NULL;
  threads=0; have_threads=false;
  channels=NULL;
  cancel=false;
  have_content_size=false;
  content_size=0;
}

httpg_info_t::~httpg_info_t(void) {
  if(channels) { free(channels); channels=NULL; };
}

#define CHECK_PROXY(msg,code) {               \
  CertInfo ci;                                \
  if(ci.TimeLeft() <= 0) {                    \
    odlog(1)<<msg<<": proxy expired"<<endl;   \
    code=                                     \
     DataHandle::credentials_expired_failure; \
  };                                          \
}

static int read_callback(unsigned long long offset,unsigned long long size,unsigned char** buf,unsigned long long* bufsize,void* arg) {
  httpg_state_t* tstat = (httpg_state_t*)arg;
  if(tstat->buffer == NULL) return -1;
  if(tstat->buffer_used == 0) {
    tstat->offset=offset;
  };
  // move data from buf to buffer
  for(;;) {
    if(size==0) break;
    unsigned long long l = (tstat->buffer_size - tstat->buffer_used);
    if(l>size) l=size; 
    memcpy(tstat->buffer+tstat->buffer_used,*buf,l);
    tstat->buffer_used+=size;
    size-=l; offset+=l; 
    if(tstat->buffer_used == tstat->buffer_size) { // full buffer
      if(!tstat->stat->buffer->is_read(tstat->buffer_h,
			            tstat->buffer_used,tstat->offset)) {
        tstat->buffer=NULL; tstat->buffer_h=-1;
        tstat->stat->buffer->error_read(true);
        return -1;

      };
      tstat->offset+=tstat->buffer_used;
      // new buffer needed
      if(!tstat->stat->buffer->for_read(tstat->buffer_h,
			                   tstat->buffer_size,true)) {
        tstat->buffer=NULL; tstat->buffer_h=-1;
        return -1;
      };
      tstat->buffer=(*(tstat->stat->buffer))[tstat->buffer_h];
      tstat->buffer_used=0;
    };
  };
  return 0;
}

static void* read_thread(void *arg) {
  httpg_info_t* istat = (httpg_info_t*)arg;
  istat->lock.block();
  if(istat->cancel) { istat->lock.unblock(); return NULL; };
  httpg_state_t* tstat = istat->channels+istat->threads;
  istat->threads++; istat->have_threads=true;
  bool encryption = true;
  if(strncasecmp(istat->url,"https://",8) == 0) encryption=false;
  HTTP_Client s(istat->url,encryption); // use relaxed encryption (integrity only)
  tstat->s=&s;
  istat->lock.unblock();
  tstat->stat=istat;
  bool failed = true;
  if(s) for(;;) {
    // get first buffer
    if(!istat->buffer->for_read(tstat->buffer_h,tstat->buffer_size,true)) break;
    tstat->buffer=(*(istat->buffer))[tstat->buffer_h];
    tstat->buffer_used=0;
    // connect
    odlog(2)<<"read_thread: calling connect"<<endl;
    if(s.connect() != 0) break;
    // acquire chunk
    istat->lock.block();
    tstat->offset = istat->requested_size;
    // check for eof and cancel
    if((istat->cancel) || 
       (istat->have_content_size && (tstat->offset >= istat->content_size))) {
      odlog(2)<<"read_thread: exiting due to eof or cancel: "<<tstat->offset<<" - "<<istat->content_size<<endl;
      istat->lock.unblock();
      istat->buffer->is_read(tstat->buffer_h,0,tstat->offset);
      if(!istat->cancel) failed=false;
      break;
    };
    istat->requested_size+=istat->chunk_size;
    istat->lock.unblock();
    unsigned long long int offset = tstat->offset;
    // recieve chunk
    odlog(2)<<"read_thread: calling GET: offset: "<<tstat->offset<<endl;
    odlog(2)<<"read_thread: calling GET: size: "<<istat->chunk_size<<endl;
    int get_res = s.GET("",tstat->offset,istat->chunk_size,&read_callback,tstat);
    // accept last received data
    if(tstat->buffer) {
      odlog(2)<<"read_thread: calling is_read: "<<tstat->buffer_used<<endl;
      istat->buffer->is_read(tstat->buffer_h,tstat->buffer_used,tstat->offset);
    };
    tstat->offset+=tstat->buffer_used;

    if(get_res != 0) { // failed
    odlog(2)<<"read_thread: GET failed"<<endl;
      istat->buffer->error_read(true);
      break;
    } else {
      unsigned long long int cs = s.response().ContentSize();
      if(cs) {
        istat->point->meta_size(cs);
        istat->content_size=cs; istat->have_content_size=true;
        istat->buffer->speed.set_max_data(cs); // little hack
      };
      if(s.response().haveLastModified()) {
        istat->point->meta_created(s.response().LastModified());
      };
      // if 0 bytes transfered - behind eof
      odlog(2)<<"read_thread: check for eof: "<<offset<<" - "<<tstat->offset<<endl;
      if(offset == tstat->offset) { failed=false; break; }; // eof
    };
  };
  odlog(2)<<"read_thread: loop exited"<<endl;
  istat->lock.block();
  istat->threads--;
  if(istat->threads==0) {
  odlog(2)<<"read_thread: last thread: failure: "<<failed<<endl;
    if(failed) {
      istat->buffer->error_read(true);
      CHECK_PROXY("read_thread",istat->failure_code);
    };
    istat->buffer->eof_read(true);
  };
  tstat->s=NULL;
  istat->lock.signal_nonblock();
  istat->lock.unblock();
  return NULL;  
}

bool DataHandle::start_reading_httpg(DataBufferPar &buf) {
  string store_url = c_url;
  // Try to obtain faster URL if can choose
  if(strncasecmp(url->current_location(),"se://",5) == 0) {
    odlog(1)<<"Talking to SOAP service at "<<c_url<<endl;
    struct soap soap;
    HTTP_ClientSOAP s(c_url.c_str(),&soap);
    soap.namespaces=file_soap_namespaces;
    if(s.connect() != 0) {
      odlog(0)<<"Failed to connect to "<<c_url<<endl;
      CHECK_PROXY("start_reading_httpg",failure_code);
      return false;
    };
    ns__infoResponse rr;
    string soap_url = c_url;
    int n = soap_url.find(':');
    // make gSOAP create http headers (should be removed in a future)
    if(n != string::npos) soap_url.replace(0,n,"http");
    if(soap_call_ns__info(&soap,soap_url.c_str(),"info",NULL,rr) != SOAP_OK) {
      odlog(1)<<"Failed to execute remote soap call 'info' at "<<c_url<<endl;
      return false;
    };
    if(rr.error_code != 0) {
      odlog(1)<<"Failed ("<<rr.error_code<<") to find remote file "<<c_url<<endl;
      return false;
    };
    if(rr.__size_file <= 0) { // it must be at least one entry
      odlog(1)<<"Failed to find remote file "<<c_url<<endl;
      return false;
    };
    // and it must be only one
    if((rr.file[0].__size_url > 0) && (rr.file[0].url[0])) {
      store_url=rr.file[0].url[0];
      // hack
      for(n=0;n<rr.file[0].__size_url;n++) {
        odlog(1)<<"File is available at: "<<rr.file[0].url[n]<<endl;
        if(strncasecmp(rr.file[0].url[n],"https://",8) == 0) // prefer SSL
          store_url=rr.file[0].url[n];
      };
    };
    if(rr.file[0].size) {
      /* provide some metadata */
      odlog(1)<<"start_reading_httpg: obtained size: "<<*(rr.file[0].size)<<endl;
      url->meta_size(*(rr.file[0].size));
    };
    if(rr.file[0].created) {
      time_t t;
      if(stringtotime(t,rr.file[0].created) == 0) url->meta_created(t);
    };
  };
  if(httpg_stat == NULL) httpg_stat=new httpg_info_t;
  httpg_stat->buffer=&buf; buffer=&buf;
  httpg_stat->url=store_url.c_str();
  httpg_stat->chunk_size=buf.buffer_size();
  httpg_stat->streams=ftp_threads;
  httpg_stat->requested_size=0; // amount of data currently being downloaded
  httpg_stat->threads=0;
  httpg_stat->cancel=false;
  httpg_stat->have_threads=false;
  if(httpg_stat->channels) free(httpg_stat->channels);
  httpg_stat->channels=(httpg_state_t*)malloc(sizeof(httpg_state_t)*httpg_stat->streams);
  if(httpg_stat->channels==NULL) return false;
  httpg_stat->handle = this;
  httpg_stat->point = url;
  httpg_stat->failure_code=failure_code;
  // start threads: one per stream
  httpg_stat->lock.reset();
  int started = 0;
  pthread_attr_t attr;
  if(pthread_attr_init(&attr) != 0) return false;
  if(pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED) != 0) return false;
  for(;started < httpg_stat->streams;started++) { 
    httpg_state_t* tstat = httpg_stat->channels+started;
    if(pthread_create(&(tstat->thr),&attr,&read_thread,httpg_stat) != 0) break;
  };
  odlog(2)<<"start_reading_httpg: streams started: "<<started<<endl;
  pthread_attr_destroy(&attr);
  if(started==0) return false;
  return true;
}

bool DataHandle::stop_reading_httpg(void) {
  httpg_stat->lock.block();
  failure_code=httpg_stat->failure_code;
  if(!buffer->eof_read()) {
    buffer->error_read(true);
    // cancel transfer
    httpg_stat->cancel=true;
    for(int i = 0;i<httpg_stat->streams;i++) {
      if(httpg_stat->channels[i].s) httpg_stat->channels[i].s->disconnect();
    };
  };
  while(httpg_stat->threads > 0) {
    httpg_stat->lock.wait_nonblock();
  };
  httpg_stat->lock.unblock();
  delete httpg_stat;
  return true;
}

static void* write_thread(void *arg) {
  httpg_info_t* istat = (httpg_info_t*)arg;
  istat->lock.block();
  if(istat->cancel) { istat->lock.unblock(); return NULL; };
  httpg_state_t* tstat = istat->channels+istat->threads;
  istat->threads++; istat->have_threads=true;
  odlog(2)<<"write_thread: threads: "<<istat->threads<<endl;
  bool encryption = true;
  if(strncasecmp(istat->url,"https://",8) == 0) encryption=false;
  HTTP_Client s(istat->url,encryption); // use relaxed encryption (integrity only)
  tstat->s=&s;
  istat->lock.unblock();
  tstat->stat=istat;
  bool failed = true;

  if(s) for(;;) {
    /// get first buffer
    if(!istat->buffer->for_write(tstat->buffer_h,tstat->buffer_size,tstat->offset,true)) {
      if(!istat->buffer->error()) failed=false; // eof
      break;
    };
    tstat->buffer=(*(istat->buffer))[tstat->buffer_h];
    // connect
    if(s.connect() != 0) break;
    // send chunk
    int get_res = s.PUT("",tstat->offset,tstat->buffer_size,
                        (unsigned char*)(tstat->buffer),
                        istat->have_content_size?istat->content_size:0,false);
    if(get_res != 0) { // failed
      istat->buffer->is_notwritten(tstat->buffer_h);
      istat->buffer->error_write(true);
      break;
    } else {
      istat->buffer->is_written(tstat->buffer_h);
    };
  };
  istat->lock.block();
  istat->threads--;
  if(istat->threads==0) {
    if(failed) odlog(2)<<"write_thread: last thread: failured"<<endl;
    DataHandle* handle = istat->handle;
    DataPoint* point = istat->point;
    if(point && !failed) {
      if(!istat->buffer->checksum_valid()) { failed=true; }
      else if(strncasecmp(point->current_location(),"se://",5) == 0) {
        // pass meta information to an SE
        struct soap soap;
        HTTP_ClientSOAP s(istat->url,&soap);
        soap.namespaces=file_soap_namespaces;
        if(s.connect() != 0) {
          odlog(0)<<"Failed to connect to "<<istat->url<<endl;
          failed=true;
        } else {
          ns__fileinfo info;
          info.size=NULL;
          // only cheksum is important here
          info.checksum=NULL;
          char checksum_[100];
          info.checksum=NULL;
          if(istat->buffer->checksum_valid()) { 
            const CheckSum* cs = istat->buffer->checksum_object();
            if(cs) {
              cs->print(checksum_,100);
              // int l_ = snprintf(checksum_,99,"cksum:%16LX",cs->end()); 
              // checksum_[l_]=0;
              info.checksum=checksum_;
            };
          };
          info.acl=NULL;
          info.id=NULL; 
          info.created=NULL;
          string created("");
          if(point->meta_created_available()) {
            struct tm* t;
            struct tm tt;
            time_t t_ = point->meta_created();
            t=gmtime_r(&t_,&tt);
            if(t) {
              if(timetostring(*t,created) == 0) {
                info.created=(char*)(created.c_str());
              };
            };
          };
          odlog(2)<<"write_thread: update: created: "<<info.created<<endl;
          odlog(2)<<"write_thread: update: created: "<<created<<endl;
          ns__updateResponse rr;
          int soap_err = SOAP_OK;
          string soap_url = istat->url;
          int n = soap_url.find(':');
          // make gSOAP create http headers (should be removed in a future)
          if(n != string::npos) soap_url.replace(0,n,"http");
          if((soap_err=soap_call_ns__update(&soap,soap_url.c_str(),"update",
                                    &info,rr)) != SOAP_OK) {
            odlog(1)<<"Failed to execute remote soap call 'update' at "<<istat->url<<endl;
            failed=true;
          } else if(rr.error_code != 0) {
            odlog(1)<<"Failed ("<<rr.error_code<<") to update remote file "<<info.id<<" at "<<istat->url<<endl;
            failed=true;
          };
        };
      };
    };
    if(failed) {
      istat->buffer->error_write(true);
      CHECK_PROXY("write_thread",istat->failure_code);
    };
    istat->buffer->eof_write(true);
  };
  tstat->s=NULL;
  istat->lock.signal_nonblock();
  istat->lock.unblock();
  return NULL;  
}

bool DataHandle::start_writing_httpg(DataBufferPar &buf) {
  string store_url(c_url.c_str());
  // registration at SE in case of SE url
  if(strncasecmp(url->current_location(),"se://",5) == 0) {
    // pass meta information directly to an SE
    string service_url = url->current_location();
    service_url.replace(0,2,"httpg");
    string service_path = "";
    {
      int n = service_url.find('?');
      if(n != string::npos) {
        service_path=service_url.c_str()+n+1;
        service_url.resize(n);
      };
    };
    canonic_url(service_url);
    store_url=service_url;
    odlog(1)<<"Talking to SOAP service at "<<service_url<<endl;
    struct soap soap;
    HTTP_ClientSOAP s(service_url.c_str(),&soap);
    soap.namespaces=file_soap_namespaces;
    if(s.connect() != 0) {
      odlog(0)<<"Failed to connect to "<<service_url<<endl;
      CHECK_PROXY("start_writing_httpg",failure_code);
      return false;
    };
    ns__fileinfo info;
    unsigned long long int size_ = url->meta_size();
    if(url->meta_size_available()) info.size=&size_;
    char checksum_[100];
    info.checksum=NULL;
    int l_ = snprintf(checksum_,99,"%16LX",url->meta_checksum()); 
    checksum_[l_]=0;
    if(url->meta_checksum_available()) info.checksum=checksum_;
    info.acl=NULL;
    string acl("");
    {
      CertInfo ci;
      acl=ci.GetSN();
      if(acl.length()) acl="<gacl><entry><person><dn>"+acl+"</dn></person><allow><read/><write/><list/><admin/></allow></entry></gacl>";
      info.acl=(char*)(acl.c_str());
    };
    if(url->lfn() && url->lfn()[0]) {
      info.id=strdup(url->lfn()); 
    } else if(service_path.length()) {
      info.id=strdup(service_path.c_str()); 
    } else {
      odlog(0)<<"Missing file id for SE URL"<<endl;
      return false;
    };
    info.created=NULL;
    if(url->meta_created_available()) {
      struct tm* t;
      struct tm tt;
      time_t t_ = url->meta_created();
      t=gmtime_r(&t_,&tt);
      if(t) {
        string created;
        if(timetostring(*t,created) == 0) {
          info.created=strdup(created.c_str());
        };
      };
    };
    odlog(2)<<"start_writing_httpg: id(lfn): "<<info.id<<endl;
    odlog(2)<<"start_writing_httpg: acl: "<<info.acl<<endl;
    odlog(2)<<"start_writing_httpg: created: "<<info.created<<endl;
    ns__addResponse rr;
    int soap_err = SOAP_OK;
    string soap_url = service_url;
    int n = soap_url.find(':');
    // make gSOAP create http headers (should be removed in a future)
    if(n != string::npos) soap_url.replace(0,n,"http");
    if((soap_err=soap_call_ns__add(&soap,soap_url.c_str(),"add",&info,0,NULL,
                                                 rr)) != SOAP_OK) {
      odlog(1)<<"Failed to execute remote soap call 'add' at "<<service_url<<endl;
      if(info.id) free(info.id);
      if(info.created) free(info.created);
      return false;
    };
    if(rr.error_code != 0) {
      odlog(1)<<"Failed ("<<rr.error_code<<") to create remote file "<<info.id<<" at "<<service_url<<endl;
      if(info.id) free(info.id);
      if(info.created) free(info.created);
      return false;
    };
    if(info.id) free(info.id);
    if(info.created) free(info.created);
    if((rr.file.url) && (rr.file.__size_url>0) && (rr.file.url[0])) {
      store_url=rr.file.url[0];
      // hack
      for(n=0;n<rr.file.__size_url;n++) {
        odlog(1)<<"File is available at: "<<rr.file.url[n]<<endl;
        if(strncasecmp(rr.file.url[n],"https://",8) == 0) // prefer SSL
          store_url=rr.file.url[n];
      };
    } else {
      if(store_url[store_url.length()-1] != '/') store_url+="/";
      if(url->lfn()[0] == '/') {
        store_url+=(url->lfn()+1);
      } else {
        store_url+=url->lfn();
      };
    };
    odlog(2)<<"Created slot at: "<<store_url<<endl;
  };
  // starting data transfer
  if(httpg_stat == NULL) httpg_stat=new httpg_info_t;
  httpg_stat->buffer=&buf; buffer=&buf;
  httpg_stat->url=strdup(store_url.c_str());
  odlog(2)<<"start_writing_httpg: url: "<<httpg_stat->url<<endl;
  httpg_stat->chunk_size=buf.buffer_size(); // not actually needed for writing
  httpg_stat->streams=ftp_threads;
  odlog(2)<<"start_writing_httpg: streams: "<<httpg_stat->streams<<endl;
  httpg_stat->requested_size=0;
  httpg_stat->threads=0;
  httpg_stat->cancel=false;
  httpg_stat->have_threads=false;
  httpg_stat->have_content_size=url->meta_size_available();
  httpg_stat->content_size=url->meta_size();
  if(httpg_stat->channels) free(httpg_stat->channels);
  httpg_stat->handle = this;
  httpg_stat->point = url;
  httpg_stat->failure_code=failure_code;
  httpg_stat->channels=(httpg_state_t*)malloc(sizeof(httpg_state_t)*httpg_stat->streams);
  if(httpg_stat->channels==NULL) return false;
  // start threads: one per stream
  httpg_stat->lock.reset();
  int started = 0;
  pthread_attr_t attr;
  if(pthread_attr_init(&attr) != 0) return false;
  if(pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED) != 0) return false;
  for(;started < httpg_stat->streams;started++) { 
  odlog(2)<<"start_writing_httpg: started: "<<started<<endl;
    httpg_state_t* tstat = httpg_stat->channels+started;
    if(pthread_create(&(tstat->thr),&attr,&write_thread,httpg_stat) != 0) break;
  };
  pthread_attr_destroy(&attr);
  odlog(2)<<"start_writing_httpg: started: "<<started<<endl;
  if(started==0) return false;
  return true;
}

extern double cpu_clock_1;
extern double cpu_clock_2;
extern double cpu_clock_3;
extern double cpu_clock_4;
extern double cpu_clock_5;
extern double cpu_clock_6;
extern double cpu_clock_7;
extern double cpu_clock_8;
extern double cpu_clock_9;
extern double cpu_clock_10;
extern double cpu_clock_11;
extern double wal_clock_1;
extern double wal_clock_2;
extern double wal_clock_3;
extern double wal_clock_4;
extern double wal_clock_5;
extern double wal_clock_6;
extern double wal_clock_7;
extern double wal_clock_8;
extern double wal_clock_9;
extern double wal_clock_10;
extern double wal_clock_11;

bool DataHandle::stop_writing_httpg(void) {
  httpg_stat->lock.block();
  failure_code=httpg_stat->failure_code;
  if(!buffer->eof_read()) {
    buffer->error_read(true);
    // cancel transfer
    httpg_stat->cancel=true;
    for(int i = 0;i<httpg_stat->streams;i++) {
      if(httpg_stat->channels[i].s) httpg_stat->channels[i].s->disconnect();
    };
  };
  while(httpg_stat->threads > 0) {
    httpg_stat->lock.wait_nonblock();
  };
  httpg_stat->lock.unblock();
  free((char*)(httpg_stat->url));
  delete httpg_stat;
/*
cerr<<"CPU clock 1: "<<cpu_clock_1/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 1: "<<wal_clock_1/1000<<endl;
cerr<<"CPU clock 2: "<<cpu_clock_2/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 2: "<<wal_clock_2/1000<<endl;
cerr<<"CPU clock 3: "<<cpu_clock_3/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 3: "<<wal_clock_3/1000<<endl;
cerr<<"CPU clock 4: "<<cpu_clock_4/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 4: "<<wal_clock_4/1000<<endl;
cerr<<"CPU clock 5: "<<cpu_clock_5/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 5: "<<wal_clock_5/1000<<endl;
cerr<<"CPU clock 6: "<<cpu_clock_6/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 6: "<<wal_clock_6/1000<<endl;
cerr<<"CPU clock 7: "<<cpu_clock_7/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 7: "<<wal_clock_7/1000<<endl;
cerr<<"CPU clock 8: "<<cpu_clock_8/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 8: "<<wal_clock_8/1000<<endl;
cerr<<"CPU clock 9: "<<cpu_clock_9/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 9: "<<wal_clock_9/1000<<endl;
cerr<<"CPU clock 10: "<<cpu_clock_10/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 10: "<<wal_clock_10/1000<<endl;
cerr<<"CPU clock 11: "<<cpu_clock_11/CLOCKS_PER_SEC<<endl;
cerr<<"WALL clock 11: "<<wal_clock_11/1000<<endl;
*/
  return true;
}

bool DataHandle::remove_httpg(void) {
  odlog(2)<<"DataHandle::remove_httpg: "<<url->current_location()<<" ("<<c_url<<")"<<endl;
  if(strncasecmp(url->current_location(),"se://",5) != 0) {
    odlog(0)<<"Removing for URL "<<url->current_location()<<" is not supported"<<endl;
    return false;
  };
  struct soap soap;
  HTTP_ClientSOAP s(c_url.c_str(),&soap);
  soap.namespaces=file_soap_namespaces;
  odlog(2)<<"DataHandle::remove_httpg: created HTTP_ClientSOAP"<<endl;
  if(s.connect() != 0) {
    odlog(0)<<"Failed to connect to "<<c_url<<endl;
    return false;
  };
  odlog(2)<<"DataHandle::remove_httpg: HTTP_ClientSOAP connected"<<endl;
  ns__delResponse rr;
  int soap_err = SOAP_OK;
  string soap_url = c_url;
  int n = soap_url.find(':');
  // make gSOAP create http headers (should be removed in a future)
  if(n != string::npos) soap_url.replace(0,n,"http");
  odlog(2)<<"DataHandle::remove_httpg: calling soap_call_ns__del"<<endl;
  if((soap_err=soap_call_ns__del(&soap,soap_url.c_str(),"del",rr))
                                                             != SOAP_OK) {
    odlog(1)<<"Failed to execute remote soap call 'del' at "<<c_url<<endl;
    return false;
  };
  if(rr.error_code != 0) {
    odlog(1)<<"Failed ("<<rr.error_code<<") to delete remote file "<<c_url<<endl;
    return false;
  };
  odlog(2)<<"DataHandle::remove_httpg: soap_call_ns__del finished"<<endl;
  return true;
}

static int check_callback(unsigned long long offset,unsigned long long size,unsigned char** buf,unsigned long long* bufsize,void* arg) {
  return 0;
}

bool DataHandle::check_httpg(void) {
  HTTP_Client s(c_url.c_str());
  if(s.connect() != 0) return false; // failed to connect
  int get_res = s.GET("",0,1,&check_callback,NULL); // recieve chunk
  if(get_res != 0) return false;
  unsigned long long int size = s.response().ContentSize();
  if(size) url->meta_size(size);
  if(s.response().haveLastModified()) {
    url->meta_created(s.response().LastModified());
  };
  return true;
}

bool DataHandle::list_files_httpg(list<DataPoint::FileInfo> &files,bool resolve) {
  odlog(2)<<"list_files_httpg"<<endl;

  if(strncasecmp(url->current_location(),"se://",5) != 0) {
    odlog(0)<<"Listing for URL "<<url->current_location()<<" is not supported"<<endl;
    return false;
  };
  struct soap soap;
  const char* service_pattern = NULL;
  string pattern(""); 
  get_url_option(url->base_url(),"pattern",0,pattern);
  if(strchr(url->current_location(),'?') == NULL) {
    service_pattern=".*";
    if(pattern.length()) service_pattern=pattern.c_str();
  };
  string service_url = c_url;

  HTTP_ClientSOAP s(service_url.c_str(),&soap);
  soap.namespaces=file_soap_namespaces;
  odlog(2)<<"Connecting to service at "<<service_url<<endl;
  if(s.connect() != 0) return false; // failed to connect
  ns__infoResponse rr;
  string soap_url = service_url;
  int n = soap_url.find(':');
  // make gSOAP create http headers (should be removed in a future)
  if(n != string::npos) soap_url.replace(0,n,"http");
  if(soap_call_ns__info(&soap,soap_url.c_str(),"info",(char*)service_pattern,rr) != SOAP_OK) {
    odlog(1)<<"Failed to execute remote soap call 'info' at "<<c_url<<endl;
    return false;
  };
  if(rr.error_code != 0) {
    odlog(1)<<"Failed ("<<rr.error_code<<") to list remote files at "<<c_url<<endl;
    return false;
  };
  for(int i = 0;i<rr.__size_file;i++) {
    char* id = rr.file[i].id;
    if(id) {
      list<DataPoint::FileInfo>::iterator f =
          files.insert(files.end(),DataPoint::FileInfo(id));
      f->type=DataPoint::FileInfo::file_type_file;
      if(rr.file[i].size) {
        f->size=*(rr.file[i].size); f->size_available=true;
      };
      // char* checksum;
      // char* acl;
      // char* created;
      // enum ns__filestate* state;
      // int __size_url;
      // char** url;
    };
  };
  return true;
}

