#include "../../std.h"

#include <sys/timeb.h>
#include <globus_io.h>

#include "../../misc/condition.h"
#include "../../misc/url_options.h"
#include "../../misc/globus_error_utils.h"
#include "../../misc/log_time.h"



#include "../../misc/inttostring.h"
// #include "../../auth/auth.h"

#include "client.h"

double cpu_clock_1 = 0.0;
double cpu_clock_2 = 0.0;
double cpu_clock_3 = 0.0;
double cpu_clock_4 = 0.0;
double cpu_clock_5 = 0.0;
double cpu_clock_6 = 0.0;
double cpu_clock_7 = 0.0;
double cpu_clock_8 = 0.0;
double cpu_clock_9 = 0.0;
double cpu_clock_10 = 0.0;
double cpu_clock_11 = 0.0;
double wal_clock_1 = 0.0;
double wal_clock_2 = 0.0;
double wal_clock_3 = 0.0;
double wal_clock_4 = 0.0;
double wal_clock_5 = 0.0;
double wal_clock_6 = 0.0;
double wal_clock_7 = 0.0;
double wal_clock_8 = 0.0;
double wal_clock_9 = 0.0;
double wal_clock_10 = 0.0;
double wal_clock_11 = 0.0;


clock_t last_cpu_clock;
unsigned long last_wal_clock;

HTTP_Client::HTTP_Client(const char* base,bool heavy_encryption):base_url(base) {
  timeout=60000; // 1 min.
  valid=false; connected=false;
  /* initialize globus io connection related objects */
  globus_io_tcpattr_init(&attr);
  globus_io_secure_authorization_data_initialize(&auth);
  globus_io_secure_authorization_data_set_callback(&auth,&authorization_callback,NULL);
  if(!base_url) return;
  if(strcasecmp(base_url.Proto().c_str(),"http") == 0) {
    const char * proxy = getenv("NORDUGRID_HTTP_PROXY");
    if(proxy) {
      proxy_hostname = proxy;
      proxy_port = 8000;
      int n = proxy_hostname.find(':');
      if(n != string::npos) {
        proxy_port=atoi(proxy_hostname.c_str()+n+1);
        proxy_hostname.resize(n);
      };
    };
    globus_io_attr_set_secure_authentication_mode(&attr,GLOBUS_IO_SECURE_AUTHENTICATION_MODE_NONE,GSS_C_NO_CREDENTIAL);
    globus_io_attr_set_secure_authorization_mode(&attr,GLOBUS_IO_SECURE_AUTHORIZATION_MODE_NONE,GLOBUS_NULL);
    globus_io_attr_set_secure_channel_mode(&attr,
                                 GLOBUS_IO_SECURE_CHANNEL_MODE_CLEAR);
    globus_io_attr_set_secure_protection_mode(&attr,
                                 GLOBUS_IO_SECURE_PROTECTION_MODE_NONE);
    globus_io_attr_set_secure_delegation_mode(&attr,
                                 GLOBUS_IO_SECURE_DELEGATION_MODE_NONE);
  } else if(strcasecmp(base_url.Proto().c_str(),"https") == 0) {
    globus_io_attr_set_secure_authentication_mode(&attr,GLOBUS_IO_SECURE_AUTHENTICATION_MODE_MUTUAL,GSS_C_NO_CREDENTIAL);
    globus_io_attr_set_secure_authorization_mode(&attr,GLOBUS_IO_SECURE_AUTHORIZATION_MODE_HOST,GLOBUS_NULL);
    globus_io_attr_set_secure_channel_mode(&attr,
                                 GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP);
    if(heavy_encryption) {
      globus_io_attr_set_secure_protection_mode(&attr,
                                 GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
    } else {
      globus_io_attr_set_secure_protection_mode(&attr,
                                 GLOBUS_IO_SECURE_PROTECTION_MODE_SAFE);
    };
    globus_io_attr_set_secure_delegation_mode(&attr,
                                 GLOBUS_IO_SECURE_DELEGATION_MODE_NONE);
  } else if(strcasecmp(base_url.Proto().c_str(),"httpg") == 0) {
    globus_io_attr_set_secure_authentication_mode(&attr,GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI,GSS_C_NO_CREDENTIAL);
    globus_io_attr_set_secure_authorization_mode(&attr,GLOBUS_IO_SECURE_AUTHORIZATION_MODE_HOST,GLOBUS_NULL);
    globus_io_attr_set_secure_channel_mode(&attr,
                                 GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP);
    if(heavy_encryption) {
      globus_io_attr_set_secure_protection_mode(&attr,
                                 GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
    } else {
      globus_io_attr_set_secure_protection_mode(&attr,
                                 GLOBUS_IO_SECURE_PROTECTION_MODE_SAFE);
    };
    globus_io_attr_set_secure_delegation_mode(&attr,
                                 GLOBUS_IO_SECURE_DELEGATION_MODE_FULL_PROXY);
  } else { return; };
  globus_io_attr_set_secure_proxy_mode(&attr,GLOBUS_IO_SECURE_PROXY_MODE_NONE);
  valid=true;
}

#if GLOBUS_IO_VERSION<4
globus_bool_t HTTP_Client::authorization_callback(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t* context_handle) {
#else
globus_bool_t HTTP_Client::authorization_callback(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t context_handle) {
#endif
  olog<<"Authenticating: "<<identity<<endl;
//  HTTPS_Connector::identity(h,identity,h->delegated_credential);
  return GLOBUS_TRUE;
}

void HTTP_Client::general_callback(void *arg,globus_io_handle_t *handle,globus_result_t  result) {
  HTTP_Client* it = (HTTP_Client*)arg;
  if(result != GLOBUS_SUCCESS) {
    olog<<"Globus error: "<<GlobusResult(result)<<endl;
    it->cond.signal(-1);
  } else {
    it->cond.signal(0);
  };
}

void HTTP_Client::write_callback(void *arg,globus_io_handle_t *handle,globus_result_t result,globus_byte_t *buf,globus_size_t nbytes) {
  HTTP_Client* it = (HTTP_Client*)arg;
  if(result != GLOBUS_SUCCESS) {
    olog<<"Globus error (write): "<<GlobusResult(result)<<endl;
    it->cond_write.signal(-1);
  } else {
    it->cond_write.signal(0);
  };
}

void HTTP_Client::write_header_callback(void *arg,globus_io_handle_t *handle,globus_result_t result,globus_byte_t *buf,globus_size_t nbytes) {

  unsigned long new_wal_clock;
  struct timeb wal_clock;
  clock_t new_cpu_clock;
  new_cpu_clock=clock();
  cpu_clock_1 += new_cpu_clock-last_cpu_clock;
  last_cpu_clock=new_cpu_clock;
  ftime(&wal_clock); new_wal_clock=wal_clock.time*1000+wal_clock.millitm;
  wal_clock_1 += new_wal_clock-last_wal_clock;
  last_wal_clock=new_wal_clock;

  HTTP_Client* it = (HTTP_Client*)arg;
  it->head_counter++;
  if(result != GLOBUS_SUCCESS) {
    olog<<"Globus error (write): "<<GlobusResult(result)<<endl;
    it->cond_write.signal(-1);
  } else {
    globus_result_t res=globus_io_register_write(&it->s,
        (globus_byte_t*)(it->write_buf),it->write_size,&write_body_callback,it);
    if(res != GLOBUS_SUCCESS) {
      it->cond_write.signal(-1);
    } else {
      it->cond_write.signal(0);
    };
  };
}

void HTTP_Client::write_body_callback(void *arg,globus_io_handle_t *handle,globus_result_t result,globus_byte_t *buf,globus_size_t nbytes) {

  unsigned long new_wal_clock;
  struct timeb wal_clock;
  clock_t new_cpu_clock;
  new_cpu_clock=clock();
  cpu_clock_2 += new_cpu_clock-last_cpu_clock;
  last_cpu_clock=new_cpu_clock;
  ftime(&wal_clock); new_wal_clock=wal_clock.time*1000+wal_clock.millitm;
  wal_clock_2 += new_wal_clock-last_wal_clock;
  last_wal_clock=new_wal_clock;

  HTTP_Client* it = (HTTP_Client*)arg;
  it->body_counter++;
  if(result != GLOBUS_SUCCESS) {
    olog<<"Globus error (write): "<<GlobusResult(result)<<endl;
    it->cond_write.signal(-1);
  } else {
    it->cond_write.signal(0);
  };
}

void HTTP_Client::read_callback(void *arg,globus_io_handle_t *handle,globus_result_t result,globus_byte_t *buf,globus_size_t nbytes) {
  HTTP_Client* it = (HTTP_Client*)arg;
  int res = 0;
  if(result != GLOBUS_SUCCESS) {
    globus_object_t* err = globus_error_get(result);
    char* tmp=globus_object_printable_to_string(err);
    if(strstr(tmp,"end-of-file") != NULL) {
      res=1; //
      odlog(3)<<"Connection closed"<<endl;
    } else {
      olog<<"Globus error (read): "<<tmp<<endl;
      res=-1;
    };    
    free(tmp); globus_object_free(err);
  } else {
    it->answer_size=nbytes;
    odlog(3)<<"*** Server response: ";
    for(globus_size_t n=0;n<nbytes;n++) odlog_(3)<<buf[n];
    odlog_(3)<<endl;
  };
  it->cond_read.block(); it->cond_write.block();
  it->cond_read.signal_nonblock(res);
  if(!it->cond_write.check()) it->cond_write.signal_nonblock(-1);
  it->cond_write.unblock(); it->cond_read.unblock();
}

int HTTP_Client::connect(void) {
  if(connected) return 0;
  globus_result_t res;
  cond.reset();
  if(proxy_hostname.length() == 0) {
    if((res=globus_io_tcp_register_connect(
                    (char*)(base_url.Host().c_str()),base_url.Port(),
                    &attr,&general_callback,this,&s)) != GLOBUS_SUCCESS) {
      olog<<"Connect failed: "<<GlobusResult(res)<<endl;
      return -1;
    };
  } else {
    if((res=globus_io_tcp_register_connect(
                    (char*)(proxy_hostname.c_str()),proxy_port,
                    &attr,&general_callback,this,&s)) != GLOBUS_SUCCESS) {
      olog<<"Connect to proxy failed: "<<GlobusResult(res)<<endl;
      return -1;
    };
  };
  int r;
  globus_thread_blocking_will_block(); // ????
  if(!cond.wait(r,timeout)) {  // timeout
    olog<<"Connection timeout"<<endl;
    globus_io_cancel(&s,GLOBUS_FALSE);
    globus_io_close(&s);
    return -1;
  };
  if(r != 0) {
    olog<<"Connection failed"<<endl;
    globus_io_close(&s); // just in case
    return -1;
  };
  connected=true;
  return 0;
}

int HTTP_Client::disconnect(void) {
  if(!connected) return 0;
  globus_io_cancel(&s,GLOBUS_FALSE);
  globus_io_close(&s);
  connected=false;
  return 0;
}

HTTP_Client::~HTTP_Client(void) {
  disconnect();
  globus_io_secure_authorization_data_destroy(&auth);
  globus_io_tcpattr_destroy(&attr);
}

void HTTP_Client::clear_input(void) {
  if(!valid) return;
  globus_byte_t buf[256];
  globus_size_t l;
  for(;;) {
    if(globus_io_read(&s,buf,256,0,&l) != GLOBUS_SUCCESS) return;
    if(l == 0) return;
    odlog(3)<<"clear_input: ";
    for(int n=0;n<l;n++) odlog_(3)<<buf[n];
    odlog_(3)<<endl;
  };
}

int HTTP_Client::analyze_response_line(char* line) {
  for(;*line;line++) if(!isspace(*line)) break;
  int len = strlen(line);
  if(len<2) return -1; // too short
  if(answer_count==0) {  // first line
    // odlog(2)<<"analyze_response_line: first line: "<<line<<endl;
    bool answer_keep_alive = true;
    answer_code=0;
    char* p = line;
    char* http_version = p;
    for(;*p;p++) if(isspace(*p)) break; (*p)=0; p++;
    for(;*p;p++) if(!isspace(*p)) break;
    char* code = p;
    for(;*p;p++) if(isspace(*p)) break; (*p)=0; p++;
    for(;*p;p++) if(!isspace(*p)) break;
    char* reason = p;
    char *e;
    answer_code=strtoul(code,&e,10);
    if((*e) != 0) return -1;
    answer_reason=reason;
    answer_count++;
    if(strcmp(http_version,"HTTP/1.1") == 0) { answer_keep_alive=true; }
    else { answer_keep_alive=false; };
    fields.reset(answer_keep_alive);
    // odlog(2)<<"analyze_response_line: answer_code:"<<answer_code<<endl;
    // odlog(2)<<"analyze_response_line: http_version:"<<http_version<<endl;
    // odlog(2)<<"analyze_response_line: reason:"<<reason<<endl;
  } else  {
    // odlog(2)<<"analyze_response_line: line: "<<line<<endl;
    char* token = line; for(;*token;token++) if(isspace(*token)) break;
    int l = token-line;
    if(*token) {
      (*token)=0; token++;
      for(;*token;token++) if(!isspace(*token)) break;
    };
    // odlog(2)<<"analyze_response_line: name: "<<line<<endl;
    // odlog(2)<<"analyze_response_line: value: "<<token<<endl;
    fields.set(line,token);
  };
}

// read server response
int HTTP_Client::read_response_header(void) {
  int r;
  answer_count=0; // line counter
  // if(!cond_read.wait(r,timeout)) {
  if(!cond_read.wait(r,-1)) { // let external procedures manage timeouts
    globus_io_cancel(&s,GLOBUS_FALSE); return -1;
  };

  unsigned long new_wal_clock;
  struct timeb wal_clock;
  clock_t new_cpu_clock;
  new_cpu_clock=clock();
  cpu_clock_3 += new_cpu_clock-last_cpu_clock;
  last_cpu_clock=new_cpu_clock;
  ftime(&wal_clock); new_wal_clock=wal_clock.time*1000+wal_clock.millitm;
  wal_clock_3 += new_wal_clock-last_wal_clock;
  last_wal_clock=new_wal_clock;

  if(r != 0) return -1;
  char line_buf[256];
  int line_p = 0;
  for(;;) {
    answer_buf[answer_size]=0;
    char* p = strchr(answer_buf,'\n');
    int l = answer_size;
    if(p) l=(p-answer_buf)+1; // available data
    int ll = (sizeof(line_buf)-1-line_p);
    if(ll>l) ll=l; // available space
    memcpy(line_buf+line_p,answer_buf,ll); line_p+=ll; line_buf[line_p]=0;
    if(l<answer_size) { memmove(answer_buf,answer_buf+l,answer_size-l); };
    answer_size-=l;
    if(p) {  // eol 
      for(;line_p;) {
        if((line_buf[line_p-1] != '\r') && (line_buf[line_p-1] != '\n')) break;
        line_p--;
      };
      line_buf[line_p]=0;
      if(line_p == 0) break; // end of header
      odlog(3)<<"read_response_header: line: "<<line_buf<<endl;
      analyze_response_line(line_buf);
      line_p=0;
    };
    if(answer_size>0) continue;
    // request new portion
    globus_result_t res;
    res=globus_io_register_read(&s,(globus_byte_t*)answer_buf,
                        sizeof(answer_buf)-1,1,&read_callback,this);
    if(res != GLOBUS_SUCCESS) {
      olog<<"globus_io_register_read failed: "<<GlobusResult(res)<<endl;
      disconnect();
      return -1;
    };
    if(!cond_read.wait(r,timeout)) {
      olog<<"Timeout while reading response header"<<endl;
      disconnect();
      return -1;
    };
    if(r != 0) {
      olog<<"Error while reading response header"<<endl;
      disconnect();
      return -1;
    };
  };
  odlog(3)<<"read_response_header: header finished"<<endl;
  return 0;
}

int HTTP_Client::skip_response_entity(void) {
  odlog(3)<<"skip_response_entity"<<endl;
  if(fields.haveContentLength() || fields.haveContentRange()) {
    unsigned long long int size = fields.ContentLength();
    odlog(3)<<"skip_response_entity: size: "<<size<<endl;
    if(size<=answer_size) {
      memmove(answer_buf,answer_buf+size,answer_size-size);
      answer_size-=size;
      odlog(3)<<"skip_response_entity: already have all"<<endl;
      return 0;
    };
    size-=answer_size;
    odlog(3)<<"skip_response_entity: size left: "<<size<<endl;
    // read size bytes
    char buf[1024];
    for(;size;) {
    odlog(3)<<"skip_response_entity:  to read: "<<size<<endl;
      cond_read.reset();
      globus_result_t res=globus_io_register_read(&s,(globus_byte_t*)buf,
                                     sizeof(buf),1,&read_callback,this);
      if(res != GLOBUS_SUCCESS) { // failure
        olog<<GlobusResult(res)<<endl; return -1;
      };
      int r;
      if(!cond_read.wait(r,timeout)) {
        odlog(3)<<"skip_response_entity: timeout"<<size<<endl;
        globus_io_cancel(&s,GLOBUS_FALSE); return -1; // timeout
      };
      odlog(3)<<"skip_response_entity: cond_read: "<<r<<endl;
      if(r != 0) return -1; // failure
      size-=answer_size;
      odlog(3)<<"skip_response_entity: read: "<<answer_size<<" ("<<size<<")"<<endl;
    };
    odlog(3)<<"skip_response_entity: read all"<<endl;
    return 0;
  };
  if(fields.KeepAlive()) {
    odlog(3)<<"skip_response_entity: no entity"<<endl;
    // no entity passed at all
    return 0;
  };
  odlog(3)<<"skip_response_entity: unknown size"<<endl;
  // can't handle unknow size - let connection be closed
  return 0;
}

//int HTTP_Client::put(const char* path,
//    unsigned long long int offset,unsigned long long int size,
//    int fd,unsigned long long int fd_offset,unsigned long long int fd_size) {

/*
  HTTP PUT method
  Content of 'buf' with size 'size' is sent to 'path'. 
  Treated as part size 'fd_size' starting from 'offset'.
*/
int HTTP_Client::make_header(const char* path,
      unsigned long long int offset,unsigned long long int size,
      unsigned long long int fd_size,string& header) {
  // create header
  if(path[0] == '/') path++;
  header = "PUT ";
  string url_path;
  if(proxy_hostname.length() == 0) {
    url_path="/"+base_url.Path();
  } else {
    url_path=base_url.Proto()+"://"+base_url.Host()+":"+inttostring(base_url.Port())+"/"+base_url.Path();
  };
  if(path[0]) {
    if(url_path[url_path.length()-1] != '/') url_path+="/";
    url_path+=path;
  };
  string url_host = base_url.Host()+":"+inttostring(base_url.Port());
  header+=url_path; header+=" HTTP/1.1\r\n";
  header+="Host: "+url_host+"\r\n";
  header+="Connection: keep-alive\r\n";
  header+="Content-Length: "+inttostring(size)+"\r\n";
  header+="Content-Range: bytes "+inttostring(offset)+"-"+inttostring(offset+size-1);
  if(fd_size>=size) {
    header+="/"+inttostring(fd_size);
  };
  header+="\r\n";
  header+="\r\n";
  return 0;
}

int HTTP_Client::PUT_header(const char* path,
      unsigned long long int offset,unsigned long long int size,
      unsigned long long int fd_size,bool wait) {
  // create header
//  if(path[0] == '/') path++;
  string header;
  make_header(path,offset,size,fd_size,header);
  globus_result_t res;
  int r;
  cond_write.reset(); clear_input();
  head_counter=0; body_counter=0;
  // get ready for negative answer
  answer_size=0;
  res=globus_io_register_read(&s,(globus_byte_t*)answer_buf,
                      sizeof(answer_buf)-1,1,&read_callback,this);
  if(res != GLOBUS_SUCCESS) {
    olog<<GlobusResult(res)<<endl;
    return -1;
  };
  // send header
  odlog(3)<<"header: "<<header<<endl;
  if(!wait) {
    globus_byte_t* headbuf = (globus_byte_t*)malloc(header.length()+1);
    if(!headbuf) {
      globus_io_cancel(&s,GLOBUS_FALSE); // cancel waiting read
      olog<<"malloc failed"<<endl; return -1;
    };
    strcpy((char*)headbuf,header.c_str());
    res=globus_io_register_write(&s,headbuf,
                      header.length(),&write_header_callback,this);
    if(res != GLOBUS_SUCCESS) {
      globus_io_cancel(&s,GLOBUS_FALSE); // cancel waiting read
      olog<<GlobusResult(res)<<endl; return -1;
    };
    return 0;
  };
  res=globus_io_register_write(&s,(globus_byte_t*)(header.c_str()),
                      header.length(),&write_callback,this);
  if(res != GLOBUS_SUCCESS) {
    globus_io_cancel(&s,GLOBUS_FALSE); // cancel waiting read
    olog<<GlobusResult(res)<<endl;
    return -1;
  };
  globus_thread_blocking_will_block(); // ????
  if(!cond_write.wait(r,timeout)) {
    olog<<"Timeout while sending header"<<endl;
    globus_io_cancel(&s,GLOBUS_FALSE);
    return -1;
  };
  if(cond_read.check()) {  // server already answered
    olog<<"Early answer from server"<<endl;
    globus_io_cancel(&s,GLOBUS_FALSE);
    return -1;
  };
  if(r != 0) {
    olog<<"Failed to send header"<<endl;
    globus_io_cancel(&s,GLOBUS_FALSE);
    return -1;
  };
  return 0;
}

int HTTP_Client::PUT(const char* path,
      unsigned long long int offset,unsigned long long int size,
      const unsigned char* buf,unsigned long long int fd_size,bool wait) {
  if(!connected) {
    olog<<"Not connected"<<endl;
    return -1;
  };

  unsigned long new_wal_clock;
  struct timeb wal_clock;
  clock_t new_cpu_clock;

  // send header
  // create header
  string header;
  make_header(path,offset,size,fd_size,header);
  globus_result_t res;
  int r;
  cond_write.reset(); clear_input();
  head_counter=0; body_counter=0;
  // get ready for negative answer
  answer_size=0;
  res=globus_io_register_read(&s,(globus_byte_t*)answer_buf,
                      sizeof(answer_buf)-1,1,&read_callback,this);
  if(res != GLOBUS_SUCCESS) {
    olog<<GlobusResult(res)<<endl; disconnect(); return -1;
  };
  // prepare body
  write_size=size; write_buf=buf;
  // send header
  odlog(3)<<"header: "<<header<<endl;
last_cpu_clock = clock();
ftime(&wal_clock); last_wal_clock=wal_clock.time*1000+wal_clock.millitm;
  res=globus_io_register_write(&s,(globus_byte_t*)(header.c_str()),
                      header.length(),&write_header_callback,this);
  if(res != GLOBUS_SUCCESS) {
    globus_io_cancel(&s,GLOBUS_FALSE); // cancel waiting read
    olog<<GlobusResult(res)<<endl; disconnect(); return -1;
  };
  // read server response
  if(read_response_header() != 0) {
    olog<<"No response from server received"<<endl;
    disconnect(); return -1;
  };
  // make sure all buffers sent
  while((!head_counter) || (!body_counter)) {
    if(!cond_write.wait(r,timeout)) { 
      olog<<"Timeout waiting for buffers"<<endl;
      disconnect(); return -1;
    };
    if(r != 0) { // at least one operation failed - no need to wait
      olog<<"Error waiting for buffers"<<endl;
      disconnect(); return -1;
    };
  };
  if(fields.KeepAlive()) {  // skip entity only if trying to keep connection
    if(skip_response_entity() != 0) {
      olog<<"Failure while receiving entity"<<endl;
      disconnect();
      return -1;
    };
    globus_io_cancel(&s,GLOBUS_FALSE); // just in case
  } else {
    disconnect();
  };
  if(answer_code != HTTP_OK) return -1;
  return 0;
}

/*
  HTTP GET method
*/
int HTTP_Client::GET_header(const char* path,
      unsigned long long int offset,unsigned long long int size) {
  // create header
  if(path[0] == '/') path++;
  string header = "GET ";
  string url_path;
  if(proxy_hostname.length() == 0) {
    url_path="/"+base_url.Path();
  } else {
    url_path=base_url.Proto()+"://"+base_url.Host()+":"+inttostring(base_url.Port())+"/"+base_url.Path();
  };
  if(path[0]) {
    if(url_path[url_path.length()-1] != '/') url_path+="/"; 
    url_path+=path;
  };
  string url_host = base_url.Host()+":"+inttostring(base_url.Port());
  header+=url_path; header+=" HTTP/1.1\r\n";
  header+="Host: "+url_host+"\r\n";
  header+="Connection: keep-alive\r\n";
  header+="Range: bytes="+inttostring(offset)+"-"+inttostring(offset+size-1)+"\r\n";
  header+="\r\n";
  odlog(3)<<"header: "<<header<<endl;
  globus_result_t res;
  int r;
  cond_write.reset(); clear_input();
  // get ready for any answer
  answer_size=0;
  res=globus_io_register_read(&s,(globus_byte_t*)answer_buf,
                      sizeof(answer_buf)-1,1,&read_callback,this);
  if(res != GLOBUS_SUCCESS) {
    olog<<"globus_io_register_read failed: "<<GlobusResult(res)<<endl;
    disconnect();
    return -1;
  };
  // send header
  res=globus_io_register_write(&s,(globus_byte_t*)(header.c_str()),
                      header.length(),&write_callback,this);
  if(res != GLOBUS_SUCCESS) {
    olog<<"globus_io_register_write failed: "<<GlobusResult(res)<<endl;
    disconnect();
    return -1;
  };
  globus_thread_blocking_will_block(); // ????
  if(!cond_write.wait(r,timeout)) {
    olog<<"Timeout while sending header"<<endl;
    disconnect();
    return -1;
  };
  if(r != 0) {
    olog<<"Failed to send header"<<endl;
    disconnect();
    return -1;
  };
  return 0;
}

int HTTP_Client::GET(const char* path,
      unsigned long long int offset,unsigned long long int size,
      get_callback_t callback,void* arg,
      unsigned char* buf,unsigned long long int bufsize) {
  odlog(2)<<"GET: path: "<<path<<endl;
  odlog(2)<<"GET: offset: "<<offset<<endl;
  odlog(2)<<"GET: size: "<<size<<endl;
  if(!connected) {
    olog<<"Not connected"<<endl;
    return -1;
  };
  // send header
  if(GET_header(path,offset,size)) return -1;
  // read server response
  if(read_response_header() != 0) {
    olog<<"No response from server received"<<endl;
    disconnect();
    return -1;
  };
  if(answer_code == 416) { // out of range
    if(skip_response_entity() != 0) {
      disconnect(); return -1;
    };
    if(!fields.KeepAlive()) {
    odlog(2)<<"GET: connection to be closed"<<endl;
      disconnect();
    };
    return 0;
  };
  if((answer_code != 200) && (answer_code != 206)) {
    if(skip_response_entity() != 0) {
      disconnect(); return -1;
    };
    if(!fields.KeepAlive()) {
    odlog(2)<<"GET: connection to be closed"<<endl;
      disconnect();
    };
    return -1;
  };
  odlog(2)<<"GET: header is read - rest: "<<answer_size<<endl;
  unsigned long long c_offset = 0;
  if(fields.haveContentRange()) c_offset=fields.ContentStart();
  bool have_length = fields.haveContentLength() || fields.haveContentRange();
  unsigned long long length = fields.ContentLength();
  // take rest of already read data
  if(answer_size) {
    if(have_length) if(answer_size > length) answer_size=length;
    odlog(3)<<"GET: calling callback(rest): content: "<<answer_buf<<endl;
    odlog(3)<<"GET: calling callback(rest): size: "<<answer_size<<endl;
    odlog(3)<<"GET: calling callback(rest): offset: "<<c_offset<<endl;
    unsigned char* in_buf = (unsigned char*)answer_buf;
    unsigned long  in_size = answer_size;
    for(;;) {
      if(in_size == 0) break;
      if(buf) {
        unsigned long l = in_size;
        if(l>bufsize) l=bufsize;
        memcpy(buf,in_buf,l);
        if(callback(c_offset,l,&buf,&bufsize,arg) != 0) {
          olog<<"GET callback returned error"<<endl;
          disconnect(); // ?????????????????
          return -1;
        };
        in_buf+=l; c_offset+=l; in_size-=l;
      } else {
        unsigned char* in_buf_ = in_buf;
        if(callback(c_offset,in_size,&in_buf_,&bufsize,arg) != 0) {
          olog<<"GET callback returned error"<<endl;
          disconnect(); // ?????????????????
          return -1;
        };
        if(in_buf_ != in_buf) buf=in_buf_;
        in_buf+=in_size; c_offset+=in_size; in_size=0;
      };
    };
    if(have_length) length-=answer_size;
  };
  unsigned char* in_buf = NULL;
  for(;;) {
    if(have_length) if(length == 0) break; 
    if(buf == NULL) {
      if(in_buf == NULL) in_buf=(unsigned char*)malloc(65536);
      if(in_buf == NULL) {
        olog<<"Failed to allocate memory"<<endl;
        disconnect();
        return -1;
      };
      buf=in_buf; bufsize=65536;
    };
    cond_read.reset();
    odlog(3)<<"GET: calling globus_io_register_read"<<endl;
    globus_result_t res=globus_io_register_read(&s,(globus_byte_t*)buf,
                                     bufsize,1,&read_callback,this);
    if(res != GLOBUS_SUCCESS) { // failure
      olog<<"Failed while reading response content: "<<GlobusResult(res)<<endl;
      disconnect();
      if(in_buf) free(in_buf);
      return -1;
    };
    int r;
    if(!cond_read.wait(r,timeout)) {
      olog<<"Timeout while reading response content: "<<endl;
      disconnect();
      if(in_buf) free(in_buf);
      return -1; // timeout
    };
    if(r != 0) { // failure
      if(r == 1) { // eof - ok if length not specified
        if(!have_length) { disconnect(); break; };
      };
      olog<<"Error while reading response content: "<<endl;
      disconnect();
      if(in_buf) free(in_buf);
      return -1;
    };
    odlog(3)<<"GET: calling callback: content: "<<buf<<endl;
    odlog(3)<<"GET: calling callback: size: "<<answer_size<<endl;
    odlog(3)<<"GET: calling callback: offset: "<<c_offset<<endl;
    if(callback(c_offset,answer_size,&buf,&bufsize,arg) != 0) {
      olog<<"GET callback returned error"<<endl;
      disconnect();
      if(in_buf) free(in_buf);
      return -1;
    };
    c_offset+=answer_size;
    if(have_length) length-=answer_size;
  };
  globus_io_cancel(&s,GLOBUS_FALSE); // just in case
  if(in_buf) free(in_buf);
  if(!fields.KeepAlive()) {
    odlog(2)<<"GET: connection to be closed"<<endl;
    disconnect();
  };
  return 0;
}

