#include "../../std.h"
#include <iostream>
#include <fstream>

#include "../../misc/inttostring.h"
#include "../../misc/stringtoint.h"
#include "../../misc/escaped.h"
#include "../../misc/log_time.h"
#include "../../config/environment.h"
#include "../../auth/auth.h"

#include "../logger.h"
#include "logger_soap.nsmap"

class acl_t {
 public:
  list<string> read;
  list<string> write;
  list<string> query;
  string sqluser;
  string sqlpassword;
};

HTTP_Service* logger_service_creator(HTTPS_Connector& c,const char* uri,void* arg) {
  acl_t* acl = (acl_t*)arg;
  /* check authorization */
  bool acl_read=false;
  bool acl_write=false;
  bool acl_query=false;
  list<AuthEvaluator*>& auths = c.authorizations();
  for(list<AuthEvaluator*>::iterator ii=auths.begin();ii!=auths.end();++ii){
    if(!acl_read)
    for(list<string>::iterator i=acl->read.begin();i!=acl->read.end();++i) {
      if((*(*ii)) == (*i)) {
        if((*(*ii)).evaluate(c.identity()) == AAA_POSITIVE_MATCH) {
          odilog(1,c.pid)<<"User is granted 'read' access through acl '"<<(*(*ii)).get_name()<<"'"<<endl;
          acl_read=true;
        };
        break;
      };
    };
    if(!acl_write)
    for(list<string>::iterator i=acl->write.begin();i!=acl->write.end();++i) {
      if((*(*ii)) == (*i)) {
        if((*(*ii)).evaluate(c.identity()) == AAA_POSITIVE_MATCH) {
          odilog(1,c.pid)<<"User is granted 'write' access through acl '"<<(*(*ii)).get_name()<<"'"<<endl;
          acl_write=true;
        };
        break;
      };
    };
    if(!acl_query)
    for(list<string>::iterator i=acl->query.begin();i!=acl->query.end();++i) {
      if((*(*ii)) == (*i)) {
        if((*(*ii)).evaluate(c.identity()) == AAA_POSITIVE_MATCH) {
          odilog(1,c.pid)<<"User is granted 'full query' access through acl '"<<(*(*ii)).get_name()<<"'"<<endl;
          acl_query=true;
        };
        break;
      };
    };
    if(acl_read && acl_write && acl_query) break;
  };
  HTTP_Logger* h = new HTTP_Logger(&c,acl_read,acl_write,acl_query,
              acl->sqluser.length()?acl->sqluser.c_str():NULL,
              acl->sqlpassword.length()?acl->sqlpassword.c_str():NULL);
  return h;
}

bool logger_service_configurator(istream& f,const char* uri,HTTP_Service_Properties &prop) {
  acl_t* acl = new acl_t;
  acl->sqluser="nglogger";
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    for(;*p;p++) if(!isspace(*p)) break;
    char* command = p;
    if(*command == '#') continue;
    for(;*p;p++) if(isspace(*p)) break;
    int command_len = p-command;
    if(!command_len) continue;
    if((command_len==3) && (strncmp(command,"end",command_len)==0)) { break; }
    else if((command_len==8) && (strncmp(command,"acl_read",command_len)==0)) {
      for(;;) {
        string name;
        int n = input_escaped_string(p,name);
        if(n == 0) break;
        p+=n;
        acl->read.push_back(name);
      };
    } else if((command_len==9) && (strncmp(command,"acl_write",command_len)==0)) {
      for(;;) {
        string name;
        int n = input_escaped_string(p,name);
        if(n == 0) break;
        p+=n;
        acl->write.push_back(name);
      };
    } else if((command_len==9) && (strncmp(command,"acl_query",command_len)==0)) {
      for(;;) {
        string name;
        int n = input_escaped_string(p,name);
        if(n == 0) break;
        p+=n;
        acl->query.push_back(name);
      };

    } else if((command_len==10) && (strncmp(command,"sqlcontact",command_len)==0)) {
      int n;
      n=input_escaped_string(p,acl->sqluser);
      if(n) {
        p+=n;
        n=input_escaped_string(p,acl->sqlpassword);
      } else {
        acl->sqlpassword="";
      };
    } else if((command_len==16) && (strncmp(command,"sqlcontactsource",command_len)==0)) {
      string path;
      int n;
      n=input_escaped_string(p,path);
      acl->sqluser="";
      acl->sqlpassword="";
      if(n) {
        ifstream sf(path.c_str());
        if(sf) {
          char buf[1024];
          istream_readline(sf,buf,sizeof(buf));
          char* p = buf;
          n=input_escaped_string(p,acl->sqluser);
          if(n) {
            p+=n;
            n=input_escaped_string(p,acl->sqlpassword);
          };
        };
      };
    };
  };
  prop.arg=acl;
  prop.subtree=false;
  return true;
}

static char* soap_strdup_l(soap* sp,const char* s,int l = -1) {
  if(l<0) l=strlen(s);
  char* s_ = (char*)soap_malloc(sp,l+1);
  if(s_ == NULL) return NULL;
  memcpy(s_,s,l); s_[l]=0;
  return s_;
}

static string sql_string(const char* s) {
  string str(s);
  int p = 0;
  for(;;) {
    p=str.find('\'',p);
    if(p == string::npos) break;
    str.insert(p,"\\"); p+=2;
  };
  return str;
}

static inline string sql_string(const string& s) {
  return sql_string(s.c_str());
}

static time_t time2stamp(const char* s) {
  struct tm t;
  sscanf(s,"%u-%u-%u %u:%u:%u",&t.tm_year,&t.tm_mon,&t.tm_mday,&t.tm_hour,&t.tm_min,&t.tm_sec);
  t.tm_year-=1900; t.tm_mon--;
  time_t tl=mktime(&t);
  struct tm t2_;
  struct tm *t2;
  t2=gmtime_r(&tl,&t2_);
  time_t tl2=mktime(t2);
  tl=tl-(tl2-tl);
  return tl;
}

static string stamp2time(time_t t) {
  struct tm t2_;
  struct tm *t2;
  if(t==0) return string("EPOCH");
  if((t2=gmtime_r(&t,&t2_)) == NULL) return string("---"); 
  string s=inttostring(t2->tm_year + 1900)+"-"+
     inttostring(t2->tm_mon+1)+"-"+
     inttostring(t2->tm_mday)+" "+
     inttostring(t2->tm_hour)+":"+
     inttostring(t2->tm_min)+":"+
     inttostring(t2->tm_sec);
  return s;
} 

static void remove_hex(char* s) {
  char* p = s;
  char c[5];
  unsigned int i;
  c[0]='0'; c[1]='x'; c[4]=0;
  for(;*p;) {
    if((*p) == '%') {
      p++; if((*p) == 0) { (*s)='%'; s++; break; };
      c[2]=(*p); p++; if((*p) == 0) { (*s)='%'; s++; (*s)=c[2]; s++; break; };
      c[3]=(*p);
      if(sscanf(c,"%x",&i) != 1) {
        (*s)='%'; s++; (*s)=c[2]; s++; (*s)=c[3]; s++;
        p++;
        continue;
      };
      (*s)=(char)i; s++; p++; continue;
    };
    (*s)=(*p); if((*s) == '+') (*s)=' '; s++; p++;
  };
  (*s)=0;
};

static inline void as(string &s,const char* p) { if(p) s+=p; };

/* MySQL helpers */
static int mysql_field_num(const MYSQL_FIELD* fields,unsigned int num_fields,const char* name) {
  unsigned int num = 0;
  for(;num<num_fields;num++) {
    if(strcmp(fields[num].name,name)==0) break;
  };
  if(num>=num_fields) return -1;
  return num;
}

static int mysql_get_datetime(MYSQL_ROW sql_row,int n,time_t* t) {
  char* value;
  if(n>=0) {
    value=sql_row[n];
    if(value) {
      (*t)=time2stamp(value);
      return 0;
    };
  };
  (*t)=0;
  return 1;
}

static int mysql_get_string(const MYSQL_ROW sql_row,int n,char** s,struct soap* sp = NULL) {
  char* value;
  if(n>=0) {
    value=sql_row[n];
    if(value) {
      if(sp) {
        (*s)=soap_strdup_l(sp,value);
      } else {
        (*s)=strdup(value);
      };
      return 0;
    };
  };
  (*s)=NULL;
  return 1;
}

int mysql_get_int(const MYSQL_ROW sql_row,int n,int* i) {
  char* value;
  if(n>=0) {
    value=sql_row[n];
    if(value) {
      char* e;
      (*i)=strtol(value,&e,0);
      if((*e) == 0) return 0;
    };
  };
  (*i)=0;
  return 1;
}

HTTP_Logger::HTTP_Logger(HTTPS_Connector *c_,bool acl_read_,bool acl_write_,bool acl_query_,const char* user,const char* password):HTTP_ServiceAdv(c_),acl_read(acl_read_),acl_write(acl_write_),acl_query(acl_query_) {
  /* SOAP */
  soap_init();
  sp.namespaces=logger_soap_namespaces;
  sp.user=this;
  /* MySQL */
  mysql_init(&sql);
  if(mysql_real_connect(&sql,"localhost",user,password,"nglogger",0,NULL,0) == NULL) {
    odlog(0)<<"Faled to connect to MySQL server: "<<mysql_error(&sql)<<endl;
  };
}

HTTP_Logger::~HTTP_Logger(void) {
  soap_deinit();
  mysql_close(&sql);
}

#ifdef OLD_GSOAP
#error gSOAP older than 2.5.2 is not supported.
#else
void HTTP_Logger::soap_methods(void) {
  odlog(3)<<"soap_methods: tag: "<<sp.tag<<endl;
  if(!soap_match_tag(&sp,sp.tag,"nl:add")) soap_serve_nl__add(&sp);
  else if(!soap_match_tag(&sp,sp.tag,"nl:get")) soap_serve_nl__get(&sp);
  else sp.error = SOAP_NO_METHOD;
}
#endif

/*
 * HTTP methods
 */
HTTP_Error HTTP_Logger::get(const char* uri,int &keep_alive) {
  // Plain web page is not supported anymore
  return HTTP_NOT_IMPLEMENTED;
}

HTTP_Error HTTP_Logger::put(const char* uri,int &keep_alive) {
  return HTTP_NOT_IMPLEMENTED;
}

HTTP_Error HTTP_Logger::post(const char* uri,int &keep_alive) {
  HTTP_Error r = soap_post(uri,keep_alive);
  return r;
}

/*
 *  SOAP interface functions
 */
int nl__add(struct soap *sp, nl__jobinfo *info, int &r) {
  HTTP_Logger* it = (HTTP_Logger*)(sp->user);
  struct tm t_;
  struct tm *t;

  r=1; // error by default

  if(!(it->acl_write)) {
    odilog(0,it->c->pid)<<"Client has no write access"<<endl;
    return SOAP_OK;
  };
  if(info->start == 0) {
    odilog(0,it->c->pid)<<"Missing needed argument (start time)"<<endl;
    return SOAP_OK;
  };
/*  cluster is identified by subject
  if(info->cluster == NULL) {
    odilog(0,c->pid)<<"Missing needed argument (start time)"<<endl;
    return SOAP_OK;
  };
*/
  if(info->user == NULL) {
    odilog(0,it->c->pid)<<"Missing needed argument (user)"<<endl;
    return SOAP_OK;
  };
  if(info->id == NULL) {
    odilog(0,it->c->pid)<<"Missing needed argument (job id)"<<endl;
    return SOAP_OK;
  };
  if(mysql_real_query(&(it->sql),"LOCK TABLES jobs WRITE",22) != 0) {
    odilog(0,it->c->pid)<<"Failed to lock table: "<<mysql_error(&(it->sql))<<endl;
    return SOAP_OK;
  };
  string tt;
  tt = stamp2time(info->start);
  string query = "UPDATE jobs SET ";
  string set = "start='"+tt+"'";
  if(info->end) {
    if(info->end > 86400) { // workaround for old bug
      set+=" , end='"+stamp2time(info->end)+"'";
    };
  };
//  set+=string(" , cluster='")+sql_string(info->cluster)+"'";
  set+=string(" , cluster='")+sql_string(it->c->identity_subject())+"'";
  set+=string(" , user='")+sql_string(info->user)+"'";
  set+=string(" , id='")+sql_string(info->id)+"'";
  if(info->name) set+=string(" , name='")+sql_string(info->name)+"'";
  if(info->failure) set+=string(" , failure='")+sql_string(info->failure)+"'";
  if(info->lrms) set+=string(" , lrms='")+sql_string(info->lrms)+"'";
  if(info->queue) set+=string(" , queue='")+sql_string(info->queue)+"'";
  if(info->rsl) set+=string(" , rsl='")+sql_string(info->rsl)+"'";
  if(info->ui) set+=string(" , ui='")+sql_string(info->ui)+"'";
//  query+=set+" WHERE cluster='"+info->cluster+
  query+=set+" WHERE cluster='"+sql_string(it->c->identity_subject())+
               "' && user='"+sql_string(info->user)+
               "' && id='"+sql_string(info->id)+
               "' && start='"+tt+"'";
  if(mysql_real_query(&(it->sql),query.c_str(),query.length()) != 0) {
    odilog(0,it->c->pid)<<"Failed to query MySQL server: "<<query<<endl;
    odilog(0,it->c->pid)<<"MySQL reported error: "<<mysql_error(&(it->sql))<<endl;
    return SOAP_OK;
  };
  const char* sql_info = mysql_info(&(it->sql));
  if(sql_info==NULL) {
    odilog(0,it->c->pid)<<"Query did not return an information"<<endl;
    return SOAP_OK;
  };
  unsigned long long int matched, changed, warnings;
  if(sscanf(sql_info,"Rows matched: %Lu Changed: %Lu Warnings: %Lu",
                                          &matched,&changed,&warnings) != 3) {
    odilog(0,it->c->pid)<<"Query did not return an interpreatable information: "<<info<<endl;
    return SOAP_OK;
  };
  if(matched == 0) {
    query = "INSERT jobs SET "+set;
    if(mysql_real_query(&(it->sql),query.c_str(),query.length()) != 0) {
      odilog(0,it->c->pid)<<"Failed to query MySQL server: "<<query<<endl;
      odilog(0,it->c->pid)<<"MySQL reported error: "<<mysql_error(&(it->sql))<<endl;
      return SOAP_OK;
    };
  };
  if(mysql_real_query(&(it->sql),"UNLOCK TABLES",13) != 0) {
    odilog(0,it->c->pid)<<"Warning: Failed to unlock table: "<<mysql_error(&(it->sql))<<endl;
  };
  odilog(1,it->c->pid)<<"Logger: stored data from "<<it->c->identity_subject()<<" about "<<info->id<<endl;
  r=0; // passed
  return SOAP_OK;
}

int nl__get(struct soap* sp, char *query, ULONG64 offset, array_jobinfo &r) {
  HTTP_Logger* it = (HTTP_Logger*)(sp->user);

  r.job=NULL; r.__size_job=0; // kind of error. TODO - report errors properly 

  if((!(it->acl_read)) && (!(it->acl_query))) return SOAP_OK;

  string q("SELECT * FROM jobs");
  if((query != NULL) && ((*query) != 0)) {
    int l = strlen(query);
    char* query_ = (char*)malloc(2*l+1);
    if(query_ == NULL) return SOAP_OK;
    query_[0]=0;
    mysql_escape_string(query_,query,l);
    q+=" WHERE "; q+=query_;
    if(!(it->acl_query)) {
      q+=" AND user='"; q+=sql_string(it->c->identity_subject()); q+="'";
    };
    free(query_);
  } else {
    if(!(it->acl_query)) {
      q+=" WHERE user='"; q+=sql_string(it->c->identity_subject()); q+="'";
    };
  };
  q+=" ORDER BY start LIMIT "; q+=inttostring(offset)+",100";
  if(mysql_real_query(&(it->sql),q.c_str(),q.length()) != 0) {
    odilog(0,it->c->pid)<<"Failed to query MySQL server with: "<<q<<endl;
    odilog(0,it->c->pid)<<"MySQL error: "<<mysql_error(&(it->sql))<<endl;
    return SOAP_OK;
  };
  MYSQL_RES* sql_res;
  if((sql_res=mysql_use_result(&(it->sql))) == NULL) {
    odilog(0,it->c->pid)<<"Failed to initiate retrieving results from MySQL server: "<<mysql_error(&(it->sql))<<endl;
    return SOAP_OK;
  }; 
  unsigned int num_fields = mysql_num_fields(sql_res);
  MYSQL_FIELD* fields = mysql_fetch_fields(sql_res);
  unsigned int start_num = mysql_field_num(fields,num_fields,"start");
  unsigned int end_num = mysql_field_num(fields,num_fields,"end");
  unsigned int cluster_num = mysql_field_num(fields,num_fields,"cluster");
  unsigned int id_num = mysql_field_num(fields,num_fields,"id");
  unsigned int user_num = mysql_field_num(fields,num_fields,"user");
  unsigned int name_num = mysql_field_num(fields,num_fields,"name");
  unsigned int failure_num = mysql_field_num(fields,num_fields,"failure");
  unsigned int lrms_num = mysql_field_num(fields,num_fields,"lrms");
  unsigned int queue_num = mysql_field_num(fields,num_fields,"queue");
  unsigned int rsl_num = mysql_field_num(fields,num_fields,"rsl");
  unsigned int ui_num = mysql_field_num(fields,num_fields,"ui");
  unsigned int usedcpu_num = mysql_field_num(fields,num_fields,"usedcpu");
  unsigned int usedmem_num = mysql_field_num(fields,num_fields,"usedmem");
  int count = 0;
  // r.job = new nl__jobinfo[100]; // limit is always 100
  // It is not right to use malloc instead of new, but this class 
  // has empty constructor anyway.
  r.job = (nl__jobinfo*)soap_malloc(sp,100*sizeof(nl__jobinfo));
  if(r.job == NULL) {
    mysql_free_result(sql_res);
    return SOAP_OK;
  };
  for(;;) {
    if(count >= 100) break;
    MYSQL_ROW sql_row;
    if((sql_row=mysql_fetch_row(sql_res)) == NULL) break;
    mysql_get_datetime(sql_row,start_num,&(r.job[count].start));
    mysql_get_datetime(sql_row,end_num,&(r.job[count].end));
    mysql_get_string(sql_row,cluster_num,&(r.job[count].cluster),sp);
    mysql_get_string(sql_row,user_num,&(r.job[count].user),sp);
    mysql_get_string(sql_row,id_num,&(r.job[count].id),sp);
    mysql_get_string(sql_row,name_num,&(r.job[count].name),sp);
    mysql_get_string(sql_row,failure_num,&(r.job[count].failure),sp);
    mysql_get_string(sql_row,lrms_num,&(r.job[count].lrms),sp);
    mysql_get_string(sql_row,queue_num,&(r.job[count].queue),sp);
    mysql_get_string(sql_row,rsl_num,&(r.job[count].rsl),sp);
    mysql_get_string(sql_row,ui_num,&(r.job[count].ui),sp);
    mysql_get_int(sql_row,usedcpu_num,&(r.job[count].usedcpu));
    mysql_get_int(sql_row,usedmem_num,&(r.job[count].usedmem));
    count++;
  };
  r.__size_job=count;
  mysql_free_result(sql_res);
  return SOAP_OK;
}

