#include "../../std.h"
#include <arpa/inet.h>

#include <globus_common.h>

#include "../../misc/log_time.h"
#include "../../misc/inttostring.h"
#include "../../misc/globus_error_utils.h"
#include "../../config/environment.h"

#include "../httpsd.h"

#if GLOBUS_IO_VERSION<5
extern "C" {
extern void globus_i_io_handle_destroy(globus_io_handle_t * handle);
};
#else
void globus_i_io_handle_destroy(globus_io_handle_t * handle) {
};
#endif


static int curr_connections = 0;
static int pre_connections = 2;
static int max_connections = 100;

extern unsigned int conections_processed;

static char* service_not_found_str = "Service not found\r\n";

list<HTTPS_Connector*> HTTPS_Connector::cons;
pthread_mutex_t HTTPS_Connector::cons_lock = PTHREAD_MUTEX_INITIALIZER;

void HTTPS_Connector::loop_in_thread(void) {
  globus_thread_t handle;
  int r;
  r=globus_thread_create(GLOBUS_NULL,GLOBUS_NULL,&loop_thread,(void*)this);
  if(r != GLOBUS_SUCCESS) {
    odilog(0,pid)<<"Failed to create thread"<<endl;
    delete this;
    conections_processed++;
  };
}

void* HTTPS_Connector::loop_thread(void* arg) {
  globus_thread_yield();
  HTTPS_Connector* it = (HTTPS_Connector*)arg;
  it->loop();
  delete it;
  conections_processed++;
  return NULL;
}

void HTTPS_Connector::loop(void) {
  char buf[1024];
  odilog(0,pid)<<"Session started"<<endl;
  {
    char abuf[1024]; abuf[sizeof(abuf)-1]=0;
    int host[4] = {0,0,0,0};
    unsigned short port = 0;
    if(globus_io_tcp_get_local_address(&h,host,&port) == GLOBUS_SUCCESS) {
      struct hostent he;
      struct hostent* he_p;
      snprintf(abuf,sizeof(abuf)-1,"%u.%u.%u.%u",
                              host[0],host[1],host[2],host[3]);
      struct in_addr a;
      if(inet_aton(abuf,&a) != 0) {
        int h_errnop;
        he_p=globus_libc_gethostbyaddr_r((char*)&a,strlen(abuf),AF_INET,
                                             &he,buf,sizeof(buf),&h_errnop);
        if(he_p) {
          if(strcmp(he_p->h_name,"localhost") == 0) {
            abuf[sizeof(abuf)-1]=0;
            if(globus_libc_gethostname(abuf,sizeof(abuf)-1) != 0) {
              strcpy(abuf,"localhost");
            };
          };
        };
      };
    };
    odilog(1,pid)<<"My name is: "<<abuf<<" ("<<host[0]<<"."<<host[1]<<"."<<host[2]<<"."<<host[3]<<":"<<port<<")"<<endl;
    abuf[0]=0;
    if(globus_io_tcp_get_remote_address(&h,host,&port) == GLOBUS_SUCCESS) {
      snprintf(abuf,sizeof(abuf)-1,"%u.%u.%u.%u:%u",
                     host[0],host[1],host[2],host[3],(unsigned int)port);
    };
    odilog(1,pid)<<"Connection from: "<<abuf<<endl;
  };
  int keep_alive = 1;
  for(;;) {
    /* read first line of request header */
    odilog(2,pid)<<"Waiting for request"<<endl;
    size_t l = readline(buf,sizeof(buf)-1);
    if(l == 0) break;
    odilog(2,pid)<<"Request: "<<buf<<endl;
    // parse request
    char* http_URI;
    char* http_version;
    char* http_method = buf;
    char* t = buf;
    t=strchr(t,' ');
    if(t == NULL) {
      odilog(0,pid)<<"Broken HTTP request (no URI found) - "<<buf<<endl; break;
    };
    (*t)=0; t++; for(;(*t)==' ';t++);
    http_URI=t;
    t=strchr(t,' ');
    if(t == NULL) {
      odilog(0,pid)<<"Broken HTTP request (no version found) - "<<http_URI<<endl; break;
    };
    (*t)=0; t++; for(;(*t)==' ';t++);
    if(strncmp("HTTP/",t,5) != 0) {
      odilog(0,pid)<<"Unknown protocol: "<<t<<endl; break;
    };
    http_version=t+5;
    if(keep_alive) {
      if(strcmp(http_version,"1.0") == 0) {
        keep_alive=1;
      } else {
        keep_alive=2;
      };
    };
    HTTP_Error err = HTTP_OK;
    // adjust URI
    string http_url = http_URI;
    if(strncasecmp("https://",http_url.c_str(),8) &&
       strncasecmp("httpg://",http_url.c_str(),8) &&
       strncasecmp("http://",http_url.c_str(),7)) {
      if(http_url[0] != '/') http_url="/"+http_url;
      http_url=base_url+http_url;
    };
    if(http_url[http_url.length()-1] == '/') http_url.resize(http_url.length()-1);
    odilog(2,pid)<<"Request on "<<http_url<<endl;
    // make sure request is for proper url
    if(strncasecmp(http_url.c_str(),base_url.c_str(),base_url.length()) != 0) {
      odilog(0,pid)<<"Improper URL in request: "<<http_url<<" != "<<base_url<<endl;
      err=HTTP_NOT_FOUND;
    } else {
      // obtain service to use 
      HTTP_Service* service = services.get(*this,http_url.c_str());
      if(service == NULL) {
        odilog(0,pid)<<"Corresponding service not found: "<<http_url<<endl;
        err=HTTP_NOT_FOUND;
      } else {
        if(strcasecmp(http_method,"GET") == 0) {
          odilog(2,pid)<<"Calling GET"<<endl;
          err=service->get(http_url.c_str(),keep_alive);
        } else if(strcasecmp(http_method,"PUT") == 0) {
          odilog(2,pid)<<"Calling PUT"<<endl;
          err=service->put(http_url.c_str(),keep_alive);
        } else if(strcasecmp(http_method,"POST") == 0) {
          odilog(2,pid)<<"Calling POST"<<endl;
          err=service->post(http_url.c_str(),keep_alive);
        } else {
          odilog(2,pid)<<"Unknown method: "<<http_method<<endl;
          err=HTTP_NOT_IMPLEMENTED;
        };
      };
    };
    if((err==HTTP_NOT_IMPLEMENTED) || (err==HTTP_NOT_FOUND)) {
      odilog(1,pid)<<"Method is not implemented in requested service"<<endl;
      if(skip_request(*this,keep_alive) != 0) break;
    };
    if((err != HTTP_OK) && (err != HTTP_FAILURE)) { /* send error response */
      odilog(1,pid)<<"Sending error response: "<<err<<endl;
      if(send_error_response(*this,keep_alive,err,NULL,NULL) != 0) {
        keep_alive=0; break;
      };
    };
    if(keep_alive == 0) {
      odilog(2,pid)<<"Can't keep connection alive"<<endl;
    } else {
      odilog(2,pid)<<"Keeping connection alive"<<endl;
    };
    if(keep_alive == 0) break;
  };
  odilog(1,pid)<<"Closing connection"<<endl;
  // close connection
  services.close(*this);
}

void HTTPS_Connector::accept_cb(void *callback_arg,globus_io_handle_t *handle,globus_result_t result) {
  HTTPS_Connector* it = (HTTPS_Connector*)callback_arg;
  if(result != GLOBUS_SUCCESS) {
    char abuf[1024]; abuf[0]=0;
    int host[4] = {0,0,0,0};
    unsigned short port = 0;
    if(globus_io_tcp_get_remote_address(handle,host,&port) == GLOBUS_SUCCESS) {
      snprintf(abuf,sizeof(abuf)-1,"%u.%u.%u.%u:%u",
                     host[0],host[1],host[2],host[3],(unsigned int)port);
    };
    odilog(0,it->pid)<<"Connection from: "<<abuf<<endl;
    odilog(0,it->pid)<<"Failed to accept connection: "<<GlobusResult(result)<<endl;
    delete it;
    conections_processed++;
    return;
  };
  odilog(1,it->pid)<<"Connection accepted"<<endl;
//  globus_thread_blocking_will_block();  // instead of creating new thread
//  it->loop();
//  delete it;
  it->loop_in_thread();
}

HTTPS_Connector::HTTPS_Connector(globus_io_handle_t *s,const char* url,HTTP_Services& serv,list<AuthEvaluator*>& auths_):services(serv),auths(auths_) {
  initialized=false;
  if(url) base_url=url;
  r_start = 0;
  r_end = 0;

  // add entry to global list
  pthread_mutex_lock(&cons_lock);
  cons.push_back(this);
  pthread_mutex_unlock(&cons_lock);

  globus_result_t res;
  memset(&h,0,sizeof(h));
  res=globus_io_tcp_register_accept(s,GLOBUS_NULL,&h,&accept_cb,this);
  if(res != GLOBUS_SUCCESS) {
    odilog(0,pid)<<"Failed to accept connection: "<<GlobusResult(res)<<endl;
    delete this;
    conections_processed++;
    return;
  };
  initialized=true;
  return;
}

HTTPS_Connector::~HTTPS_Connector(void) {
  // remove own entry from global list
  pthread_mutex_lock(&cons_lock);
  bool in_list = false;
  for(list<HTTPS_Connector*>::iterator i = cons.begin();i!=cons.end();++i) {
    if((*i)==this) { cons.erase(i); in_list=true; break; };
  };
  pthread_mutex_unlock(&cons_lock);
  if(in_list) {
    if(initialized) globus_io_close(&h);
  };
  globus_i_io_handle_destroy(&h);
}

void HTTPS_Connector::identity(globus_io_handle_t* handle,const char* subject) {
  // modify entry in global list to reflect identity of client
#if GLOBUS_IO_VERSION<5
  gss_cred_id_t cred = handle->delegated_credential;
#else
  gss_cred_id_t cred;
  if(globus_io_tcp_get_delegated_credential(handle,&cred) != GLOBUS_SUCCESS) {
    cred=GSS_C_NO_CREDENTIAL;
  };
#endif
  pthread_mutex_lock(&cons_lock);
  for(list<HTTPS_Connector*>::iterator i = cons.begin();i!=cons.end();++i) {
    if(&((*i)->h) == handle) {
      (*i)->user.set(subject,cred);
      odilog(0,(*i)->pid)<<"Subject is "<<(*i)->user.DN()<<endl;
      if((*i)->user.proxy()[0]) {
        odilog(0,(*i)->pid)<<"Delegated credentials in "<<(*i)->user.proxy()<<endl;
      } else {
        odilog(0,(*i)->pid)<<"No delegated credentials"<<endl;
      };
      break;
    };
  };
  pthread_mutex_unlock(&cons_lock);
}

// Hide EOF error, this is not an error. 
#define print_globus_result_not_eof(o,prefix,res) {                 \
  globus_object_t* err = globus_error_get(res);                     \
  if(err == NULL) { o<<prefix<<"<undefined>"<<endl; }               \
  else {                                                            \
    char* tmp = globus_object_printable_to_string(err);             \
    if(tmp == NULL) {                                               \
      o<<prefix<<"<undefined>"<<endl; globus_object_free(err);      \
    } else {                                                        \
      if(strstr(tmp,"end-of-file") == NULL) o<<prefix<<tmp<<endl;   \
      free(tmp);                                                    \
    };                                                              \
    globus_object_free(err);                                        \
  };                                                                \
}

size_t HTTPS_Connector::readline(char* buf, size_t l) {
  char c = 0;
  unsigned int ll = 0;
  for(;;) {
    if(ll > 65536) return 0; // must be veeeeery strange line
    if(r_start>=r_end) {
      globus_size_t ll_;
      globus_result_t res;  
      if((res=globus_io_read(&h,(globus_byte_t*)r_buf,sizeof(r_buf),1,&ll_))
                                                          !=GLOBUS_SUCCESS) {
        print_globus_result_not_eof(odilog(0,pid),"Failed to read from network: ",res);
        return 0; // unexpected eof
      };
      if(ll_ == 0) return 0; // unexpected eof
      r_start=0; r_end=ll_;
    };
    char c = r_buf[r_start];
//cerr<<"readline: char: "<<c<<endl;
    if(ll<l) buf[ll]=c;
    ++r_start; ++ll;
    if(c == 0x0a) break; // eol
  };
//cerr<<"readline: eol at "<<ll<<endl;
  // eol found
  if(ll > l) ll=l;
  if(ll == 0) return 0;
  l=ll;
  for(;l>0;) {
    l--;
    if((buf[l] != 0x0a) && (buf[l] != 0x0d)) break;
    buf[l]=0;
  };
//buf[ll]=0;
//cerr<<"readline: line: "<<buf<<endl;
  return ll;
}

size_t HTTPS_Connector::read(char* buf, size_t l) {
  globus_result_t res;
  globus_size_t ll = 0;
  if(r_start<r_end) {
    ll=r_end-r_start;
    if(ll<l) { memcpy(buf,r_buf+r_start,ll); r_end=0; r_start=0; }
    else { memcpy(buf,r_buf+r_start,l); r_start+=l; return l; };
    l-=ll; buf+=ll;
  };
  globus_size_t ll_;
  if(l > sizeof(r_buf)) {
    if((res=globus_io_read(&h,(globus_byte_t*)buf,l,ll?0:1,&ll_))
                                                          !=GLOBUS_SUCCESS) {
      if(ll) return ll;
      print_globus_result_not_eof(odilog(0,pid),"Failed to read from network: ",res);
      return 0;
    };
    return (ll+ll_);
  };
  if((res=globus_io_read(&h,(globus_byte_t*)r_buf,sizeof(r_buf),ll?0:1,&ll_))
                                                          !=GLOBUS_SUCCESS) {
    if(ll) return ll;
    print_globus_result_not_eof(odilog(0,pid),"Failed to read from network: ",res);
    return 0;
  };
  r_start=0; r_end=ll_; if(ll_>l) ll_=l;
  memcpy(buf,r_buf,ll_); r_start+=ll_;
  return (ll+ll_);
}

int HTTPS_Connector::write(const char* buf, size_t l) {
  size_t n = 0;
  for(;n<l;) {
    globus_size_t ll = 0;
    globus_result_t res;
    if((res=globus_io_write(&h,(globus_byte_t*)buf+n,l-n,&ll)) != GLOBUS_SUCCESS) {
      odilog(0,pid)<<"Failed to write to network: "<<GlobusResult(res)<<endl;
      return -1;
    };
    n+=ll;
  };
  return 0;
}

int skip_header(HTTPS_Connector &c,int &keep_alive) {
/* TODO - control keep_alive */
  char buf[1024], *s;
  for (;;) {
    if(c.readline(buf,1023) == 0) return 1; /* connection broken */
    if(buf[0] == 0) break; /* empty line */
  };
  return 0;
}

int skip_request(HTTPS_Connector &c,int &keep_alive) {
  bool length_passed=false;
  unsigned long long int length;
  char buf[1024];
  // skip header
  for (;;) {
    if(c.readline(buf,1023) == 0) return HTTP_ERROR; /* connection broken */
    if(buf[0] == 0) break; /* empty line - end of header */
    char* token = buf; for(;*token;token++) if(isspace(*token)) break;
    int l = token-buf;
    for(;*token;token++) if(!isspace(*token)) break;
    if(strncasecmp("Connection:",buf,l) == 0) {
      if(strcasecmp("close",token) == 0) { keep_alive=0; return HTTP_OK; }
      else if(strcasecmp("keep-alive",token) == 0) { keep_alive=2; };
    } else if(strncasecmp("Content-Length:",buf,l) == 0) {
      char *e;
      length=strtoull(token,&e,10);
      if((*e) == 0) length_passed=true;
      if(length > 1024) { keep_alive=0; return HTTP_OK; };        
    };
  };
  // skip body
  if(!length_passed) { keep_alive=0; return HTTP_OK; };
  for (;;) {
    size_t l = 1024;
    if(l>length) l=length;
    l=c.read(buf,0);
    if(l==0) { keep_alive=0; return HTTP_OK; };
    length-=l; 
    if(length == 0) break;
  };
  return HTTP_OK;
}

int send_error_response(HTTPS_Connector &c,int keep_alive,int code,char* type,char* content) {
  unsigned long long int size = 0;
  if(content) {
    size=strlen(content);
    if(send_response_header(c,keep_alive,code,type,size) != 0) return -1;
    return c.write(content,strlen(content));
  } else {
    string fname = nordugrid_loc+"/share/error"+inttostring(code)+".html";
    stat_file(fname.c_str(),size);
    if(send_response_header(c,keep_alive,code,"text/html",size) != 0) return -1;
    if(size) return send_file(fname.c_str(),c);
  };
  return 0;
}

int send_response_header(HTTPS_Connector &c,int keep_alive,int code,char* type,int size) {
  string header = "HTTP/1.1 "+inttostring(code);
  if(code == 200) { header+=" OK\r\n"; } else { header+=" NOREASON\r\n"; };
  if(keep_alive && (size>=0)) {
    header+="Connection: Keep-Alive\r\n";
  } else {
    header+="Connection: close\r\n";
  };
  if(size>=0) header+="Content-Length: "+inttostring(size)+"\r\n";
  if(type) { header+="Content-Type: "; header+=type; header+="\r\n"; };
  header+="\r\n";
  olog<<"Sending response:\n"<<header<<endl;
  return c.write(header.c_str(),header.length());
}

int stat_file(const char* fname,unsigned long long int &size) {
  struct stat64 st;
  if(stat64(fname,&st) != 0) return -1;
  if(!S_ISREG(st.st_mode)) return -1;
  size=st.st_size;
  return 0;
}

int send_file(const char* fname,HTTPS_Connector &c) {
  int h = ::open(fname,O_RDONLY);
  if(h == -1) return 1;
  char buf[1024];
  for(;;) {
    int l = read(h,buf,sizeof(buf));
    if(l == 0) break;
    if(l == -1) { close(h); return 1; };
    if(c.write(buf,l)) { close(h); return 1; };
  };
  close(h);
  return 0;
}

