#ifndef __URL_OPTIONS_H__
#define __URL_OPTIONS_H__

#include <string>
#include <iostream>

#define PORT_DEFAULT_RC      389
#define PORT_DEFAULT_RLS     39281
#define PORT_DEFAULT_LDAP    389
#define PORT_DEFAULT_HTTP    80
#define PORT_DEFAULT_HTTPS   443
#define PORT_DEFAULT_HTTPG   8000
#define PORT_DEFAULT_FTP     21
#define PORT_DEFAULT_GSIFTP  2811

int remove_url_options(string& url);
int get_url_option(const string& url,const char* name,int num,string& value);
int add_url_option(string& url,const char* name,const char* value,int num);
int add_url_option(string& url,const string& option,int num,const char* name = NULL);
int del_url_option(string& url,const char* name,int num);
int get_url_options(const char* host,string& options);
int add_url_options(string& url,const char* options,int num);
int canonic_url(string &url);
const char* get_url_path(const char* url);
string get_url_host(const char* url);

class URL {
 friend ostream& operator<<(ostream& o,const URL& u);
 private:
  int port;
  string host;
  string proto;
  string path;
  bool valid;
  static string empty;
 public:
  URL(const char*);
  ~URL(void) { };
  int Port() const { if(!valid) return 0; return port; };
  const string& Proto() { if(!valid) return empty; return proto; };
  const string& Host() { if(!valid) return empty; return host; };
  const string& Path() { if(!valid) return empty; return path; };
  operator bool(void) { return valid; };
  void operator =(const URL& u) { 
    proto=u.proto; port=u.port; host=u.host; path=u.path; valid=u.valid;
  };
};

ostream& operator<<(ostream& o,const URL& u);
 
#endif // __URL_OPTIONS_H__
