#include "std.h"
#include "datamove/datapoint.h"
#include "datamove/datamove.h"
/* 
  Do not include file_soap.nsmap directly here. It is already included 
  in datahandle_httpg. Hence this will cause multiple definition of
  namespaces. This problem should get resolved after switching to 
  gSOAP 2.5.x
#include "https/se/file_soap.nsmap"
*/
#include "https/se/file_soapH.h"
extern SOAP_NMAC struct Namespace file_soap_namespaces[];

#include "https/client/client.h"
#include "misc/globus_modules.h"
#include "misc/log_time.h"

extern "C" {
int ngacl(int argc,char** argv) {
  LogTime::active=false;
  bool verbose = false;
  for(;;) {
    opterr=0;
    int optc=getopt(argc,argv,"+hvd:c");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"ngacl [-h] [-v] [-d level] get|set url"<<endl;
        olog<<" -h - print help"<<endl;
        olog<<" -v - print version"<<endl;
        olog<<" -d - set debug level"<<endl;
        return 1;
      }; break;
      case 'v': {
        olog<<"ngacl: version "<<VERSION<<endl;
        return 0;
      };
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          return 1;
        };
        verbose=true;
        LogTime::level=i;
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        return 1;
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        return 1;
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        return 1;
      };
    };
  };
  char* command = argv[optind+0];
  if(command == NULL) {
    olog<<"Missing action" << endl; return 1;
  };
  char* file_url = argv[optind+1];
  if(file_url == NULL) {
    olog<<"Missing url" << endl; return 1;
  };
  if(argv[optind+2] != NULL) {
    olog<<"Unrecognized option: "<<argv[optind+1]<<endl; return 1;
  };
  if(strncmp(file_url,"gsiftp://",9) == 0) {
    string gacl_url(file_url);
    int n = gacl_url.rfind('/');
    if(n < 9) {
      olog<<"URL contains no path"<<endl; return 1;
    };
    if(n == (gacl_url.length()-1)) {
      gacl_url.insert(n+1,".gacl");
    } else {
      gacl_url.insert(n+1,".gacl-");
    };
    DataPoint url(gacl_url.c_str());
    if(!url) {
      olog<<"Unsupported URL given"<<endl; return 1;
    };
    char* tmp_name = strdup("/tmp/ngacl.XXXXXX");
    int tmp_h = mkstemp(tmp_name);
    if(tmp_h == -1) {
      olog<<"Failed to create temporary file"<<endl; return 1;
    };
    fchmod(tmp_h,S_IRUSR | S_IWUSR);
    close(tmp_h);
    char* tmp_url = (char*)malloc(7+strlen(tmp_name)+1);
    if(tmp_url == NULL) {
      unlink(tmp_name); free(tmp_name);
      olog<<"Failed to allocate memory"<<endl; return 1;
    };
    strcpy(tmp_url,"file://"); strcat(tmp_url,tmp_name);
    DataPoint tmp(tmp_url); free(tmp_url);
    DataMove mover;
    mover.secure(false);
    mover.passive(true);
    mover.verbose(verbose);
    DataCache cache;
    if(!strcmp(command,"get")) {
      if(mover.Transfer(url,tmp,cache,UrlMap()) != DataMove::success) {
        olog<<"ACL transfer FAILED."<<endl;
        unlink(tmp_name); free(tmp_name);
        return 1;
      };
      tmp_h=::open(tmp_name,O_RDONLY);
      if(tmp_h == -1) {
        unlink(tmp_name); free(tmp_name); 
        olog<<"Failed to open temporary file"<<endl; return 1;
      };
      char buf[256];
      for(;;) {
        int l = ::read(tmp_h,buf,sizeof(buf)-1);
        if(l==0) break;
        if(l==-1) {
          close(tmp_h); unlink(tmp_name); free(tmp_name); 
          olog<<"Failed to read temporary file"<<endl; return 1;
        };
        buf[l]=0;
        cout<<buf;
      };
      close(tmp_h);
    } else if(!strcmp(command,"set")) {
      tmp_h=::open(tmp_name,O_WRONLY);
      if(tmp_h == -1) {
        unlink(tmp_name); free(tmp_name);
        olog<<"Failed to open temporary file"<<endl; return 1;
      };
      char buf[256];
      for(;;) {
        int l = ::read(0,buf,sizeof(buf));
        if(l==0) break;
        if(l==-1) {
          close(tmp_h); unlink(tmp_name); free(tmp_name);
          olog<<"Failed to read stdin"<<endl; return 1;
        };
        char* p = buf;
        for(;l>0;) {
          int ll=::write(tmp_h,p,l);
          if(ll==-1) {
            close(tmp_h); unlink(tmp_name); free(tmp_name);
            olog<<"Failed to write temporary file"<<endl; return 1;
          };
          l-=ll; p+=ll;
        };
      };
      close(tmp_h);
      if(mover.Transfer(tmp,url,cache,UrlMap()) != DataMove::success) {
        olog<<"ACL transfer FAILED."<<endl;
        unlink(tmp_name); free(tmp_name);
        return 1;
      };
    } else {
      unlink(tmp_name); free(tmp_name);
      olog<<"Only get and set commands are supported"<<endl; return 1;
    };
    unlink(tmp_name);
    if(verbose) olog<<"Operation succeeded"<<endl;
    return 0;
  } else if(strncmp(file_url,"se://",5) == 0) {
    GlobusModuleIO io_mod;
    string service_url = file_url;
    int n;
    n=service_url.find('?');
    if(n != string::npos) {
      service_url[n]='/';
    };
    service_url.replace(0,2,"httpg");
    odlog(1)<<"Talking to SOAP service at "<<service_url<<endl;
    struct soap soap;
    HTTP_ClientSOAP s(service_url.c_str(),&soap);
    soap.namespaces=file_soap_namespaces;
    string soap_url = service_url;
    n=soap_url.find(':');
    // make gSOAP create http headers (should be removed in a future)
    if(n != string::npos) soap_url.replace(0,n,"http");
    if(!strcmp(command,"get")) {
      if(s.connect() != 0) {
        odlog(0)<<"Failed to connect to "<<service_url<<endl;
        return 1;
      };
      ns__aclResponse rr;
      int soap_err = SOAP_OK;
      if((soap_err=soap_call_ns__acl(&soap,soap_url.c_str(),"acl",NULL,rr))
                                                             != SOAP_OK) {
        odlog(1)<<"Failed to execute remote soap call 'acl' at "<<service_url<<endl;
        return 1;
      };
      if(rr.error_code != 0) {
        odlog(1)<<"Failed ("<<rr.error_code<<") to get acl at "<<service_url<<endl;
        return 1;
      };
      if(rr.acl == NULL) {
        odlog(1)<<"No acl in answer from "<<service_url<<endl;
        return 1;
      };
      cout<<rr.acl;
      return 0;
    } else if(!strcmp(command,"set")) {
      if(s.connect() != 0) {
        odlog(0)<<"Failed to connect to "<<service_url<<endl;
        return 1;
      };
      string acl("");
      char buf[256];
      for(;;) {
        int l = ::read(0,buf,sizeof(buf)-1);
        if(l==0) break;
        if(l==-1) {
          olog<<"Failed to read stdin"<<endl; return 1;
        };
        buf[l]=0; acl+=buf;
      };
      ns__aclResponse rr;
      int soap_err = SOAP_OK;
      if((soap_err=soap_call_ns__acl(&soap,soap_url.c_str(),"acl",
                                     (char*)acl.c_str(),rr)) != SOAP_OK) {
        odlog(1)<<"Failed to execute remote soap call 'acl' at "<<service_url<<endl;
        return 1;
      };
      if(rr.error_code != 0) {
        odlog(1)<<"Failed ("<<rr.error_code<<") to set acl at "<<service_url<<endl;
        return 1;
      };
      return 0;
    } else {
      olog<<"Only get and set commands are supported"<<endl; return 1;
    };
    return 1;
  } else {
    olog<<"Only GridFTP servers with GACL and SSE are supported"<<endl;
  };
  return 1;
}
};

