#include "std.h"
#include "datamove/datacache.h"
#include "datamove/datapoint.h"
#include "datamove/datahandle.h"
#include "datamove/datamove.h"
#include "misc/globus_modules.h"
#include "misc/log_time.h"


void progress(FILE* o,const char* s,unsigned int t,
       unsigned long long int all,unsigned long long int max,
       double instant,double average) {
  static int rs = 0;
  const char rs_[4] = { '|','/','-','\\' };
  if(max) {
    fprintf(o,"\r|");
    unsigned int l = (74*all+37)/max; if(l>74) l=74;
    unsigned int i = 0;
    for(;i<l;i++) fprintf(o,"=");
    fprintf(o,"%c",rs_[rs++]); if(rs>3) rs=0;
    for(;i<74;i++) fprintf(o," ");
    fprintf(o,"|\r"); fflush(o);
    return;
  };
  fprintf(o,"\r%Lu kB                    \r",all/1024);
}

extern "C" {
int ngcopy(int argc,char** argv) {
  LogTime::active=false;
  LogTime::level=0;
  bool verbose = false;
  char* cache_path = NULL;
  char* cache_data_path = NULL;
  char* id = "<ngcopy>";
  bool secure = false;
  bool passive = false;
  bool force = false;
  for(;;) {
    opterr=0;
    int optc=getopt(argc,argv,"+hvpfd:sc:C:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"ngcopy [-h] [-v] [-d level] [-s] [-c cache [-C data]] source_url destination_url"<<endl;
        olog<<" -h - print help"<<endl;
        olog<<" -v - print version"<<endl;
        olog<<" -d - set debug level (use 0 for verbose)"<<endl;
        olog<<" -c - path to local cache (use to put file into cache)"<<endl;
        olog<<" -C - path for cache data (if different from -c)"<<endl;
        olog<<" -s - use secure transfer (insecure by default)"<<endl;
        olog<<" -p - use passive transfer (does not work if secure is on)"<<endl;
        olog<<" -f - force transfer to metaurl even if LFN exists"<<endl;
        return 1;
      }; break;
      case 'v': {
        olog<<"ngcopy: version "<<VERSION<<endl;
        return 0;
      };
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          return 1;
        };
        verbose=true;
        LogTime::level=i;
      }; break;
      case 's': {
        secure=true;
      }; break;
      case 'p': {
        passive=true;
      }; break;
      case 'f': {
        force=true;
      }; break;
      case 'c': {
        cache_path=optarg;
      }; break;
      case 'C': {
        cache_data_path=optarg;
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        return 1;
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        return 1;
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        return 1;
      };
    };
  };
  if(argv[optind+0] == NULL) {
    olog<<"Missing source url" << endl; return 1;
  };
  if(argv[optind+1] == NULL) {
    olog<<"Missing destination url" << endl; return 1;
  };
  if(argv[optind+2] != NULL) {
    olog<<"Unrecognised option: "<<argv[optind+2]<<endl; return 1;
  };
  string source_url = argv[optind+0];
  string destination_url = argv[optind+1];
  if(source_url.length() == 0) {
    olog<<"Empty source url" << endl; return 1;
  };
  if(destination_url.length() == 0) {
    olog<<"Empty destination url" << endl; return 1;
  };
  if(source_url != "-") {
    int p = source_url.find("://");
    if((p == string::npos) || (source_url.find('/') < p)) {
      if(source_url[0] == '/') {
        source_url="file://"+source_url;
      } else {
        char buf[1024]; buf[0]=0;
        getcwd(buf,sizeof(buf));
        source_url=string("file://")+buf+"/"+source_url;
      };
    };
  };
  if(destination_url != "-") {
    int p = destination_url.find("://");
    if((p == string::npos) || (destination_url.find('/') < p)) {
      if(destination_url[0] == '/') {
        destination_url="file://"+destination_url;
      } else {
        char buf[1024]; buf[0]=0;
        getcwd(buf,sizeof(buf));
        destination_url=string("file://")+buf+"/"+destination_url;
      };
    };
  };
  if(destination_url[destination_url.length()-1] == '/') {
    if(source_url[source_url.length()-1] == '/') {
      // olog<<"Fileset copy is not supported yet" << endl; return 1;
      // Find out if source can be listed (TODO - through datapoint)
      if(strncasecmp(source_url.c_str(),"rc://",5) &&
         strncasecmp(source_url.c_str(),"rls://",6) &&
         strncasecmp(source_url.c_str(),"se://",5) &&
         strncasecmp(source_url.c_str(),"gsiftp://",9) &&
         strncasecmp(source_url.c_str(),"ftp://",6)) {
        olog<<"Fileset copy for this kind of source is not supported yet"<<endl;
        return 1;
      };
      DataPoint source(source_url.c_str());
      if(!source) {
        olog<<"Unsupported source url"<<endl; return 1;
      };
      list<DataPoint::FileInfo> files;
      if(source.meta()) {
        if(!source.list_files(files,false)) {
          odlog(1)<<"Failed listing metafiles"<<endl;
          return 1;
        };
      } else {
        DataHandle h(&source);
        if(!h.list_files(files,false)) {
          odlog(1)<<"Failed listing files"<<endl;
          return 1;
        };
      };
      bool failures = false;
      for(list<DataPoint::FileInfo>::iterator i = files.begin();
                                             i!=files.end();++i) {
        if(verbose) olog<<"Name: "<<i->name<<endl;
        string s_url(source_url.c_str());
        string d_url(destination_url.c_str());
        s_url+=i->name; d_url+=i->name;
        odlog(1)<<"Source: "<<s_url<<endl;
        odlog(1)<<"Destination: "<<d_url<<endl;
        DataPoint source(s_url.c_str());
        DataPoint destination(d_url.c_str());
        if(!source) { olog<<"Unsupported source url"<<endl; continue; };
        if(!destination) { olog<<"Unsupported destination url"<<endl; continue; };
        DataMove mover;
        mover.secure(secure);
        mover.passive(passive);
        mover.verbose(verbose);
        mover.force_to_meta(force);
        DataCache cache(cache_path,cache_data_path,NULL,id,getuid(),getgid());
        if(mover.Transfer(source,destination,cache,UrlMap()) 
                                                != DataMove::success) {
          olog<<"Current transfer FAILED."<<endl; failures=true;
        } else {
          if(verbose) cout<<"Current transfer complete."<<endl;
        };
      };
      return failures?1:0;
    } else {
      int p = source_url.rfind('/');
      if(p == string::npos) {
        olog<<"Bad source url" << endl; return 1;
      };
      destination_url+=source_url.c_str()+p+1;
    };
  } else {
    if(source_url[source_url.length()-1] == '/') {
      olog<<"Fileset copy to single object is not supported yet" << endl; return 1;
    };
  };
  DataPoint source(source_url.c_str());
  DataPoint destination(destination_url.c_str());
  if(!source) { olog<<"Unsupported source url"<<endl; return 1; };
  if(!destination) { olog<<"Unsupported destination url"<<endl; return 1; };
  DataMove mover;
  mover.secure(secure);
  mover.passive(passive);
  mover.verbose(verbose);
  mover.force_to_meta(force);
  DataCache cache(cache_path,cache_data_path,NULL,id,getuid(),getgid());
  if(LogTime::level == 0) mover.set_progress_indicator(&progress);
  if(mover.Transfer(source,destination,cache,UrlMap()) != DataMove::success) {
    olog<<"Transfer FAILED."<<endl;
    return 1;
  }
  else {
    if(verbose) cout<<"Transfer complete."<<endl;
    return 0;
  };
}
};

