#include "std.h"
#include "datamove/datapoint.h"
#include "datamove/datahandle.h"
#include "misc/log_time.h"

void print_time(time_t t);

extern "C" {
int ngls(int argc,char** argv) {
  LogTime::active=false;
  bool verbose = false;
  bool show_details=false;
  bool show_urls=false;
  for(;;) {
    opterr=0;
    int optc=getopt(argc,argv,"+hvd:lL");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"ngls [-h] [-v] [-d level] [-l] [-L] url"<<endl;
        olog<<" -h - print help"<<endl;
        olog<<" -v - print version"<<endl;
        olog<<" -d - set debug level"<<endl;
        olog<<" -l - show available file details"<<endl;
        olog<<" -L - show URL(s) from which file can be obtained"<<endl;
        return 1;
      }; break;
      case 'v': {
        olog<<"ngls: version "<<VERSION<<endl;
        return 0;
      };
      case 'l': {
        show_details=true;
      }; break;
      case 'L': {
        show_details=true; show_urls=true;
      }; break;
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          return 1;
        };
        verbose=true;
        LogTime::level=i;
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        return 1;
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        return 1;
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        return 1;
      };
    };
  };
  char* dir_url = argv[optind+0];
  if(dir_url == NULL) {
    olog<<"Missing url" << endl; return 1;
  };
  if(argv[optind+1] != NULL) {
    olog<<"Unrecognized option: "<<argv[optind+1]<<endl; return 1;
  };
  DataPoint url(dir_url);
  if(!url) {
    olog<<"Unsupported url given"<<endl; return 1;
  };
  list<DataPoint::FileInfo> files;
  if(url.meta()) {
    if(!url.list_files(files,show_details)) {
      if(files.size() == 0) {
        odlog(1)<<"Failed listing metafiles"<<endl;
        return 1;
      };
      odlog(1)<<"Warning: Failed listing metafiles but some information is obtained"<<endl;
    };
  } else {
    DataHandle h(&url);
    h.secure(false);
    if(!h.list_files(files,show_details)) {
      if(files.size() == 0) {
        odlog(1)<<"Failed listing files"<<endl;
        return 1;
      };
      odlog(1)<<"Warning: Failed listing files but some information is obtained"<<endl;
    };
  };
  for(list<DataPoint::FileInfo>::iterator i = files.begin();
                                               i!=files.end();++i) {
    cout<<i->name;
    if(show_details) {
      switch(i->type) {
        case DataPoint::FileInfo::file_type_file: cout<<" file"; break;
        case DataPoint::FileInfo::file_type_dir: cout<<" dir"; break;
        default: cout<<" unknown"; break;
      };
      if(i->size_available) { cout<<" "<<i->size; } else { cout<<" *"; };
      if(i->created_available) { print_time(i->created); } else { cout<<" *"; };
      if(i->valid_available) { print_time(i->valid); } else { cout<<" *"; };
      if(i->checksum_available) { cout<<" "<<i->checksum; } else { cout<<" *"; };
    };
    cout<<endl;
    if(show_urls) {
      for(list<string>::iterator u = i->urls.begin();u!=i->urls.end();++u) {
        cout<<"\t"<<*u<<endl;
      };
    };
  };
  return 0;
}

};

void print_time(time_t t) {
  char buf[1024];
  char *t_;
  t_=ctime_r(&t,buf);
  int l = strlen(t_);
  if(l>0) { if(t_[l-1] == '\n') { l--; t_[l]=0; }; };
  if((t_ == NULL) || (l==0)) { cout<<" *"; return; };
  cout<<" \""<<t_<<"\"";
  return;
}


