#include "../std.h"

#include "../rsl/parse_rsl.h"
#include "../transfer/url_map.h"
#include "../transfer/proxy.h"
#include "../datamove/datapoint.h"
#include "../datamove/datahandle.h"
#include "../misc/condition.h"
#include "../misc/log_time.h"

CondSimple cond;

class lfn_t {
 public:
  const char* lfn;
  bool done;
  bool failed;
  lfn_t(const char* l):lfn(l),done(false),failed(false) { };
};

void* check_url(void *arg) {
  lfn_t* lfn = (lfn_t*)arg;
  cerr<<lfn->lfn<<endl;
  DataPoint source(lfn->lfn);
  if(!source) {
    cerr<<"Failed to acquire source: "<<source<<endl;
    lfn->failed=true; lfn->done=true; cond.signal();
    return NULL;
  };
  if(!source.meta_resolve(true,UrlMap())) {
    cerr<<"Failed to resolve "<<source<<endl;
    lfn->failed=true; lfn->done=true; cond.signal();
    return NULL;
  };
  source.tries(1);
  DataHandle source_h(&source);
  // TODO. Run every URL in separate thread.
  // TODO. Do only connection (optionally)
  for(;source.have_location();) {
    if(source_h.check()) break;
    source.next_location();
  };
  if(!source.have_location()) {
    cerr<<"Failed to check "<<source<<endl;
    lfn->failed=true; lfn->done=true; cond.signal();
    return NULL;
  };
  lfn->done=true; cond.signal();
  return NULL;
}


int main(int argc,char* argv[]) {
  if(argc<2) {
    cerr<<"Usage: inputcheck RSL_file [proxy_file]"<<endl;
    return -1;
  };
  string rsl = argv[1];
  const char* proxy = NULL;
  if(argc>=3) proxy=argv[2];
  
  JobLocalDescription job;

  LogTime::level = 2;

  if(!parse_rsl(rsl,job)) return 1;

  if(proxy) {
    setenv("X509_USER_PROXY",proxy,1);
    setenv("X509_USER_CERT",proxy,1);
    setenv("X509_USER_KEY",proxy,1);
  };
  prepare_proxy();
  
  list<FileData>::iterator file;
  bool has_lfns = false;
  list<lfn_t*> lfns;
  for(file=job.inputdata.begin();file!=job.inputdata.end();++file) {
    if(file->has_lfn()) {
      pthread_t thread;
      pthread_attr_t thread_attr;
      lfn_t* lfn = new lfn_t(file->lfn.c_str());
      lfns.push_back(lfn);
      pthread_attr_init(&thread_attr);
      pthread_attr_setdetachstate(&thread_attr,PTHREAD_CREATE_DETACHED);
      pthread_create(&thread,&thread_attr,&check_url,lfn);
      has_lfns=true;
    };
  };
  for(;has_lfns;) {
    cond.wait();
    has_lfns=false;
    for(list<lfn_t*>::iterator l = lfns.begin();l!=lfns.end();++l) {
      if((*l)->done) {
        if((*l)->failed) {
          remove_proxy();
          exit(1);
        };
      } else {
        has_lfns=true; 
      };
    };
  };
  remove_proxy();
  exit(0);
}

