#include "../std.h"

#include "proxy.h"

int prepare_proxy(void) {
  char* proxy_file = NULL;
  char* proxy_file_tmp = NULL;
  struct stat stx;
  int h = -1;
  off_t len;
  char* buf = NULL;
  off_t l,ll;
  int res=-1;

  if(getuid() == 0) { /* create temporary proxy */
    proxy_file=getenv("X509_USER_PROXY");
    if(proxy_file==NULL) goto exit;
    h=open(proxy_file,O_RDONLY);
    if(h==-1) goto exit;
    if((len=lseek(h,0,SEEK_END))==-1) goto exit;
    lseek(h,0,SEEK_SET);
    buf=malloc(len);
    if(buf==NULL) goto exit;
    for(l=0;l<len;) {
      ll=read(h,buf+l,len-l);
      if(ll==-1) goto exit;
      if(ll==0) break;
      l+=ll;
    };
    close(h); h=-1; len=l;
    proxy_file_tmp=malloc(strlen(proxy_file)+5);
    if(proxy_file_tmp==NULL) goto exit;
    strcpy(proxy_file_tmp,proxy_file); strcat(proxy_file_tmp,".tmp");
    h=open(proxy_file_tmp,O_WRONLY | O_CREAT,S_IRUSR | S_IWUSR);
    if(h==-1) goto exit;
    chmod(proxy_file_tmp,S_IRUSR | S_IWUSR);
    for(l=0;l<len;) {
      ll=write(h,buf+l,len-l);
      if(ll==1) goto exit;
      l+=ll;
    };
    close(h); h=-1; 
    setenv("X509_USER_PROXY",proxy_file_tmp,1);
  };
  res=0;
 exit:
  if(proxy_file_tmp) free (proxy_file_tmp);
  if(buf) free(buf);
  if(h!=-1) close(h);
  return res;
}

int remove_proxy(void) {
  char* proxy_file = NULL;
  if(getuid() == 0) {
    proxy_file=getenv("X509_USER_PROXY");
    if(proxy_file == NULL) return 0;
    remove(proxy_file);
  };
  return 0;
}

int renew_proxy(const char* old_proxy,const char* new_proxy) {
  int h = -1;
  off_t len,l,ll;
  char* buf = NULL;
  char* proxy_file_tmp = NULL;
  struct stat st;
  int res = -1;

  h=open(new_proxy,O_RDONLY);
  if(h==-1) goto exit;
  if((len=lseek(h,0,SEEK_END))==-1) goto exit;
  lseek(h,0,SEEK_SET);
  buf=malloc(len);
  if(buf==NULL) goto exit;
  for(l=0;l<len;) {
    ll=read(h,buf+l,len-l);
    if(ll==-1) goto exit;
    if(ll==0) break;
    l+=ll;
  };
  close(h); h=-1; len=l;
  proxy_file_tmp=malloc(strlen(old_proxy)+7);
  if(proxy_file_tmp==NULL) goto exit;
  strcpy(proxy_file_tmp,old_proxy); strcat(proxy_file_tmp,".renew");
  remove(proxy_file_tmp);
  h=open(proxy_file_tmp,O_WRONLY | O_CREAT | O_EXCL,S_IRUSR | S_IWUSR);
  if(h==-1) goto exit;
  chmod(proxy_file_tmp,S_IRUSR | S_IWUSR);
  for(l=0;l<len;) {
    ll=write(h,buf+l,len-l);
    if(ll==1) goto exit;
    l+=ll;
  };
  if(stat(old_proxy,&st) != 0) goto exit;  /* old proxy must exist */
  fchown(h,st.st_uid,st.st_gid);
  close(h); h=-1;
  if(remove(old_proxy) != 0) goto exit; 
  if(rename(proxy_file_tmp,old_proxy) != 0) goto exit;
  res=0;
 exit:
  if(h!=-1) close(h);
  if(buf) free(buf);
  if(proxy_file_tmp) { remove(proxy_file_tmp); free (proxy_file_tmp); };
  return res;
}

char* write_proxy(gss_cred_id_t cred) {
  char* proxy_fname = NULL;
  OM_uint32 major_status = 0;
  OM_uint32 minor_status = 0;
  gss_buffer_desc deleg_proxy_filename;
  major_status = gss_export_cred(&minor_status,
                                 cred,
                                 NULL,
                                 1,
                                 &deleg_proxy_filename);
  if (major_status == GSS_S_COMPLETE) {
    char * cp;
    cp = strchr((char *)deleg_proxy_filename.value, '=');
    if(cp != NULL) {
      cp++;
      proxy_fname=strdup(cp);
    };
    free(deleg_proxy_filename.value);
  };
  return proxy_fname;
}

int proxy_expired(void) {


}
 
