/*
  
   Copyright 2004 Jonas Bardino

   This file is part of NorduGRID GridSSH.

    NorduGRID GridSSH is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    NorduGRID GridSSH is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with NorduGRID GridSSH; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 *
 * Functions for obtaining the home directory of a user
 *
 */

#include "home_dir.h"
#include "log.h"
#include "runtime_conf.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <assert.h>

/* sysconf */
#include <unistd.h>

/* Get homedir with getpw */
#include <pwd.h>

/* Dummy init/clean so far */
int initialize_home_dir(void)
{
  return 0;
}

int finalize_home_dir(void)
{
  return 0;
}

char *get_user_home(char *user, void *uid)
{
  /* Find home dir of supplied user using passwd file */
  char *home = NULL; 
  struct passwd pw_entry, *pw_entry_p = &pw_entry;
  char *buf = NULL;
  long bufsize = -1;

  bufsize = sysconf(_SC_GETPW_R_SIZE_MAX);
  assert(bufsize > 0);
  buf = malloc(bufsize*sizeof(char));
  assert(buf != NULL);
  /* use name for lookup if supplied, otherwise use uid */
  if(user != NULL) {
    debugmsg(6,"find homedir from name %s", user);
    if(getpwnam_r(user, &pw_entry, buf, bufsize, &pw_entry_p) != 0) {
      errmsg("couldn't extract user homedir from name - can't continue");
      free(buf);
      return NULL;
    }
  } else {
    debugmsg(6,"find homedir from uid %d", uid);
    if(getpwuid_r((uid_t)uid, &pw_entry, buf, bufsize, &pw_entry_p) != 0) {
      errmsg("couldn't extract user homedir from uid - can't continue");
      free(buf);
      return NULL;
    }
  }

  debugmsg(5,"found homedir %s for user %s", pw_entry.pw_dir, user);

  home = malloc((strlen(pw_entry.pw_dir)+1)*sizeof(char));
  assert(home != NULL);
  
  debugmsg(9,"copying homedir name");
  sprintf(home, "%s",pw_entry.pw_dir);

  debugmsg(9,"set home to %s",home);

  free(buf);
  pw_entry_p = NULL;

  return home;
}

char *get_home(void)
{
  /* Find home of calling user */
  uid_t uid = -1;

  debugmsg(7,"get uid of calling user");
  uid = getuid();

  debugmsg(5,"got uid %d for calling user", uid);

  return get_user_home(NULL, (void *)uid);
}
