#include "qstring.h"

#if QT_VERSION >= 0x030100

#include "qtextedit.h"
#include <iostream>
#include "xRSLHighlighter.h"

using namespace std;

xRSLHighlighter::xRSLHighlighter( QTextEdit* textEdit ) : QSyntaxHighlighter( textEdit )
{ 
  red.setRgb( 255, 0, 0 );
  blue.setRgb( 0, 0, 255 );
  black.setRgb( 0, 0, 0 );

  string tmp[] = { "executable", "arguments", "inputfiles", "executables", "cache", "outputfiles", "cputime", "memory", "disk", "runtimeenvironment", "middleware", "stdin", "stdout", "stderr", "join", "gmlog", "jobname", "ftpthreads", "cluster", "queue", "startime", "lifetime", "notify", "replicacollection", "rerun", "architecture", "nodeaccess", "dryrun", "rsl_substitution", "count" };
  unsigned int tmpsize = sizeof( tmp )/sizeof( *tmp );

  //                ftpthreads, count, rerun, cputime, memory, disk (numbers)
  //                nodeaccess (inbound, outbound)

  xrslattributes.assign( tmp, tmp + tmpsize );
};

xRSLHighlighter::~xRSLHighlighter( ) 
{ };

int xRSLHighlighter::highlightParagraph( const QString& paragraph, int endStateOfLastPara )
{
  unsigned int begin = 0;
  unsigned int middle = 0;
  unsigned int parenthesis = 0;
  
  string line = paragraph;
  string attributetyped;

  setFormat( 0, line.size(), black );

  for( unsigned int i=0; i<line.size(); i++ ) {
    
    if( line[i]=='(' ) {
      begin = i+1;
      parenthesis++;
      continue;
    }

    if( begin>0 && line[i]=='=' ) {
      attributetyped = line.substr( begin, i-begin );
      unsigned int j;
      for( j=0; j<attributetyped.size(); j++ )
	attributetyped[j] = tolower(attributetyped[j]);
      for( j=0; j<xrslattributes.size(); j++ ) {
	if( strcmp( attributetyped.c_str(), xrslattributes[j].c_str() )== 0 )
	  setFormat( begin, i-begin, red );
      }
      begin = 0;
      middle = i+1;
      continue;
    }

    if( middle>0 && ( (strcmp( attributetyped.c_str(), "dryrun" ) == 0) || (strcmp( attributetyped.c_str(), "join" ) == 0) || (strcmp( attributetyped.c_str(), "cache" ) == 0) ) ) {
      if( i==middle+2 && strcmp( line.substr( middle, 2 ).c_str(), "no" ) == 0 )
	setFormat( middle, 2, blue );
      if( i==middle+3 && strcmp( line.substr( middle, 3 ).c_str(), "yes" ) == 0 )
	setFormat( middle, 3, blue );

      continue;
    }

    if( middle>0 && (strcmp( attributetyped.c_str(), "nodeaccess" ) == 0) ) {
      if( i==middle+7 && strcmp( line.substr( middle, 7 ).c_str(), "inbound" ) == 0 )
	setFormat( middle, 7, blue );
      if( i==middle+8 && strcmp( line.substr( middle, 8 ).c_str(), "outbound" ) == 0 )
	setFormat( middle, 8, blue );

      continue;
    }

    if( middle>0 && line[i] != '"' && line[i] != ' ' && line[i] != ')') {
      setFormat( i, 1, blue );
      continue;
    }

    if( middle>0 && line[i]==')' ) { 
      parenthesis--;
      if( parenthesis==0 )
	middle = 0;
    }

  }

  return endStateOfLastPara;
};

#endif
