#include <iostream>
#include <string>
using namespace std;

#include "ftpClient.h"
#include "condition.h"

ftpClient::ftpClient( int answertimeout )
{
  if( globus_module_activate( GLOBUS_FTP_CLIENT_MODULE ) != GLOBUS_SUCCESS )
    goto notinitialized;
  if( globus_ftp_client_handleattr_init( &attr ) != GLOBUS_SUCCESS )
    goto notinitialized;
  if( globus_ftp_client_handle_init( &handle, &attr ) != GLOBUS_SUCCESS ) {
    globus_ftp_client_handleattr_destroy( &attr );
    goto notinitialized;
  }
  if( globus_ftp_client_operationattr_init( &opattr ) != GLOBUS_SUCCESS ) {
    globus_ftp_client_handleattr_destroy( &attr );
    globus_ftp_client_handle_destroy( &handle );
    goto notinitialized;
  }

  ready = true;
  timeout = answertimeout;
  return;

 notinitialized:
  ready = false;
}

ftpClient::~ftpClient( )
{
  if( ready == true ) {
    globus_ftp_client_operationattr_destroy( &opattr );
    globus_ftp_client_handle_destroy( &handle );
    globus_ftp_client_handleattr_destroy( &attr );
  }
  globus_module_deactivate( GLOBUS_FTP_CLIENT_MODULE );
}

ftpInfo ftpClient::fileExists( string url )
{
  ftpInfo info;

  if( !ready ) {
    info.error = "Ftp-handles not correctly initialized";
    return info;
  }

  if( globus_ftp_client_exists( &handle, url.c_str(), &opattr, ftpClientCallback, (void*)&cond ) != GLOBUS_SUCCESS ) {
    info.error = "Could not call globus_ftp_client_exists";
    return info;
  }

  bool valid;
  if( !cond.wait( valid, timeout ) ) {
    globus_ftp_client_abort( &handle );
    info.error = "Timeout waiting for response from gridftp-server";
    return info;
  }

  info.exists = valid;
  return info;
}

ftpInfo ftpClient::fileSize( string url )
{
  ftpInfo info;

  if( !ready ) {
    info.error = "Ftp-handles not correctly initialized";
    return info;
  }

  globus_off_t fsize;
  if( globus_ftp_client_size( &handle, url.c_str(), &opattr, &fsize, ftpClientCallback, (void*)&cond ) != GLOBUS_SUCCESS ) {
    info.error = "Could not call globus_ftp_client_exists";
    return info;
  }

  bool valid;
  if( !cond.wait( valid, timeout ) ) {
    globus_ftp_client_abort( &handle );
    info.error = "Timeout waiting for response from gridftp-server";
    return info;
  }

  if( valid ) {
    info.filesize = fsize;
  } else {
    info.filesize = -1;
  }

  return info;
}

void ftpClientCallback( void* condition, globus_ftp_client_handle_t* handle, globus_object_t* error ) {

  Condition<bool>* cond = (Condition<bool>*)condition;
  if( error != GLOBUS_SUCCESS ) {
    cond->signal( false );
  } else {
    cond->signal( true );
  }
}
