#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <vector>
#include <map>
#include <iostream>
#include <fstream>

#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
#include <globus_replica_catalog.h>
#include <globus_replica_catalog_sets.h>
#endif
#include <globus_ftp_client.h>
#include <globus_gass_transfer.h>
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
extern "C" {
#include <globus_rls_client.h>
}
#endif

#include "RemoteFile.h"
#include "MdsQuery.h"
#include "Xrsl.h"
#include "Utils.h"
#include "misc/condition.h"


RemoteFileInfo::RemoteFileInfo (const char T,
				const int timeout_) : type (T),
						      timeout (timeout_),
						      size (0),
						      queried (false) {}

bool RemoteFileInfo::IsType (const char T) const { return (type == T); }

bool RemoteFileInfo::IsLocal (Cluster * c) const { return false; }


long long int RemoteFileInfo::GetSize () {

  if (Query()) return -1;
  return size;
}


void RemoteFileInfo::RegisterCachedFile (const string & cluster) {

  if (!IsCached (cluster)) cached.push_back (cluster);
}


bool RemoteFileInfo::IsCached (const string & cluster) const {

  for (vector <string>::const_iterator vsi = cached.begin();
       vsi != cached.end(); vsi++)
    if (*vsi == cluster) return true;
  return false;
}


#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
RcLocationInfo::RcLocationInfo (const string & name_,
				const string & uc,
				const string & path) : name (name_) {

  int pos = uc.find ("://");
  if (pos == string::npos)
    url = "gsiftp://" + uc;
  else
    url = uc;
  if (!path.empty()) {
    if (url [url.length() - 1] != '/' && path [0] != '/') url += '/';
    url += path;
  }
}


const string & RcLocationInfo::GetName () const { return name; }


const string & RcLocationInfo::GetUrl () const { return url; }


const string RcLocationInfo::GetHost () const {

  int pos0 = url.find ("://");
  if (pos0 == string::npos)
    pos0 = 0;
  else
    pos0 += 3;
  int pos1 = url.find_first_of (":/", pos0);
  if (pos1 == string::npos)
    return url.substr (pos0);
  else
    return url.substr (pos0, pos1 - pos0);
}


RcFileInfo::RcFileInfo (const string & rc_,
			const string & lfn_,
			const int timeout) : rc (rc_),
					     lfn (lfn_),
					     RemoteFileInfo ('R', timeout) {}


bool RcFileInfo::operator== (const RemoteFileInfo & file) const {

  if (!file.IsType ('R')) return false;
  RcFileInfo * rcfile = (RcFileInfo *) &file;
  return (rc == rcfile->rc && lfn == rcfile->lfn);
}


int RcFileInfo::Query () {

  if (queried) return 0;
  queried = true;

  char * filelist [] = {(char *) lfn.c_str(), NULL};
  char * fileattrlist [] = {"size", NULL};
  char * locationattrlist [] = {"uc", "path", NULL};

  globus_result_t err;

  globus_replica_catalog_collection_handle_t handle;
  globus_replica_catalog_collection_handleattr_t handleattrs;
  err = globus_replica_catalog_collection_handleattr_init (&handleattrs);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_replica_catalog_collection_open (&handle, &handleattrs,
						(char *) rc.c_str());
  if (err != GLOBUS_SUCCESS) goto errorexit;

  globus_replica_catalog_attribute_set_t fileattrset;
  err = globus_replica_catalog_attribute_set_init (&fileattrset);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_replica_catalog_logicalfile_list_attributes (&handle, (char *)
							    lfn.c_str(),
							    fileattrlist,
							    &fileattrset);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_replica_catalog_attribute_set_first (&fileattrset);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  while (globus_replica_catalog_attribute_set_more (&fileattrset)) {
    char * aname;
    char ** values;
    err = globus_replica_catalog_attribute_set_get_name (&fileattrset, &aname);
    if (err != GLOBUS_SUCCESS) goto errorexit;
    err = globus_replica_catalog_attribute_set_get_values (&fileattrset,
							   &values);
    if (err != GLOBUS_SUCCESS) goto errorexit;
    if (strcasecmp (aname, "size") == 0) size = atoll (values [0]);
    globus_libc_free (aname);
    err = globus_replica_catalog_attribute_set_free_values (values);
    if (err != GLOBUS_SUCCESS) goto errorexit;
    err = globus_replica_catalog_attribute_set_next (&fileattrset);
    // Don't know why I can't check this
    // if (err != GLOBUS_SUCCESS) goto errorexit;
  }
  err = globus_replica_catalog_attribute_set_destroy (&fileattrset);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  globus_replica_catalog_entry_set_t locations;
  err = globus_replica_catalog_entry_set_init (&locations);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_replica_catalog_collection_find_locations (&handle, filelist,
							  true,
							  locationattrlist,
							  &locations);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_replica_catalog_entry_set_first (&locations);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  while (globus_replica_catalog_entry_set_more (&locations)) {
    char * lname;
    err = globus_replica_catalog_entry_set_get_name (&locations, &lname);
    if (err != GLOBUS_SUCCESS) goto errorexit;
    string name = lname;
    globus_libc_free (lname);
    globus_replica_catalog_attribute_set_t locationattrset;
    err = globus_replica_catalog_attribute_set_init (&locationattrset);
    if (err != GLOBUS_SUCCESS) goto errorexit;
    err = globus_replica_catalog_entry_set_get_attributes (&locations,
							   &locationattrset);
    if (err != GLOBUS_SUCCESS) goto errorexit;
    err = globus_replica_catalog_attribute_set_first (&locationattrset);
    if (err != GLOBUS_SUCCESS) goto errorexit;
    string uc;
    string path;
    while (globus_replica_catalog_attribute_set_more (&locationattrset)) {
      char * aname;
      char ** values;
      err = globus_replica_catalog_attribute_set_get_name (&locationattrset,
							   &aname);
      if (err != GLOBUS_SUCCESS) goto errorexit;
      err = globus_replica_catalog_attribute_set_get_values (&locationattrset,
							     &values);
      if (err != GLOBUS_SUCCESS) goto errorexit;
      if (strcasecmp (aname, "uc") == 0) uc = values [0];
      if (strcasecmp (aname, "path") == 0) path = values [0];
      globus_libc_free (aname);
      err = globus_replica_catalog_attribute_set_free_values (values);
      if (err != GLOBUS_SUCCESS) goto errorexit;
      err = globus_replica_catalog_attribute_set_next (&locationattrset);
      // Don't know why I can't check this
      // if (err != GLOBUS_SUCCESS) goto errorexit;
    }
    RcLocationInfo locinfo (name, uc, path);
    if (locinfo.GetUrl().substr (0, 8) == "cache://")
      RegisterCachedFile (locinfo.GetHost());
    loclist.push_back (locinfo);
    err = globus_replica_catalog_attribute_set_destroy (&locationattrset);
    if (err != GLOBUS_SUCCESS) goto errorexit;
    err = globus_replica_catalog_entry_set_next (&locations);
    // Don't know why I can't check this
    // if (err != GLOBUS_SUCCESS) goto errorexit;
  }
  err = globus_replica_catalog_entry_set_destroy (&locations);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  err = globus_replica_catalog_collection_close (&handle);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_replica_catalog_collection_handleattr_destroy (&handleattrs);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  return 0;

 errorexit:
  char * tmp = globus_object_printable_to_string (globus_error_get (err));
  cerr << "Replica Catalog error: " << tmp << endl;
  free (tmp);
  cerr << "when querying " << rc << " for " << lfn << endl;
  return 1;
}
#endif


#ifdef HAVE_GLOBUS_RLS_CLIENT_H
RlsLocationInfo::RlsLocationInfo (const string & url_) : url (url_) {}


const string & RlsLocationInfo::GetUrl () const { return url; }


const string RlsLocationInfo::GetHost () const {

  int pos0 = url.find ("://");
  if (pos0 == string::npos)
    pos0 = 0;
  else
    pos0 += 3;
  int pos1 = url.find_first_of (":/", pos0);
  if (pos1 == string::npos)
    return url.substr (pos0);
  else
    return url.substr (pos0, pos1 - pos0);
}


RlsFileInfo::RlsFileInfo (const string & server_,
			  const string & lfn_,
			  const int timeout) : server (server_),
					       lfn (lfn_),
					       RemoteFileInfo ('Q', timeout) {}


bool RlsFileInfo::operator== (const RemoteFileInfo & file) const {

  if (!file.IsType ('Q')) return false;
  RlsFileInfo * rlsfile = (RlsFileInfo *) &file;
  return (server == rlsfile->server && lfn == rlsfile->lfn);
}


int RlsFileInfo::Query () {

  if (queried) return 0;
  queried = true;

  globus_result_t err;

  globus_rls_client_set_timeout (timeout);

  globus_rls_handle_t * handle;
  err = globus_rls_client_connect ((char *) (("rls://" + server).c_str()),
				   &handle);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  globus_list_t * attr_list;
  err = globus_rls_client_lrc_attr_value_get (handle, (char *) (lfn.c_str()),
					      "size", globus_rls_obj_lrc_lfn,
					      &attr_list);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  globus_list_t * a_list;
  for (a_list = attr_list; a_list; a_list = globus_list_rest (a_list)) {
    globus_rls_attribute_t * attr =
      (globus_rls_attribute_t *) globus_list_first (a_list);
    if (attr->type != globus_rls_attr_type_str) continue;
    if (strcmp (attr->name, "size") == 0) size = atoll (attr->val.s);
  }

  err = globus_rls_client_free_list (attr_list);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  globus_list_t * pfn_list;
  err = globus_rls_client_lrc_get_pfn (handle, (char *) (lfn.c_str()), 0, 0,
				       &pfn_list);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  globus_list_t * p_list;
  for (p_list = pfn_list; p_list; p_list = globus_list_rest (p_list)) {
    globus_rls_string2_t * str2 =
      (globus_rls_string2_t *) globus_list_first (p_list);
    RlsLocationInfo locinfo (str2->s2);
    if (locinfo.GetUrl().substr (0, 8) == "cache://")
      RegisterCachedFile (locinfo.GetHost());
    loclist.push_back (locinfo);
  }

  err = globus_rls_client_free_list (pfn_list);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  err = globus_rls_client_close (handle);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  return 0;

 errorexit:
  char tmp [1024];
  globus_rls_client_error_info (err, NULL, tmp, 1024, false);
  cerr << "Replica Location Service error: " << tmp << endl;
  cerr << "when querying rls://" << server << " for " << lfn << endl;
  return 1;
}
#endif


FtpFileInfo::FtpFileInfo (const string & url_,
			  const int timeout) : url (url_),
					       RemoteFileInfo ('F', timeout) {}


bool FtpFileInfo::operator== (const RemoteFileInfo & file) const {

  if (!file.IsType ('F')) return false;
  FtpFileInfo * ftpfile = (FtpFileInfo *) &file;
  return (url == ftpfile->url);
}


void FtpCallback (void * arg,
		  globus_ftp_client_handle_t * handle,
		  globus_object_t * err) {

  Condition <bool> * cond = (Condition <bool> *) arg;
  if (err) {
    char * tmp = globus_object_printable_to_string (err);
    string errstr = tmp;
    free (tmp);
    int pos;
    while ((pos = errstr.find ("\r\n")) != string::npos) errstr.erase (pos, 2);
    if (errstr.find ("end-of-file") != string::npos)
      errstr = "server unexpectedly closed connection";
    cerr << "Gridftp error: " << errstr << endl;
    cond->signal (false);
  }
  else
    cond->signal (true);
}


int FtpFileInfo::Query () {

  if (queried) return 0;
  queried = true;

  Condition <bool> completed;
  globus_result_t err;
  globus_off_t gsize = 0;

  globus_ftp_client_handle_t handle;
  globus_ftp_client_handleattr_t attr;
  err = globus_ftp_client_handleattr_init (&attr);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_ftp_client_handle_init (&handle, &attr);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  globus_ftp_client_operationattr_t oattr;
  err = globus_ftp_client_operationattr_init (&oattr);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  err = globus_ftp_client_size (&handle, url.c_str(), &oattr,
				&gsize, FtpCallback, &completed);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  bool OK;
  if (!completed.wait (OK, timeout * 1000)) {
    cerr << "Gridftp timeout waiting for size" << endl;
    cerr << "when querying " << url << endl;
    globus_ftp_client_abort (&handle);
    completed.wait (OK, -1);
    return 1;
  }
  if (!OK) {
    cerr << "when querying " << url << endl;
    return 1;
  }

  size = gsize;

  err = globus_ftp_client_operationattr_destroy (&oattr);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_ftp_client_handle_destroy (&handle);
  if (err != GLOBUS_SUCCESS) goto errorexit;
  err = globus_ftp_client_handleattr_destroy (&attr);
  if (err != GLOBUS_SUCCESS) goto errorexit;

  return 0;

 errorexit:
  char * tmp = globus_object_printable_to_string (globus_error_get (err));
  string errstr = tmp;
  free (tmp);
  int pos;
  while ((pos = errstr.find ("\r\n")) != string::npos) errstr.erase (pos, 2);
  cerr << "Gridftp error: " << errstr << endl;
  cerr << "when querying " << url << endl;
  return 1;
}


bool FtpFileInfo::IsLocal (Cluster * c) const {

  return c->MatchLocalSe (url);
}


HttpFileInfo::HttpFileInfo (const string & url_,
			    const int tiemout) : url (url_),
						 RemoteFileInfo ('H',
								 timeout) {}


bool HttpFileInfo::operator== (const RemoteFileInfo & file) const {

  if (!file.IsType ('H')) return false;
  HttpFileInfo * httpfile = (HttpFileInfo *) &file;
  return (url == httpfile->url);
}


int HttpFileInfo::Query () {

  if (queried) return 0;
  queried = true;

  int err;

  globus_gass_transfer_request_t request;

  err = globus_gass_transfer_get (&request, NULL, (char *) url.c_str());
  if (err) goto errorexit;

  size = globus_gass_transfer_request_get_length (request);

  err = globus_gass_transfer_request_destroy (request);
  // This gives err = 12 for some reason, so the next line is commented out
  // if (err) goto errorexit;

  return 0;

 errorexit:
  cerr << "Http error number " << err << endl;
  cerr << "when querying " << url << endl;
  return 1;
}


bool HttpFileInfo::IsLocal (Cluster * c) const {

  return c->MatchLocalSe (url);
}


LocalFileInfo::LocalFileInfo (const string & fname,
			      const int timeout) : filename (fname),
						   RemoteFileInfo ('L',
								   timeout) {}


bool LocalFileInfo::operator== (const RemoteFileInfo & file) const {

  if (!file.IsType ('L')) return false;
  LocalFileInfo * localfile = (LocalFileInfo *) &file;
  return (filename == localfile->filename);
}


int LocalFileInfo::Query () {

  if (queried) return 0;
  queried = true;

  ifstream file (filename.c_str());
  if (!file) goto errorexit;
  file.seekg (0, ios::end);
  size = file.tellg();
  file.close();

  return 0;

 errorexit:
  cerr << "Error: Can not open local file " << filename << endl;
  return 1;
}


RemoteFile::RemoteFile (const string & optstring) : fileinfo (NULL) {

  options = ParseOptionString (optstring);
}


void RemoteFile::Register () {

  if (!fileinfo) return;
  bool found = false;
  for (vector <RemoteFileInfo *>::iterator vrcii = filecloud.begin();
       !found && vrcii != filecloud.end(); vrcii++)
    if (*fileinfo == **vrcii) {
      delete fileinfo;
      fileinfo = *vrcii;
      found = true;
    }
    if (!found) filecloud.push_back (fileinfo);
}


FileLocation RemoteFile::Local (Cluster * c, const bool defaultcache) const {

  if (fileinfo->Query()) return IOERROR;

  bool foundlocal = fileinfo->IsLocal (c);
  bool local = false;
  bool cache = defaultcache;
  GetBooleanOption (options, "local", &local);
  GetBooleanOption (options, "cache", &cache);
  if (!foundlocal && local) return NOLOCATION;
  if (cache && fileinfo->IsCached (c->GetName())) return CACHED;
  if (foundlocal) return (cache ? LOCAL : LOCALNOCACHE);
  return (cache ? REMOTE : REMOTENOCACHE);
}


long long int RemoteFile::GetSize () const {

  return fileinfo->GetSize();
}


bool RemoteFile::IsType (const char T) const {

  return fileinfo->IsType (T);
}


void RemoteFile::RegisterCachedFile (const string & cluster) const {

  fileinfo->RegisterCachedFile (cluster);
}


#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
RcLocation::RcLocation (const string & name_,
			const string & optstring) : name (name_) {

  options = ParseOptionString (optstring);
}


const string & RcLocation::GetName () const { return name; }


bool RcLocation::GetBooleanOption (const string & opt, bool * value) const {

  return ::GetBooleanOption (options, opt, value);
}


RcFile::RcFile (const string & rc,
		const string & lfn,
		const string & locstring,
		const string & optstring,
		const int timeout) : RemoteFile (optstring) {

  fileinfo = new RcFileInfo (rc, lfn, timeout);
  Register();

  if (locstring.empty()) return;
  int pos = 0;
  while (pos != string::npos) {
    int barpos = locstring.find ('|', pos);
    string entry;
    if (barpos == string::npos)
      entry = locstring.substr (pos);
    else
      entry = locstring.substr (pos, barpos - pos);
    int semicpos = entry.find (';');
    if (semicpos == 0)
      defaultoptions = ParseOptionString (entry.substr (1));
    else {
      string name;
      string optstring;
      if (semicpos == string::npos)
	name = entry;
      else {
	name = entry.substr (0, semicpos);
	optstring = entry.substr (semicpos + 1);
      }
      loclist.push_back (RcLocation (name, optstring));
    }
    pos = barpos;
    if (pos != string::npos) pos++;
  }
}


FileLocation RcFile::Local (Cluster * c, const bool defaultcache) const {

  if (fileinfo->Query()) return IOERROR;

  bool foundlocal = false;
  bool found = false;
  bool cache = defaultcache;

  for (vector <RcLocation>::const_iterator vrli = loclist.begin();
       !foundlocal && vrli != loclist.end(); vrli++) {
    for (vector <RcLocationInfo>::const_iterator vrlii =
	   ((RcFileInfo *) fileinfo)->loclist.begin(); !foundlocal &&
	   vrlii != ((RcFileInfo *) fileinfo)->loclist.end(); vrlii++) {
      if (vrli->GetName() == vrlii->GetName()) {
	if (c->MatchLocalSe (vrlii->GetUrl())) foundlocal = true;
	bool local = false;
	if (!vrli->GetBooleanOption ("local", &local))
	  GetBooleanOption (defaultoptions, "local", &local);
	if (foundlocal || (!found && !local)) {
	  found = true;
	  cache = defaultcache;
	  if (!vrli->GetBooleanOption ("cache", &cache))
	    GetBooleanOption (defaultoptions, "cache", &cache);
	}
      }
    }
  }

  if (!foundlocal && (loclist.empty() || !defaultoptions.empty())) {
    for (vector <RcLocationInfo>::const_iterator vrlii =
	   ((RcFileInfo *) fileinfo)->loclist.begin(); !foundlocal &&
	   vrlii != ((RcFileInfo *) fileinfo)->loclist.end(); vrlii++) {
      if (c->MatchLocalSe (vrlii->GetUrl())) foundlocal = true;
      bool local = false;
      GetBooleanOption (defaultoptions, "local", &local);
      if (foundlocal || (!found && !local)) {
	found = true;
	cache = defaultcache;
	GetBooleanOption (defaultoptions, "cache", &cache);
      }
    }
  }

  if (!found) return NOLOCATION;
  if (cache && fileinfo->IsCached (c->GetName())) return CACHED;
  if (foundlocal) return (cache ? LOCAL : LOCALNOCACHE);
  return (cache ? REMOTE : REMOTENOCACHE);
}
#endif


#ifdef HAVE_GLOBUS_RLS_CLIENT_H
RlsLocation::RlsLocation (const string & url_,
			  const string & optstring) : url (url_) {

  options = ParseOptionString (optstring);
}


const string & RlsLocation::GetUrl () const { return url; }


bool RlsLocation::GetBooleanOption (const string & opt, bool * value) const {

  return ::GetBooleanOption (options, opt, value);
}


RlsFile::RlsFile (const string & server,
		  const string & lfn,
		  const string & locstring,
		  const string & optstring,
		  const int timeout) : RemoteFile (optstring) {

  fileinfo = new RlsFileInfo (server, lfn, timeout);
  Register();

  if (locstring.empty()) return;
  int pos = 0;
  while (pos != string::npos) {
    int barpos = locstring.find ('|', pos);
    string url;
    if (barpos == string::npos)
      url = locstring.substr (pos);
    else
      url = locstring.substr (pos, barpos - pos);
    int semicpos = url.find (';');
    if (semicpos == 0)
      defaultoptions = ParseOptionString (url.substr (1));
    else {
      string url;
      string optstring;
      if (semicpos != string::npos) {
	int pos0 = url.find ("://");
	if (pos0 == string::npos)
	  pos0 = 0;
	else
	  pos0 += 3;
	int slashpos = url.find ('/', pos0);
	if (semicpos < slashpos) {
	  optstring = url.substr (semicpos + 1, slashpos - semicpos - 1);
	  url.erase (semicpos, slashpos - semicpos);
	}
      }
      loclist.push_back (RlsLocation (url, optstring));
    }
    pos = barpos;
    if (pos != string::npos) pos++;
  }
}


FileLocation RlsFile::Local (Cluster * c, const bool defaultcache) const {

  if (fileinfo->Query()) return IOERROR;

  bool foundlocal = false;
  bool found = false;
  bool cache = defaultcache;

  for (vector <RlsLocation>::const_iterator vrli = loclist.begin();
       !foundlocal && vrli != loclist.end(); vrli++) {
    for (vector <RlsLocationInfo>::const_iterator vrlii =
	   ((RlsFileInfo *) fileinfo)->loclist.begin(); !foundlocal &&
	   vrlii != ((RlsFileInfo *) fileinfo)->loclist.end(); vrlii++) {
      if (vrli->GetUrl() == vrlii->GetUrl()) {
	if (c->MatchLocalSe (vrlii->GetUrl())) foundlocal = true;
	bool local = false;
	if (!vrli->GetBooleanOption ("local", &local))
	  GetBooleanOption (defaultoptions, "local", &local);
	if (foundlocal || (!found && !local)) {
	  found = true;
	  cache = defaultcache;
	  if (!vrli->GetBooleanOption ("cache", &cache))
	    GetBooleanOption (defaultoptions, "cache", &cache);
	}
      }
    }
  }

  if (!foundlocal && (loclist.empty() || !defaultoptions.empty())) {
    for (vector <RlsLocationInfo>::const_iterator vrlii =
	   ((RlsFileInfo *) fileinfo)->loclist.begin(); !foundlocal &&
	   vrlii != ((RlsFileInfo *) fileinfo)->loclist.end(); vrlii++) {
      if (c->MatchLocalSe (vrlii->GetUrl())) foundlocal = true;
      bool local = false;
      GetBooleanOption (defaultoptions, "local", &local);
      if (foundlocal || (!found && !local)) {
	found = true;
	cache = defaultcache;
	GetBooleanOption (defaultoptions, "cache", &cache);
      }
    }
  }

  if (!found) return NOLOCATION;
  if (cache && fileinfo->IsCached (c->GetName())) return CACHED;
  if (foundlocal) return (cache ? LOCAL : LOCALNOCACHE);
  return (cache ? REMOTE : REMOTENOCACHE);
}
#endif


FtpFile::FtpFile (const string & url,
		  const string & optstring,
		  const int timeout) : RemoteFile (optstring) {

  fileinfo = new FtpFileInfo (url, timeout);
  Register();
}


HttpFile::HttpFile (const string & url,
		    const string & optstring,
		    const int timeout) : RemoteFile (optstring) {

  fileinfo = new HttpFileInfo (url, timeout);
  Register();
}


LocalFile::LocalFile (const string & filename,
		      const string & optstring,
		      const int timeout) : RemoteFile (optstring) {

  fileinfo = new LocalFileInfo (filename, timeout);
  Register();
}


FileLocation LocalFile::Local (Cluster * c, const bool defaultcache) const {

  if (fileinfo->Query()) return IOERROR;

  return REMOTENOCACHE;
}


RemoteFileQuery::RemoteFileQuery (const RemoteFileQuery & query) {

  for (vector <RemoteFile *>::const_iterator vrfi = query.filelist.begin();
       vrfi != query.filelist.end(); vrfi++) {
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
    if ((*vrfi)->IsType ('R'))
      filelist.push_back (new RcFile (* (RcFile *) *vrfi));
#endif
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
    if ((*vrfi)->IsType ('Q'))
      filelist.push_back (new RlsFile (* (RlsFile *) *vrfi));
#endif
    if ((*vrfi)->IsType ('F'))
      filelist.push_back (new FtpFile (* (FtpFile *) *vrfi));
    if ((*vrfi)->IsType ('H'))
      filelist.push_back (new HttpFile (* (HttpFile *) *vrfi));
    if ((*vrfi)->IsType ('L'))
      filelist.push_back (new LocalFile (* (LocalFile *) *vrfi));
  }
}


RemoteFileQuery::~RemoteFileQuery () {

  for (vector <RemoteFile *>::iterator vrfi = filelist.begin();
       vrfi != filelist.end(); vrfi++)
    delete *vrfi;
  filelist.clear();
}


RemoteFileQuery & RemoteFileQuery::operator= (const RemoteFileQuery & query) {

  if (this != &query) {
    for (vector <RemoteFile *>::iterator vrfi = filelist.begin();
	 vrfi != filelist.end(); vrfi++)
      delete *vrfi;
    filelist.clear();

    for (vector <RemoteFile *>::const_iterator vrfi = query.filelist.begin();
	 vrfi != query.filelist.end(); vrfi++) {
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
      if ((*vrfi)->IsType ('R'))
	filelist.push_back (new RcFile (* (RcFile *) *vrfi));
#endif
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
      if ((*vrfi)->IsType ('Q'))
	filelist.push_back (new RlsFile (* (RlsFile *) *vrfi));
#endif
      if ((*vrfi)->IsType ('F'))
	filelist.push_back (new FtpFile (* (FtpFile *) *vrfi));
      if ((*vrfi)->IsType ('H'))
	filelist.push_back (new HttpFile (* (HttpFile *) *vrfi));
      if ((*vrfi)->IsType ('L'))
	filelist.push_back (new LocalFile (* (LocalFile *) *vrfi));
    }
  }
  return *this;
}


void RemoteFileQuery::AddFile (RemoteFile * const file) {

  filelist.push_back (file);
}


int RemoteFileQuery::AddSizes (Cluster * c,
			       long long int * cachesize,
			       long long int * sessdirsize,
			       long long int * remotesize,
			       long long int * localsize,
			       const bool defaultcache) const {

  *cachesize = 0;
  *sessdirsize = 0;
  *remotesize = 0;
  *localsize = 0;

  for (vector <RemoteFile *>::const_iterator vrfi = filelist.begin();
       vrfi != filelist.end(); vrfi++) {
    switch ((*vrfi)->Local (c, defaultcache)) {
    case REMOTE:
      *remotesize += (*vrfi)->GetSize();
      *cachesize += (*vrfi)->GetSize();
      break;
    case LOCAL:
      *localsize += (*vrfi)->GetSize();
      *cachesize += (*vrfi)->GetSize();
      break;
    case REMOTENOCACHE:
      *remotesize += (*vrfi)->GetSize();
      *sessdirsize += (*vrfi)->GetSize();
      break;
    case LOCALNOCACHE:
      *localsize += (*vrfi)->GetSize();
      *sessdirsize += (*vrfi)->GetSize();
      break;
    case CACHED:
      break;
    case NOLOCATION:
      *cachesize = *sessdirsize = *remotesize = *localsize = UNDEFINED;
      return 0;
      break;
    case IOERROR:
      *cachesize = *sessdirsize = *remotesize = *localsize = UNDEFINED;
      return 1;
      break;
    }
  }
  return 0;
}


int RemoteFileQuery::RegisterCachedFiles (Cluster * c,
					  const bool defaultcache) const {

  for (vector <RemoteFile *>::const_iterator vrfi = filelist.begin();
       vrfi != filelist.end(); vrfi++) {
    switch ((*vrfi)->Local (c, defaultcache)) {
    case REMOTE:
    case LOCAL:
      (*vrfi)->RegisterCachedFile (c->GetName());
      break;
    case REMOTENOCACHE:
    case LOCALNOCACHE:
    case CACHED:
      break;
    case NOLOCATION:
    case IOERROR:
      return 1;
      break;
    }
  }
  return 0;
}


vector <RemoteFileInfo *> RemoteFile::filecloud;
