#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "MdsQuery.h"
#include "Xrsl.h"
#include "RemoteFile.h"
#include "Target.h"



Target::Target () : xrsl (NULL),
					neededcachesize (UNDEFINED),
					neededsessdirsize (UNDEFINED),
					remotesize (UNDEFINED),
					localsize (UNDEFINED) {};

Target::Target (Cluster * c, Queue * q) : cluster (c),
					  queue (q),
					  xrsl (NULL),
					  neededcachesize (UNDEFINED),
					  neededsessdirsize (UNDEFINED),
					  remotesize (UNDEFINED),
					  localsize (UNDEFINED) {};

bool Target::operator! () const { return !xrsl; }

bool Target::valid() const { return xrsl.valid(); }

int Target::CalculateNeededFileSizes (const int timeout) {

  vector <string> inputfiles;
  if (xrsl.GetInputFiles (inputfiles)) return 1;
  string defaultrc;
  if (xrsl.GetRc (defaultrc)) return 1;
  bool defaultcache;
  if (xrsl.GetDefaultCache (&defaultcache)) return 1;

  for (vector <string>::iterator vsi = inputfiles.begin();
       vsi != inputfiles.end(); vsi++) {

    string url = *vsi;
    string optstring;

    int pos = url.find ("://");
    if (pos != string::npos) {
      pos += 3;
      int atpos = url.find ('@', pos);
      if (atpos != string::npos) {
	int slashpos = url.find ('/', pos);
	if (slashpos == string::npos || atpos < slashpos) pos = atpos + 1;
      }
      int semicpos = url.find (';', pos);
      if (semicpos != string::npos) {
	int slashpos = url.find ('/', pos);
	if (slashpos == string::npos) {
	  optstring = url.substr (semicpos + 1);
	  url.erase (semicpos);
	}
	else if (slashpos > semicpos) {
	  optstring = url.substr (semicpos + 1, slashpos - semicpos - 1);
	  url.erase (semicpos, slashpos - semicpos);
	}
      }
    }

    if (url.substr (0, 5) == "rc://") {
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
      string locstring;
      string rc;

      int pos = 5;
      int slashpos = url.find ('/', pos);
      int atpos = url.find ('@', pos);
      if (slashpos == string::npos) {
	cerr << "Error: Can not parse url " << url << endl;
	return 1;
      }
      if (atpos != string::npos && atpos < slashpos) {
	locstring = url.substr (pos, atpos - pos);
	pos = atpos + 1;
      }
      else if (strcasecmp (url.substr (slashpos + 1, 3).c_str(), "lc=") != 0) {
	locstring = url.substr (pos, slashpos - pos);
	pos = slashpos;
      }
      if (pos == slashpos) {
	rc = defaultrc;
	if (rc.empty()) {
	  cerr << "Error: XRSL attribute \"replicacollection\" missing"
	       << endl;
	  return 1;
	}
      }
      else {
	slashpos++;
	slashpos = url.find ('/', slashpos);
	if (slashpos == string::npos) {
	  cerr << "Error: Can not parse url " << url << endl;
	  return 1;
	}
	rc = "ldap://" + url.substr (pos, slashpos - pos);
      }
      string file = url.substr (slashpos + 1);
      query.AddFile (new RcFile (rc, file, locstring, optstring, timeout));
#else
      cerr << "Warning: This version of the NorduGrid toolkit user interface was compiled" << endl;
      cerr << "  without support for the globus replica catalog (rc:// urls)" << endl;
      cerr << "Warning: The location of the inputfile can not be used for brokering" << endl;
#endif
    }

    else if (url.substr (0, 6) == "rls://") {
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
      string locstring;

      int pos = 6;
      int slashpos = url.find ('/', pos);
      int atpos = url.find ('@', pos);
      if (slashpos == string::npos) {
	cerr << "Error: Can not parse url " << url << endl;
	return 1;
      }
      if (atpos != string::npos && atpos < slashpos) {
	locstring = url.substr (pos, atpos - pos);
	pos = atpos + 1;
      }
      if (pos == slashpos) {
	cerr << "Error: Can not parse url " << url << endl;
	return 1;
      }
      string server = url.substr (pos, slashpos - pos);
      string file = url.substr (slashpos + 1);
      query.AddFile (new RlsFile (server, file, locstring, optstring,
				  timeout));
#else
      cerr << "Warning: This version of the NorduGrid toolkit user interface was compiled" << endl;
      cerr << "  without support for the globus replica location service (rls:// urls)" << endl;
      cerr << "Warning: The location of the inputfile can not be used for brokering" << endl;
#endif
    }

    else if (url.substr (0, 6) == "ftp://" ||
	     url.substr (0, 9) == "gsiftp://") {
      query.AddFile (new FtpFile (url, optstring, timeout));
    }

    else if (url.substr (0, 7) == "http://" ||
	     url.substr (0, 8) == "https://") {
      query.AddFile (new HttpFile (url, optstring, timeout));
    }

    else if (url.substr (0, 7) == "file://" ||
	     url.find ("://") == string::npos) {
      if (url.substr (0, 7) == "file://")
	url = url.substr (7);
      else if (url[0] != '/') {
	char buffer[PATH_MAX];
	getcwd (buffer, PATH_MAX);
	url = (string) buffer + '/' + url;
      }
      query.AddFile (new LocalFile (url, optstring, timeout));
    }

    else {
      cerr << "Warning: This version of the NorduGrid toolkit user interface does not" << endl;
      cerr << "  understand " << url.substr (0, url.find ("://") + 3) << " urls" << endl;
      cerr << "Warning: The location of the inputfile can not be used for brokering" << endl;
    }
  }

  if (query.AddSizes (cluster, &neededcachesize, &neededsessdirsize,
		      &remotesize, &localsize, defaultcache)) return 1;

  long long int disk;
  if (xrsl.GetDisk (&disk)) return 1;
  if (disk != UNDEFINED) neededsessdirsize += disk;

  return 0;
}


long long int Target::GetNeededCacheSize () const {

  return neededcachesize;
}


long long int Target::GetNeededSessDirSize () const {

  return neededsessdirsize;
}


long long int Target::GetNeededTotalSize () const {

  if (neededcachesize == UNDEFINED || neededsessdirsize == UNDEFINED)
    return UNDEFINED;
  return neededcachesize + neededsessdirsize;
}


long long int Target::GetRemoteSize () const {

  return remotesize;
}


long long int Target::GetLocalSize () const {

  return localsize;
}


const Environment * Target::FindEnvironment (const string & attr,
					     const EnvironmentTest &
					     envtest) const {

  const Environment * cenv = cluster->FindEnvironment (attr, envtest);
  const Environment * qenv = queue->FindEnvironment (attr, envtest);

  if (!cenv) return qenv;
  if (!qenv) return cenv;
  if (cenv->GetVersion() > qenv->GetVersion())
    return cenv;
  else
    return qenv;
}


int Target::Accept () {

  long int time;
  if (GetCpuTime (&time)) return 1;
  int count;
  if (xrsl.GetCount (&count)) return 1;
  bool defaultcache;
  if (xrsl.GetDefaultCache (&defaultcache)) return 1;

  queue->Accept (time, count);

  if (cluster->GetCacheFree() != UNDEFINED)
    cluster->ClaimCache (neededcachesize);
  else {
    cluster->ClaimSessionDir (neededcachesize);
    queue->ClaimUserDiskSpace (neededcachesize);
  }

  cluster->ClaimSessionDir (neededsessdirsize);
  queue->ClaimUserDiskSpace (neededsessdirsize);

  if (query.RegisterCachedFiles (cluster, defaultcache)) return 1;

  return 0;
}


const string & Target::GetArchitecture () const {

  if (!queue->GetArchitecture().empty())
    return queue->GetArchitecture();
  else
    return cluster->GetArchitecture();
}


int Target::GetFrequency () const {

  int frequency = queue->GetFrequency();
  if (frequency == UNDEFINED) frequency = cluster->GetFrequency();
  if (frequency == UNDEFINED) {
    cout << "Warning: Could not determine CPU frequency for queue "
	 << queue->GetName() << " at cluster " << cluster->GetName()
	 << ", assuming 1 GHz" << endl;
    frequency = 1000;
  }
  return frequency;
}


int Target::GetCpuTime (long int * time) const {

  long int cputime;
  if (xrsl.GetCpuTime (&cputime)) return 1;
  long int gridtime;
  if (xrsl.GetGridTime (&gridtime)) return 1;
  map<string, pair<float, long int> > benchmarks;
  if (xrsl.GetBenchmarks (benchmarks)) return 1;

  if (cputime != UNDEFINED && gridtime != UNDEFINED) {
    cerr << "Error: Both \"cputime\" and \"gridtime\" given in the XRSL"
	 << endl;
    return 1;
  }
  if (cputime != UNDEFINED && !benchmarks.empty()) {
    cerr << "Error: Both \"cputime\" and \"benchmarks\" given in the XRSL"
	 << endl;
    return 1;
  }
  if (gridtime != UNDEFINED && !benchmarks.empty()) {
    cerr << "Error: Both \"gridtime\" and \"benchmarks\" given in the XRSL"
	 << endl;
    return 1;
  }

  long int benchmarktime = UNDEFINED;
  if (!benchmarks.empty()) {
    for (map<string, pair<float, long int> >::iterator mspi =
	   benchmarks.begin(); mspi != benchmarks.end(); mspi++) {
      long int thisbenchmarktime = UNDEFINED;
      if (queue->GetBenchmark (mspi->first) > 0)
	thisbenchmarktime += (long int) (mspi->second.second *
					 mspi->second.first /
					 queue->GetBenchmark (mspi->first));
      else if (cluster->GetBenchmark (mspi->first) > 0)
	thisbenchmarktime += (long int) (mspi->second.second *
					 mspi->second.first /
					 cluster->GetBenchmark (mspi->first));
      if (benchmarktime < thisbenchmarktime) benchmarktime = thisbenchmarktime;
    }
  }

  if (benchmarktime != UNDEFINED)
    *time = benchmarktime;
  else if (gridtime != UNDEFINED)
    *time = gridtime * 2800 / GetFrequency();
  else if (cputime != UNDEFINED)
    *time = cputime;
  else
    *time = queue->GetDefaultCpuTime();

  return 0;
}


long int Target::GetMaxCpuTime () const {

  long int cputime;
  if (GetCpuTime (&cputime)) return -2;
  // -1 is used for no limit / infinite
  return cputime;
}


int Target::GetTotalCpus () const {

  if (queue->GetTotalCpus() != UNDEFINED)
    return queue->GetTotalCpus();
  else if (cluster->GetTotalCpus() != UNDEFINED)
    return cluster->GetTotalCpus();
  return 0;
}


int Target::GetNodeMemory () const {

  if (queue->GetNodeMemory() != UNDEFINED)
    return queue->GetNodeMemory();
  else
    return cluster->GetNodeMemory();
}


int Target::GetQueued () const {

  if (queue->GetQueued() != UNDEFINED)
    return queue->GetQueued();
  else if (cluster->GetQueued() != UNDEFINED)
    return cluster->GetQueued();
  return INT_MAX;
}
