#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <map>
#include <algorithm>
#include <ctype.h>

#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#include <iomanip>

#include "Utils.h"


// need to wrap for gcc 3.2
inline int to_lower (int x) { return tolower(x); }


map <int, int> ParseStringToMap (const string & value) {

  // Parses a string of the type "2 4:711 6:710" to an integer to integer map
  // A missing second arguments are given the value INT_MAX

  map <int, int> result;

  if (value.empty()) return result;
  int pos = 0;
  while (pos != string::npos) {
    int spacepos = value.find (' ', pos);
    string entry;
    if (spacepos == string::npos)
      entry = value.substr (pos);
    else
      entry = value.substr (pos, spacepos - pos);
    int first;
    int second;
    int colonpos = entry.find (':');
    if (colonpos == string::npos) {
      first = atoi (entry.c_str());
      second = INT_MAX;
    }
    else {
      first = atoi (entry.substr (0, colonpos).c_str());
      second = atoi (entry.substr (colonpos + 1).c_str());
    }
    result [first] = second;
    pos = spacepos;
    if (pos != string::npos) pos++;
  }
  return result;
}


map <long int, int> ParseFreeCpuStringToMap (const string & value) {

  map <long int, int> result;

  if (value.empty()) return result;
  int pos = 0;
  while (pos != string::npos) {
    int spacepos = value.find (' ', pos);
    string entry;
    if (spacepos == string::npos)
      entry = value.substr (pos);
    else
      entry = value.substr (pos, spacepos - pos);
    int first;
    int second;
    int colonpos = entry.find (':');
    if (colonpos == string::npos) {
      first = atoi (entry.c_str());
      second = LONG_MAX;
    }
    else {
      first = atoi (entry.substr (0, colonpos).c_str());
      second = atol (entry.substr (colonpos + 1).c_str()) * 60;
    }
    result [second] = first;
    pos = spacepos;
    if (pos != string::npos) pos++;
  }
  return result;
}


map <string, string> ParseOptionString (const string & optstring) {

  map <string, string> result;

  if (optstring.empty()) return result;
  int pos = 0;
  while (pos != string::npos) {
    int semicpos = optstring.find (';', pos);
    string entry;
    if (semicpos == string::npos)
      entry = optstring.substr (pos);
    else
      entry = optstring.substr (pos, semicpos - pos);
    string first;
    string second;
    int equalpos = entry.find ('=');
    if (equalpos == string::npos) {
      first = entry;
      second = "";
    }
    else {
      first = entry.substr (0, equalpos);
      second = entry.substr (equalpos + 1);
    }
    // make the key case insensitive
    transform (first.begin(), first.end(), first.begin(), to_lower);
    result [first] = second;
    pos = semicpos;
    if (pos != string::npos) pos++;
  }
  return result;
}


bool GetBooleanOption (const map <string, string> & options,
		       const string & opt, bool * value) {

  const map <string, string>::const_iterator mssi = options.find (opt);
  if (mssi != options.end()) {
    if (mssi->second [0] == 'y' || mssi->second [0] == 'Y' || // yes
	mssi->second [0] == 't' || mssi->second [0] == 'T' || // true
	mssi->second [0] == '1') {
      *value = true;
      return true;
    }
    if (mssi->second [0] == 'n' || mssi->second [0] == 'N' || // no
	mssi->second [0] == 'f' || mssi->second [0] == 'F' || // false
	mssi->second [0] == '0') {
      *value = false;
      return true;
    }
  }
  return false;
}


void RemoveDefaultPort (string & url) {

  int pos = url.find ("://");
  if (pos == string::npos) return;
  string protocol = url.substr (0, pos);
  pos += 3;
  int slashpos = url.find ('/', pos);
  int colonpos = url.find (':', pos);
  if (colonpos == string::npos) return;
  if (slashpos != string::npos && slashpos < colonpos) return;
  string port;
  if (slashpos == string::npos)
    port = url.substr (colonpos + 1);
  else
    port = url.substr (colonpos + 1, slashpos - colonpos - 1);
  if ((protocol == "ftp" && port == "21") ||
      (protocol == "http" && port == "80") ||
      (protocol == "sftp" && port == "115") ||
      (protocol == "https" && port == "443") ||
      (protocol == "gsiftp" && port == "2811")) {
    if (slashpos == string::npos)
      url.erase (colonpos);
    else
      url.erase (colonpos, slashpos - colonpos);
  }
}


string UnX509 (const string & escaped) {

  string unescaped (escaped);
  int pos = 0;
  while ((pos = unescaped.find ("\\x", pos)) != string::npos) {
#ifdef HAVE_SSTREAM
    stringstream ss (unescaped.substr (pos + 2, 2));
#else
    strstream ss;
    ss << unescaped.substr (pos + 2, 2) << ends;
#endif
    int i;
    ss >> hex >> i;
    unescaped.replace (pos, 4, 1, i);
  }
  return unescaped;
}
