#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <vector>
#include <map>

#include <globus_common.h>
#include <globus_rsl.h>

#ifndef UI_XRSL
#define UI_XRSL

class Target;

class Xrsl {

 public:
  Xrsl (const string & xrslstring);
  Xrsl (const string & xrslfile, int dummy);
  Xrsl (const Xrsl & axrsl);
  Xrsl (globus_rsl_t * axrsl);
  ~Xrsl ();
  Xrsl & operator= (const Xrsl & axrsl);
  bool operator! () const;
  bool valid() const;
  Xrsl operator[] (const int ix) const;
  void Print () const;
  const string str () const;
  void AddSimpleRelation (const string & attr, const string & value);
  void AddSimpleRelationSet (const string & attr,
			     const vector <string> & values, bool select);
  int RemoveRelation (const string & attr, globus_rsl_t * axrsl = NULL);
  int UpdateAttribute (const string & attr, const string & value);
  int Check (globus_rsl_t * axrsl = NULL) const;
  int FixJoin ();
  int FixInOut (const string & stdattr, const string & inoutattr);
  int FixExec ();
  int FixStdin ();
  int FixTime (const string & attr);
  int FixPeriod (const string & attr);
  int PrepareUpload (vector <string> & filelist);
  int Collect (Target & target, const string & attr,
	       globus_rsl_t * axrsl = NULL);
  int GetJobName (string & jobname) const;
  string GetName () const;
  int GetCount (int * count) const;
  int GetCpuCount () const;
  int GetCpuTime (long int * time) const;
  int GetGridTime (long int * time) const;
  int GetBenchmarks (map<string, pair<float, long int> > & bmrk) const;
  int GetDisk (long long int * disk) const;
  int GetClientXrsl (string & xrslstring) const;
  int GetInputFiles (vector <string> & filelist) const;
  int GetRc (string & rc) const;
  int GetDefaultCache (bool * cache) const;
  void Eval ();
  int Test (Target & target, string & failattr,
	    globus_rsl_t * axrsl = NULL) const;
  string TestTarget (Target & target) const;
  // "-1" indicates error
 private:
  int FindRelation (const string & attr, globus_rsl_t ** relation,
		    globus_rsl_t * axrsl = NULL) const;
  globus_list_t ** FindHead ();
  globus_rsl_t * xrsl;
};

#endif
