#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dlfcn.h>
#include <getopt.h>
#include <string>
#include <vector>
#include <iostream>
#include <fstream>

#include "ngui.h"


void wrapper (char ** argv);


void ngusage (const string & progname, const char * optstring) {

  bool havegiisurlopt = false;
  bool havediropt = false;
  bool havestatusopt = false;

  cout << "Usage: " << progname << " [options]";
  if (progname != "ngsync")
    if (progname == "ngsub")
      cout << " [filename ...]";
    else
      cout << " [job ...]";
  cout << endl << endl;
  cout << "Options:" << endl;
  for (int i = 0; optstring [i]; i++) {
    if (optstring [i] == ':') continue;
    if (optstring [i + 1] == ':') {
      switch (optstring [i]) {
      case 'c':
	cout << "  -c, -cluster   [-]name       "
	     << "explicity select or reject a specific cluster" << endl;
	break;
      case 'C':
	cout << "  -C, -clustlist [-]filename   "
	     << "list of clusters to select or reject" << endl;
	break;
      case 'd':
	cout << "  -d, -debug     debuglevel    "
	     << "0 = none, 1 = some, 2 = more, 3 = a lot" << endl;
	break;
      case 'D':
	cout << "      -dir                     "
	     << "download directory (the job directory will be" << endl;
	cout << "                               "
	     << "  created in this directory)" << endl;
	havediropt = true;
 	break;
      case 'e':
	cout << "  -e, -xrsl      xrslstring    "
	     << "xrslstring describing the job to be submitted" << endl;
	break;
      case 'f':
	cout << "  -f, -file      filename      "
	     << "xrslfile describing the job to be submitted" << endl;
	break;
      case 'g':
	cout << "  -g, -giisurl   url           "
	     << "url to a central GIIS" << endl;
	havegiisurlopt = true;
	break;
      case 'G':
	cout << "  -G, -giislist  filename      "
	     << "list of GIIS urls" << endl;
	break;
      case 'i':
	cout << "  -i, -joblist   filename      "
	     << "file containing a list of jobids" << endl;
	break;
      case 'k':
	cout << "  -k, -kluster   [-]name       "
	     << "explicity select or reject a specific cluster" << endl;
	cout << "                               "
	     << "  as resubmission target" << endl;
	break;
      case 'K':
	cout << "  -K, -klustlist [-]filename   "
	     << "list of clusters to select or reject as" << endl;
	cout << "                               "
	     << "  resubmission target" << endl;
	break;
      case 'o':
	cout << "  -o, -joblist   filename      "
	     << "file where the jobids will be stored" << endl;
	break;
      case 's':
	cout << "  -s, -status    statusstr     "
	     << "only select jobs whose status is statusstr" << endl;
	havestatusopt = true;
	break;
      case 't':
	cout << "  -t, -timeout   time          "
	     << "timeout for MDS queries in seconds (default 40)" << endl;
	break;
      }
    }
    else {
      switch (optstring [i]) {
      case 'a':
	cout << "  -a, -all                     "
	     << "all jobs" << endl;
	break;
      case 'D':
	cout << "      -dryrun                  "
	     << "add dryrun option to the xrsl" << endl;
	break;
      case 'e':
	cout << "  -e, -stderr                  "
	     << "show stderr of the job" << endl;
	break;
      case 'f':
	if (progname == "ngclean") {
	  cout << "  -f, -force                   "
	       << "removes the job from the local list of jobs" << endl;
	  cout << "                               "
	       << "  even if the job is not found in the MDS" << endl;
	}
	else {
	  cout << "  -f, -force                   "
	       << "don't ask for verification" << endl;
	}
	break;
      case 'h':
	cout << "  -h, -help                    "
	     << "print this help" << endl;
	break;
      case 'j':
	cout << "  -j, -usejobname              "
	     << "use the jobname instead of the short ID as the" << endl;
	cout << "                               "
	     << "  job directory name" << endl;
	break;
      case 'l':
	if (progname == "ngcat") {
	  cout << "  -l, -gmlog                   "
	       << "show the grid manager's error log of the job" << endl;
	}
	else {
	  cout << "  -l, -long                    "
	       << "long format (more information)" << endl;
	}
	break;
      case 'o':
	cout << "  -o, -stdout                  "
	     << "show stdout of the job (default)" << endl;
	break;
      case 'P':
	cout << "      -dumpxrsl                "
	     << "do not submit - dump transformed xrsl to stdout" << endl;
	break;
      case 'q':
	cout << "  -q, -queues                  "
	     << "show information about clusters and queues" << endl;
	break;
      case 'r':
	cout << "      -keep                    "
	     << "keep files on gatekeeper (do not clean)" << endl;
	break;
      case 'v':
	cout << "  -v, -version                 "
	     << "print version information" << endl;
	break;
      case 'x':
	cout << "  -x, -anonymous               "
	     << "use anonymous bind for MDS queries (default)" << endl;
	break;
      case 'X':
	cout << "  -X, -gsi                     "
	     << "use gsi-gssapi bind for MDS queries" << endl;
	break;
      }
    }
  }
  if (progname != "ngsync") {
    cout << endl;
    cout << "Arguments:" << endl;
    if (progname == "ngsub")
      cout << "  filename ...                 "
	   << "xrslfiles describing the jobs to be submitted" << endl;
    else
      cout << "  job ...                      "
	   << "list of jobids and/or jobnames" << endl;
  }
  if (havegiisurlopt) {
    cout << endl;
    cout << "A url to a giis has the format ldap://hostname[:port]/basedn" << endl;
    cout << "If the port is omitted the default 2135 is used" << endl;
    cout << endl;
    cout << "If neither the -giisurl nor the -giislist option is given a list of urls is" << endl;
    cout << "read from the first existing file in the following list:" << endl;
    cout << endl;
    cout << "  1. ${HOME}/.nggiislist" << endl;
    cout << "  2. ${NORDUGRID_LOCATION}/etc/giislist" << endl;
    cout << "  3. /etc/giislist" << endl;
  }
  if (havestatusopt) {
    cout << endl;
    cout << "For finished jobs the status option will match FINISHED or FAILED depending on" << endl;
    cout << "weather the job finished successfully or not" << endl;
  }
  cout << endl;
  cout << "Default values for the MDS timeout or the debuglevel can be set by defining" << endl;
  cout << "the environment variables NGTIMEOUT or NGDEBUG, or if these are undefined by" << endl;
  cout << "defining NGTIMEOUT or NGDEBUG in ${HOME}/.ngrc" << endl;
  if (havediropt) {
    cout << endl;
    cout << "A default value for the download directory other than the current working" << endl;
    cout << "directory can be set by defining the environment variable NGDOWNLOAD, or if" << endl;
    cout << "this is undefined by defining NGDOWNLOAD in ${HOME}/.ngrc" << endl;
  }
}


int ReadList (vector <string> & vs, const string & filename) {

  ifstream listfile (filename.c_str());
  if (!listfile) {
    cerr << "Error: Can not open list file \"" << filename << "\""
	 << endl;
    return 1;
  }
  string line;
  while (getline (listfile, line)) {
    if (line.empty()) continue;
    if (line[0] == '#') continue;
    vs.push_back (line);
  }
  listfile.close();
  return 0;
}


int main (int argc, char ** argv) {

  if (!getenv ("NORDUGRID_ENVIRONMENT_SETUP")) wrapper (argv);

  char * pos = strrchr (argv[0], '/');
  if (pos) argv[0] = pos + 1;
  string progname = argv[0];

  char * optstring;

  if (progname == "ngcat")
    optstring = "ai:c:C:s:oelt:d:xXvh";
  else if (progname == "ngclean")
    optstring = "ai:c:C:s:ft:d:xXvh";
  else if (progname == "ngget")
    optstring = "ai:c:C:s:D:jrt:d:xXvh";
  else if (progname == "ngkill")
    optstring = "ai:c:C:s:rt:d:xXvh";
  else if (progname == "ngrenew")
    optstring = "ai:c:C:s:t:d:xXvh";
  else if (progname == "ngresub")
    optstring = "ai:c:C:s:k:K:g:G:o:DPrt:d:xXvh";
  else if (progname == "ngstat")
    optstring = "ai:c:C:s:g:G:qlt:d:xXvh";
  else if (progname == "ngsub")
    optstring = "c:C:g:G:e:f:o:DPt:d:xXvh";
  else if (progname == "ngsync")
    optstring = "c:C:g:G:ft:d:xXvh";
  else {
    cerr << "Unknown program name: " << progname << endl;
    return 1;
  }

  struct option longoptions[20];
  int nopt = 0;
  for (int i = 0; optstring [i]; i++) {
    if (optstring [i] == ':') continue;
    if (optstring [i + 1] == ':') {
      switch (optstring [i]) {
      case 'c':
	{
	  struct option lopt = {"cluster", 1, 0, 'c'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'C':
	{
	  struct option lopt = {"clustlist", 1, 0, 'C'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'd':
	{
	  struct option lopt = {"debug", 1, 0, 'd'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'D':
	{
	  struct option lopt = {"dir", 1, 0, 'D'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'e':
	{
	  struct option lopt = {"xrsl", 1, 0, 'e'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'f':
	{
	  struct option lopt = {"file", 1, 0, 'f'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'g':
	{
	  struct option lopt = {"giisurl", 1, 0, 'g'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'G':
	{
	  struct option lopt = {"giislist", 1, 0, 'G'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'i':
	{
	  struct option lopt = {"joblist", 1, 0, 'i'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'k':
	{
	  struct option lopt = {"kluster", 1, 0, 'c'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'K':
	{
	  struct option lopt = {"klustlist", 1, 0, 'C'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'o':
	{
	  struct option lopt = {"joblist", 1, 0, 'o'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 's':
	{
	  struct option lopt = {"status", 1, 0, 's'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 't':
	{
	  struct option lopt = {"timeout", 1, 0, 't'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      }
    }
    else {
      switch (optstring [i]) {
      case 'a':
	{
	  struct option lopt = {"all", 0, 0, 'a'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'D':
	{
	  struct option lopt = {"dryrun", 0, 0, 'D'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'e':
	{
	  struct option lopt = {"stderr", 0, 0, 'e'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'f':
	{
	  struct option lopt = {"force", 0, 0, 'f'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'h':
	{
	  struct option lopt = {"help", 0, 0, 'h'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'j':
	{
	  struct option lopt = {"usejobname", 0, 0, 'j'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'l':
	{
	  if (progname == "ngcat") {
	    struct option lopt = {"gmlog", 0, 0, 'l'};
	    longoptions [nopt++] = lopt;
	  }
	  else {
	    struct option lopt = {"long", 0, 0, 'l'};
	    longoptions [nopt++] = lopt;
	  }
	  break;
	}
      case 'o':
	{
	  struct option lopt = {"stdout", 0, 0, 'o'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'P':
	{
	  struct option lopt = {"dumpxrsl", 0, 0, 'P'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'q':
	{
	  struct option lopt = {"queues", 0, 0, 'q'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'r':
	{
	  struct option lopt = {"keep", 0, 0, 'r'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'v':
	{
	  struct option lopt = {"version", 0, 0, 'v'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'x':
	{
	  struct option lopt = {"anonymous", 0, 0, 'x'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      case 'X':
	{
	  struct option lopt = {"gsi", 0, 0, 'X'};
	  longoptions [nopt++] = lopt;
	  break;
	}
      }
    }
  }

  struct option lopt = {0, 0, 0, 0};
  longoptions [nopt++] = lopt;

  // common options
  vector <string> clusterreject;
  vector <string> clusterselect;
  vector <string> giisurls;
  bool all = false;
  vector <string> status;
  vector <string> joblists;
  int timeout = UNDEFINED;
  int debug = UNDEFINED;
  bool anonymous = true;
  // ngcat options
  int whichfile = STDOUT;
  // ngclean & ngsync options
  bool force = false;
  // ngget, ngkill & ngresub options
  bool keep = false;
  // ngget options
  string directory;
  bool usejobname = false;
  // ngstat options
  bool clusters = false;
  bool longlist = false;
  // ngsub options
  vector <string> xrslstrings;
  // ngresub options
  vector <string> klusterreject;
  vector <string> klusterselect;
  // ngsub & ngresub options
  string joblist;
  bool dryrun = false;
  bool dumpxrsl = false;

  vector <string> params;

  int opt;
  while (opt != -1) {
    opt = getopt_long_only (argc, argv, optstring, longoptions, NULL);
    if (opt == -1) continue;
    if (optarg) {
      switch (opt) {
      case 'c':  // cluster
	if (optarg[0] == '-')
	  clusterreject.push_back (&optarg[1]);
	else if (optarg[0] == '+')
	  clusterselect.push_back (&optarg[1]);
	else
	  clusterselect.push_back (optarg);
	break;
      case 'C':  // clustlist
	if (optarg[0] == '-') {
	  if (ReadList (clusterreject, &optarg[1])) return 1;
	}
	else if (optarg[0] == '+') {
	  if (ReadList (clusterselect, &optarg[1])) return 1;
	}
	else {
	  if (ReadList (clusterselect, optarg)) return 1;
	}
	break;	
      case 'd':  // debug
	debug = atoi (optarg);
	break;
      case 'D':  // dir
	directory = optarg;
	break;
      case 'e':  // xrsl
	xrslstrings.push_back (optarg);
	break;
      case 'f':  // file
	params.push_back (optarg);
	break;
      case 'g':  // giisurl
	giisurls.push_back (optarg);
	break;
      case 'G':  // giislist
	if (ReadList (giisurls, optarg)) return 1;
	break;
      case 'i':  // joblist (in)
	joblists.push_back (optarg);
	break;
      case 'k':  // kluster
	if (optarg[0] == '-')
	  klusterreject.push_back (&optarg[1]);
	else if (optarg[0] == '+')
	  klusterselect.push_back (&optarg[1]);
	else
	  klusterselect.push_back (optarg);
	break;
      case 'K':  // klustlist
	if (optarg[0] == '-') {
	  if (ReadList (klusterreject, &optarg[1])) return 1;
	}
	else if (optarg[0] == '+') {
	  if (ReadList (klusterselect, &optarg[1])) return 1;
	}
	else {
	  if (ReadList (klusterselect, optarg)) return 1;
	}
	break;	
      case 'o':  // joblist (out)
	joblist = optarg;
	break;
      case 's':  // status
	status.push_back (optarg);
	break;
      case 't':  // timeout
	timeout = atoi (optarg);
	break;
      default:
	cerr << "Use \"" << progname << " -help\" for help" << endl;
	return 1;
      }
    }
    else {
      switch (opt) {
      case 'a':  // all
	all = true;
	break;
      case 'D':  // dryrun
	dryrun = true;
	break;
      case 'e':  // stderr
	whichfile = STDERR;
	break;
      case 'f':  // force
	force = true;
	break;
      case 'h':  // help
	ngusage (progname, optstring);
	return 0;
      case 'j':  // usejobname
	usejobname = true;
	break;
      case 'l':  // gmlog or long
	if (progname == "ngcat")
	  whichfile = GMLOG;
	else
	  longlist = true;
	break;
      case 'o':  // stdout
	whichfile = STDOUT;
	break;
      case 'P':  // dumpxrsl
	dumpxrsl = true;
	break;
      case 'q':  // queues
	clusters = true;
	break;
      case 'r':  // keep
	keep = true;
	break;
      case 'v':  // version
	cout << progname << ": version " << VERSION << endl;
	return 0;
      case 'x':  // anonymous
	anonymous = true;
	break;
      case 'X':  // gsi
	anonymous = false;
	break;
      default:
	cerr << "Use \"" << progname << " -help\" for help" << endl;
	return 1;
      }
    }
  }

  while (argc > optind) params.push_back (argv [optind++]);

  if (progname == "ngsub" && params.empty() && xrslstrings.empty()) {
    string xrsl;
    string line;
    while (getline (cin, line) && line != ".") xrsl += line;
    xrslstrings.push_back (xrsl);
  }

  void * plugin = dlopen ("libngui.so", RTLD_NOW);
  if (!plugin) {
    cerr << dlerror() << endl;
    return 1;
  }

  ngversiontype ngversion = (ngversiontype) dlsym (plugin, "ngversion");
  if (!ngversion) {
    cerr << dlerror() << endl;
    cerr << "Wrong version of libngui.so found" << endl;
    dlclose (plugin);
    return 1;
  }

  if (strcmp ((*ngversion)(), VERSION) != 0) {
    cerr << "Different versions:" << endl;
    cerr << "  " << progname << " binary from nordugrid-" << VERSION << endl;
    cerr << "  libngui.so library from nordugrid-" << (*ngversion)() << endl;
    dlclose (plugin);
    return 1;
  }

  int retval;

  if (progname == "ngcat") {
    ngcatxxtype ngcat = (ngcatxxtype) dlsym (plugin, "ngcatxx");
    retval = (*ngcat) (params, joblists, clusterselect, clusterreject, status,
		       all, whichfile, timeout, debug, anonymous);
  }
  else if (progname == "ngclean") {
    ngcleanxxtype ngclean = (ngcleanxxtype) dlsym (plugin, "ngcleanxx");
    retval = (*ngclean) (params, joblists, clusterselect, clusterreject,
			 status, all, force, timeout, debug, anonymous);
  }
  else if (progname == "ngget") {
    nggetxxtype ngget = (nggetxxtype) dlsym (plugin, "nggetxx");
    retval = (*ngget) (params, joblists, clusterselect, clusterreject, status,
		       all, directory, usejobname, keep, timeout, debug,
		       anonymous);
  }
  else if (progname == "ngkill") {
    ngkillxxtype ngkill = (ngkillxxtype) dlsym (plugin, "ngkillxx");
    retval = (*ngkill) (params, joblists, clusterselect, clusterreject, status,
			all, keep, timeout, debug, anonymous);
  }
  else if (progname == "ngrenew") {
    ngrenewxxtype ngrenew = (ngrenewxxtype) dlsym (plugin, "ngrenewxx");
    retval = (*ngrenew) (params, joblists, clusterselect, clusterreject,
			 status, all, timeout, debug, anonymous);
  }
  else if (progname == "ngresub") {
    ngresubxxtype ngsub = (ngresubxxtype) dlsym (plugin, "ngresubxx");
    retval = (*ngsub) (params, joblists, clusterselect, clusterreject,
		       status, all, klusterselect, klusterreject, giisurls,
		       joblist, dryrun, dumpxrsl, keep, timeout, debug,
		       anonymous);
  }
  else if (progname == "ngstat") {
    ngstatxxtype ngstat = (ngstatxxtype) dlsym (plugin, "ngstatxx");
    retval = (*ngstat) (params, joblists, clusterselect, clusterreject, status,
			all, giisurls, clusters, longlist, timeout, debug,
			anonymous);
  }
  else if (progname == "ngsub") {
    ngsubxxtype ngsub = (ngsubxxtype) dlsym (plugin, "ngsubxx");
    retval = (*ngsub) (params, xrslstrings, clusterselect, clusterreject,
		       giisurls, joblist, dryrun, dumpxrsl, timeout, debug,
		       anonymous);
  }
  else if (progname == "ngsync") {
    ngsyncxxtype ngsync = (ngsyncxxtype) dlsym (plugin, "ngsyncxx");
    retval = (*ngsync) (clusterselect, clusterreject, giisurls, force, timeout,
			debug, anonymous);
  }

  // Avoid segmentations faults with glibc < 2.2.4
  // dlclose (plugin);

  return retval;
}
