#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <string>
#include <vector>
#include <unistd.h>
#include <stdio.h>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"

#define __UI_LIBRARY__
#include "ui_downloader.h"

#include "ngui.h"


extern "C"
int ngcatxx (const vector <string> & jobs,
	     const vector <string> & joblists,
	     const vector <string> & clusterselect_,
	     const vector <string> & clusterreject_,
	     const vector <string> & status,
	     const bool all,
	     const int whichfile,
	     int timeout,
	     int debug,
	     const bool anonymous) {

  vector <string> clusterselect = ResolveAliases (clusterselect_);
  vector <string> clusterreject = ResolveAliases (clusterreject_);

  vector <string> jobids;
  vector <Cluster> clusterlist;

  int error = 0;
  ActivateGlobus();

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) {
    error = 1;
    goto errorexit;
  }

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft()) << endl;
  }

  // finding the information

  jobids = GetJobIDs (jobs, joblists, clusterselect, clusterreject, status,
		      all);

  if (all && jobids.empty()) {
    cout << "ngcat: No jobs" << endl;
    goto errorexit;
  }

  if (jobids.empty()) {
    cerr << "ngcat: no valid jobids given" << endl;
    cerr << "Use \"ngcat -help\" for help" << endl;
    error = 1;
    goto errorexit;
  }

  for (vector <string>::iterator vsi = jobids.begin();
       vsi != jobids.end(); vsi++) {
    int pos0 = 0;
    if (vsi->substr (0, 9) == "gsiftp://") pos0 = 9;
    int pos = vsi->find_first_of (":/", pos0);
    if (pos == string::npos) {
      cerr << "ngcat: invalid jobid: " << *vsi << endl;
      error = 1;
      continue;
    }
    string clustername = vsi->substr (pos0, pos - pos0);
    bool found = false;
    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 !found && cli != clusterlist.end(); cli++)
      if (clustername == cli->GetName()) found = true;
    if (!found) clusterlist.push_back (clustername);
  }

  if (clusterlist.empty()) {
    cerr << "ngcat: no valid jobids given" << endl;
    error = 1;
    goto errorexit;
  }

  FindClusterInfo (clusterlist, Mds::JobManipulation, user.GetSNx(), anonymous,
		   timeout, debug);

  for (vector <string>::iterator vsi = jobids.begin();
       vsi != jobids.end(); vsi++) {

    Cluster * c = NULL;
    string file;
    string jobstatus;

    bool found = false;
    bool selected = false;

    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 !found && cli != clusterlist.end(); cli++)
      for (vector <Queue>::iterator qli = cli->queues.begin();
	   !found && qli != cli->queues.end(); qli++)
	for (vector <Job>::iterator jli = qli->jobs.begin();
	     !found && jli != qli->jobs.end(); jli++)
	  if (*vsi == jli->GetId()) {
	    if (jli->IsStatus (status)) {
	      selected = true;
	      c = &*cli;
	      jobstatus = jli->GetStatus();
	      if (whichfile == STDOUT)
		file = jli->GetStdout();
	      else if (whichfile == STDERR)
		file = jli->GetStderr();
	      else if (whichfile == GMLOG)
		file = "errors";
	    }
	    found = true;
	  }

    if (!found) {
      cerr << "ngcat: no job with jobid " << *vsi << " found" << endl;
      error = 1;
      continue;
    }
    if (!selected)
      continue;
    if (jobstatus == "DELETED") {
      cerr << "ngcat: job " << *vsi << " has already been deleted" << endl;
      error = 1;
      continue;
    }
    if ((whichfile == STDOUT || whichfile == STDERR) &&
	(jobstatus == "ACCEPTED" || jobstatus == "PENDING: ACCEPTED" ||
	 jobstatus == "PREPARING" || jobstatus == "PENDING: PREPARING" ||
	 jobstatus.substr (0, 9) == "INLRMS: Q")) {
      cerr << "ngcat: job " << *vsi << " has not started yet" << endl;
      error = 1;
      continue;
    }
    if (file.empty()) {
      cerr << "ngcat: cannot determine the name of the "
	   << ((whichfile == STDOUT) ? "stdout" :
	       ((whichfile == STDERR) ? "stderr" :
		((whichfile == GMLOG) ? "grid manager log directory" :
		 "ERROR")))
	   << " for " << *vsi << endl;
      error = 1;
      continue;
    }
    int pos = vsi->rfind ('/');
    if (pos == string::npos) {
      cerr << "ngcat: invalid jobid: " << *vsi << endl;
      error = 1;
      continue;
    }
    string shortid = vsi->substr (pos + 1);

    vector <string> filenames;

    filenames.push_back (file);

    int err = ui_downloader ((c->GetContact() +
			      (whichfile == GMLOG ? "/info/" : "/")
			      + shortid).c_str(), false,
			     ("/tmp/" + shortid).c_str(), filenames, true,
			     false, debug, timeout);

    if (err) {
      cerr << "ngcat: The "
	   << ((whichfile == STDOUT) ? "stdout" :
	       ((whichfile == STDERR) ? "stderr" :
		((whichfile == GMLOG) ? "grid manager error log" : "ERROR")))
	   << " of the job could not be found." << endl;
      cerr << "It either hasn't been created yet or it has been successfully moved to the" << endl;
      cerr << "location specified in the job's xrsl description." << endl;
      error = 1;
      continue;
    }

    system (("cat /tmp/" + shortid + '/' + file).c_str());
    remove (("/tmp/" + shortid + '/' + file).c_str());
    pos = file.rfind ('/');
    while (pos != string::npos) {
      rmdir (("/tmp/" + shortid + '/' + file.substr (0, pos)).c_str());
      pos = file.rfind ('/', pos - 1);
    }
    rmdir (("/tmp/" + shortid).c_str());
  }

 errorexit:
  DeactivateGlobus();
  return error;
}
