/* ngjmlib.cpp - see ngjmlib.h for futher information */
/* Copyright 2003 Henrik Thostrup Jensen and Jesper Ryge Leth 
 * All rights reserved.
 *
 * This file is part of NGProxy.
 *	
 * NGProxy is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * NGProxy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NGProxy; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <vector>
#include <ldap.h>

#include "LdapQuery.h"
#include "MdsQuery.h"
#include "CertInfo.h"

#include "ngjmlib.h"

string getClusterName(const string jobId) {
	
	string clusterName = "UNKNOWN";
	int i;

	// Sample JOBID string:	gsiftp://francis.auc.dk:2811/jobs/1725880770104498049
	int result = strncmp(jobId.c_str(),"gsiftp://",9);
	if (result) {
		cerr << "This does not appear to start with gsiftp.. bailing out" << endl;
	} // FIXME: we're in dire need of some exception handling...
	else {
		i = jobId.find("/", 9);							  // start after "gsiftp://" 9 characters 
		clusterName = jobId.substr(9, i - 9 - 5);   // -5 because we don't want the port number either		
	}
	return clusterName;
}

string getJobStatus(const string jobId) 
{	
	int timeout = 40; 		// timeout to servers - default = 40
	int debug = 1;				// debug level
	bool anonymous = true;	// anon when connecting to mds (i think)
	CertInfo userProxy;		// reference to user proxy
	string clusterName, jobStatus;
	
	clusterName = getClusterName(jobId);
	
	Cluster myCluster(clusterName);
	vector<Cluster>myClusterList;
	myClusterList.push_back(myCluster);

	// FIXME: there must be an easier way to get jobinfo when we have the job id - maybe a direct ldap query	
	FindClusterInfo(myClusterList, Mds::JobInfo, userProxy.GetSN(), anonymous, timeout,	debug);
		
	for (vector<Cluster>::iterator cli = myClusterList.begin(); cli != myClusterList.end(); cli++) {
		for (vector<Queue>::iterator qli = cli->queues.begin(); qli != cli->queues.end(); qli++) {
			for (vector<Job>::iterator jli = qli->jobs.begin(); jli != qli->jobs.end(); jli++) {
				if (jli->GetId() == jobId) {
				jobStatus = jli->GetStatus();
				}
			}
		}
	}

	return jobStatus;
}


Job* getJob(const string jobId) 
{	
	int timeout = 40;        // timeout to servers - default = 40
	int debug = 1;           // debug level
	bool anonymous = true;   // anon when connecting to mds (i think)
	CertInfo userProxy;      // user proxy object
	string clusterName;
	Job* job;
	
	clusterName = getClusterName(jobId);
	
	Cluster myCluster(clusterName);
	vector<Cluster>myClusterList;
	myClusterList.push_back(myCluster);

	// FIXME: there must be an easier way to get jobinfo when we have the job id - maybe a direct ldap query	
	FindClusterInfo(myClusterList, Mds::JobInfo, userProxy.GetSN(), anonymous, timeout,	debug);
		
	for (vector<Cluster>::iterator cli = myClusterList.begin(); cli != myClusterList.end(); cli++) {
		for (vector<Queue>::iterator qli = cli->queues.begin(); qli != cli->queues.end(); qli++) {
			for (vector<Job>::iterator jli = qli->jobs.begin(); jli != qli->jobs.end(); jli++) {
				if (jli->GetId() == jobId) {
					//job = &(*jli);
					job = new Job(*jli);
				}
			}
		}
	}

	return job;
}


vector<Job> getJobStatus(vector<string> jobIdList) {

	string clusterName;
	vector<string> clusterNameList;
	vector<Cluster> clusterList;
	vector<Job> jobs;
	
	int timeout = 40; 		// timeout to servers - default = 40
	int debug = 1;				// debug level
	bool anonymous = true;	// anon when connecting to mds (i think)
	CertInfo userProxy;		// reference to user proxy
	
	// Populate the clusterList vith names extracted from job ID's but oly add every cluster once
	for (vector<string>::iterator jli = jobIdList.begin(); jli != jobIdList.end(); jli++) {
		bool insert = true;
		string clusterName = getClusterName(*jli);
		
		// check if the name is in the list
		for (vector<string>::iterator cni = clusterNameList.begin(); cni != clusterNameList.end(); cni++) {
			if ( *cni == clusterName )
				insert = false;
		}
		// only insert if the name is not already in the list
		if (insert)
			clusterNameList.push_back(clusterName);
	}
	
	// Generate a vector containing the according cluster objects
	for (vector<string>::iterator cni = clusterNameList.begin(); cni != clusterNameList.end(); cni++) 
		clusterList.push_back(Cluster::Cluster(*cni));
	
	FindClusterInfo(clusterList, Mds::JobInfo, userProxy.GetSN(), anonymous, timeout,	debug);
	
	for (vector<Cluster>::iterator cli = clusterList.begin(); cli != clusterList.end(); cli++) {
		for (vector<Queue>::iterator qli = cli->queues.begin(); qli != cli->queues.end(); qli++) {
			for (vector<Job>::iterator jli = qli->jobs.begin(); jli != qli->jobs.end(); jli++) {
				for (vector<string>::iterator jobIdI = jobIdList.begin(); jobIdI != jobIdList.end(); jobIdI++) {
					if (jli->GetId() == (*jobIdI))
						jobs.push_back((Job) *jli);
				}
			}
		}
	}
	return jobs;
} // getJobStatus
