#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <stdio.h>

#include "Preferences.h"
#include "CertInfo.h"
#include "DateTime.h"
#include "MdsQuery.h"
#include "Giis.h"

#include "ngui.h"


extern "C"
int ngsyncxx (const vector <string> & clusterselect_,
	      const vector <string> & clusterreject_,
	      const vector <string> & giisurls,
	      const bool force,
	      int timeout,
	      int debug,
	      const bool anonymous) {

  vector <string> clusterselect = ResolveAliases (clusterselect_);
  vector <string> clusterreject = ResolveAliases (clusterreject_);

  vector <Giis> giislist;
  vector <Cluster> clusterlist;
  multimap <string, string> ngjobmap;

  int error = 0;
  ActivateGlobus();

  // setting defaults

  if (timeout == UNDEFINED) timeout = iGetEnv ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = iGetDef ("NGTIMEOUT");
  if (timeout == UNDEFINED) timeout = DEFAULT_TIMEOUT;

  if (debug == UNDEFINED) debug = iGetEnv ("NGDEBUG");
  if (debug == UNDEFINED) debug = iGetDef ("NGDEBUG");
  if (debug == UNDEFINED) debug = 0;

  // get user info

  CertInfo user;
  if (!user) {
    error = 1;
    goto errorexit;
  }

  if (debug) {
    cout << "User subject name: " << user.GetSN() << endl;
    cout << "Remaining proxy lifetime: "
	 << Period (user.TimeLeft()) << endl;
  }

  if (!force) {
    cout << "Synchronizing the local list of active jobs with the information in the MDS" << endl;
    cout << "can result in some inconsistencies. Very recently submitted jobs might not yet" << endl;
    cout << "be present in the MDS information, whereas jobs very recently scheduled for" << endl;
    cout << "deletion can still be present." << endl;
    cout << "Are you sure you want to synchronize your local job list? [y/n] ";
    string response;
    cin >> response;
    if (response != "y" && response != "Y") {
      cout << "Cancelling synchronization request" << endl;
      goto errorexit;
    }
  }

  // find available clusters

  for (vector <string>::const_iterator vsi = clusterselect.begin();
       vsi != clusterselect.end(); vsi++) {
    bool found = false;
    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 !found && cli != clusterlist.end(); cli++)
      if (*vsi == cli->GetName()) found = true;
    if (!found) clusterlist.push_back (*vsi);
  }

  if (clusterlist.empty()) {
    if (GetGiises (giisurls, giislist)) {
      error = 1;
      goto errorexit;
    }

    clusterlist = FindClusters (giislist, user.GetSNx(), anonymous,
				timeout, debug);
    if (clusterlist.empty()) {
      cerr << "ngsync: could not retrieve cluster list from giis" << endl;
      error = 1;
      goto errorexit;
    }
  }

  for (vector <string>::const_iterator vsi = clusterreject.begin();
       vsi != clusterreject.end(); vsi++)
    for (vector <Cluster>::iterator cli = clusterlist.begin();
	 cli != clusterlist.end(); cli++)
      if (cli->GetName() == *vsi) {
	if (debug) cout << "Rejecting cluster: " << *vsi << endl;
	clusterlist.erase (cli);
	break;
      }

  FindClusterInfo (clusterlist, Mds::JobInfo, user.GetSNx(), anonymous,
		   timeout, debug);

  // finding the information

  for (vector <Cluster>::iterator cli = clusterlist.begin();
       cli != clusterlist.end(); cli++)
    for (vector <Queue>::iterator qli = cli->queues.begin();
	 qli != cli->queues.end(); qli++)
      for (vector <Job>::iterator jli = qli->jobs.begin();
	   jli != qli->jobs.end(); jli++)
	ngjobmap.insert (multimap <string, string>::value_type (jli->GetName(),
								jli->GetId()));

  if (!ngjobmap.empty()) {
    string filename = GetEnv ("HOME");
    filename.append ("/.ngjobs");
    ofstream ngjobs (filename.c_str());

    if (debug) cout << "Found the following jobs:" << endl;

    for (multimap <string, string>::iterator mmssi = ngjobmap.begin();
	 mmssi != ngjobmap.end(); mmssi++) {
      ngjobs << mmssi->second << '#' << mmssi->first << endl;
      if (debug) cout << mmssi->second << endl;
    }

    ngjobs.close();
  }
  else {
    cout << "No jobs found." << endl;

    string filename = GetEnv ("HOME");
    filename.append ("/.ngjobs");
    remove (filename.c_str());
  }

 errorexit:
  DeactivateGlobus();
  return error;
}
